Package: rEDM
Type: Package
Title: Empirical Dynamic Modeling ('EDM')
Version: 1.15.4
Date: 2024-04-05
Authors@R: c( person("Joseph", "Park", role = c("aut", "cre"),
                     email = "JosephPark@IEEE.org",
                     comment = c(ORCID = "0000-0001-5411-1409")),
              person("Cameron", "Smith", role = c("aut"),
                     email = "cos008@ucsd.edu",
                     comment = c(ORCID = "0000-0003-0020-5607")),
              person("George", "Sugihara", role = c("aut", "ccp"), 
                     comment = c(ORCID = "0000-0002-2863-6946")),
              person("Ethan", "Deyle", role = c("aut"),
                     comment = c(ORCID = "0000-0001-8704-8434")),
              person("Erik", "Saberski", role = c("ctb"),
                     comment = c(ORCID = "0000-0002-6475-6187")),
              person("Hao", "Ye", role = c("ctb"), 
                     comment = c(ORCID = "0000-0002-8630-1458")),
              person("The Regents of the University of California",
                     role = c("cph") ) )
Maintainer: Joseph Park <JosephPark@IEEE.org>
Description: An implementation of 'EDM' algorithms based on research software developed for internal use at the Sugihara Lab ('UCSD/SIO').  The package is implemented with 'Rcpp' wrappers around the 'cppEDM' library.  It implements the 'simplex' projection method from Sugihara & May (1990) <doi:10.1038/344734a0>, the 'S-map' algorithm from Sugihara (1994) <doi:10.1098/rsta.1994.0106>, convergent cross mapping described in Sugihara et al. (2012) <doi:10.1126/science.1227079>, and, 'multiview embedding' described in Ye & Sugihara (2016) <doi:10.1126/science.aag0863>.
License: BSD_2_clause + file LICENSE
LazyData: true
LazyLoad: yes
Imports: methods, Rcpp (>= 1.0.1)
LinkingTo: Rcpp, RcppThread
Suggests: knitr, rmarkdown, formatR
VignetteBuilder: knitr
NeedsCompilation: yes
Packaged: 2024-04-05 17:24:24 UTC; jpark
Author: Joseph Park [aut, cre] (<https://orcid.org/0000-0001-5411-1409>),
  Cameron Smith [aut] (<https://orcid.org/0000-0003-0020-5607>),
  George Sugihara [aut, ccp] (<https://orcid.org/0000-0002-2863-6946>),
  Ethan Deyle [aut] (<https://orcid.org/0000-0001-8704-8434>),
  Erik Saberski [ctb] (<https://orcid.org/0000-0002-6475-6187>),
  Hao Ye [ctb] (<https://orcid.org/0000-0002-8630-1458>),
  The Regents of the University of California [cph]
Repository: CRAN
Date/Publication: 2024-04-06 10:30:03 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 18:44:27 UTC; unix
Archs: rEDM.so.dSYM

Package: rEDM
Type: Package
Title: Empirical Dynamic Modeling ('EDM')
Version: 2.0.2
Date: 2026-08-21
Authors@R: c( person("Joseph", "Park", role = c("aut", "cre"),
                     email = "JosephPark@IEEE.org",
                     comment = c(ORCID = "0000-0001-5411-1409")),
              person("George", "Sugihara", role = c("aut", "ccp"), 
                     comment = c(ORCID = "0000-0002-2863-6946")),
              person("Ethan", "Deyle", role = c("aut"),
                     comment = c(ORCID = "0000-0001-8704-8434")),
              person("The Regents of the University of California",
                     role = c("cph") ) )
Maintainer: Joseph Park <JosephPark@IEEE.org>
Description: An implementation of 'EDM' algorithms based on research software developed at the Sugihara Lab ('UCSD/SIO'). Primary methods include 'Simplex' projection from Sugihara & May (1990) <doi:10.1038/344734a0>, Sequential locally-weighted global linear maps 'S-map': Sugihara (1994) <doi:10.1098/rsta.1994.0106>, Convergent cross mapping described in Sugihara et al. (2012) <doi:10.1126/science.1227079>, and, 'Multiview embedding' from Ye & Sugihara (2016) <doi:10.1126/science.aag0863>.
License: BSD_2_clause + file LICENSE
Encoding: UTF-8
LazyData: true
LazyLoad: yes
Imports: stats, utils, graphics, parallel, RANN
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, formatR
Config/testthat/edition: 3
RoxygenNote: 8.0.0
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-08-29 22:30:42 UTC; jpark
Author: Joseph Park [aut, cre] (ORCID: <https://orcid.org/0000-0001-5411-1409>),
  George Sugihara [aut, ccp] (ORCID:
    <https://orcid.org/0000-0002-2863-6946>),
  Ethan Deyle [aut] (ORCID: <https://orcid.org/0000-0001-8704-8434>),
  The Regents of the University of California [cph]
Repository: CRAN
Date/Publication: 2026-08-30 05:10:18 UTC

useDynLib( rEDM, .registration = TRUE )

export( Simplex   )
export( SMap      )
export( CCM       )
export( Multiview )
export( Embed     )
export( MakeBlock )
export( ComputeError     )
export( EmbedDimension   )
export( PredictInterval  )
export( PredictNonlinear )
export( SurrogateData    )

# Legacy functions
# export( ccm           )
# export( block_lnlp    )
# export( s_map         )
# export( simplex       )
# export( multiview     )
# export( make_block    )
# export( compute_stats )
# export( make_surrogate_data )

importFrom("grDevices", "dev.cur", "dev.new", "dev.list")
importFrom("graphics", "abline", "legend", "lines", "mtext", "par", "plot")
importFrom("utils", "data", "read.csv")
importFrom("stats", "fft", "predict", "rnorm", "runif", "sd", "smooth.spline", "cov", "pnorm")
import( methods )
import( Rcpp )

rEDM
====

Overview
--------

`rEDM` is a collection of methods for Empirical Dynamic Modeling (EDM).
EDM is based on the mathematical theory of reconstructing attractor manifolds
from time series data with applications to forecasting, causal inference, 
and more. It is based on research software developed at the 
[Sugihara Lab](https://deepeco.ucsd.edu/)
(University of California San Diego, Scripps Institution of Oceanography).

Empirical Dynamic Modeling (EDM)
-------------------------------------

Introduction and documentation are are avilable
[online](https://sugiharalab.github.io/EDM_Documentation/ "EDM Docs"), or
in the package [tutorial](https://github.com/SugiharaLab/rEDM/blob/d5aafe06573be73f603488f6ee4ae68a73da5e12/doc/rEDM-tutorial.pdf "rEDM turorial").

Functionality includes:

* Simplex projection (Sugihara and May 1990)
* Sequential Locally Weighted Global Linear Maps (S-map) (Sugihara 1994)
* Multivariate embeddings (Deyle and Sugihara 2011)
* Convergent cross mapping (Sugihara et. al. 2012)
* Multiview embedding (Ye and Sugihara 2016)

Installation
------------
To install from CRAN [rEDM](https://CRAN.R-project.org/package=rEDM):

    install.packages(rEDM)

Using R devtools for latest development version:

    install.packages("devtools")
    devtools::install_github("SugiharaLab/rEDM")

Building from source:

    git clone https://github.com/SugiharaLab/rEDM.git
    cd rEDM
    R CMD INSTALL .

Example
-------

We begin by looking at annual time series of sunspots:

    df = data.frame(yr = as.numeric(time(sunspot.year)), 
                     sunspot_count = as.numeric(sunspot.year))

    plot(df$yr, df$sunspot_count, type = "l", 
         xlab = "year", ylab = "sunspots")

![](man/figures/sunspots-1.png)

First, we use `EmbedDimension()` to determine the optimal embedding
dimension, E:

    library(rEDM)   # load the package
    # If you're new to the rEDM package, please consult the tutorial:
    # vignette("rEDM-tutorial")

    E.opt = EmbedDimension( dataFrame = df,    # input data
                            lib     = "1 280", # portion of data to train
                            pred    = "1 280", # portion of data to predict
                            columns = "sunspot_count",
                            target  = "sunspot_count" )

![](man/figures/optimal-E-1.png)

    E.opt
    #     E    rho
    # 1   1 0.7413
    # 2   2 0.8954
    # 3   3 0.9156
    # 4   4 0.9169
    # 5   5 0.9207
    # 6   6 0.9167
    # 7   7 0.9109
    # 8   8 0.9078
    # 9   9 0.8890
    # 10 10 0.8787

`E = 3` represents convergence of predictive skill as a function of `E`
and we use `E = 3` to forecast the last 1/3 of data based on training 
(attractor reconstruction) from the first 2/3.

    simplex = Simplex( dataFrame = df, 
                       lib     = "1   190", # portion of data to train
                       pred    = "191 289", # portion of data to predict
                       columns = "sunspot_count",
                       target  = "sunspot_count",
                       E       = 3 )

    plot( df$yr, df$sunspot_count, type = "l", lwd = 2,
          xlab = "year", ylab = "sunspots")
    lines( simplex$yr, simplex$Predictions, col = "red", lwd = 2)
    legend( 'topleft', legend = c( "Observed", "Predicted (year + 1)" ),
            fill = c( 'black', 'red' ), bty = 'n', cex = 1.3 )

![](man/figures/simplex-projection-1.png)

Further Examples
----------------

Please see the package vignettes for more details:

    browseVignettes("rEDM")

### References

Sugihara G. and May R. 1990. Nonlinear forecasting as a way of
distinguishing chaos from measurement error in time series. Nature,
344:734–741.

Sugihara G. 1994. Nonlinear forecasting for the classification of
natural time series. Philosophical Transactions: Physical Sciences and
Engineering, 348 (1688) : 477–495.

Deyle E. R. and Sugihara G. 2011. Generalized theorems for nonlinear
state space reconstruction, PLoS One, 6(3) : e18295. 

Sugihara G., May R., Ye H., Hsieh C., Deyle E., Fogarty M., Munch S.,
2012. Detecting Causality in Complex Ecosystems. Science 338:496-500.

Ye H., and G. Sugihara, 2016. Information leverage in interconnected 
ecosystems: Overcoming the curse of dimensionality. Science 353:922–925.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(echo = TRUE, warning = FALSE, tidy = TRUE,
                      fig.pos = 'h', fig.align = 'center',
                      fig.width = 4, fig.height = 3.3 )
knitr::opts_knit$set(global.par = TRUE, progress = FALSE)
options(digits = 2)
par.orig = par()
par(mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))

## ----CRAN installation instructions, eval = FALSE-----------------------------
#  install.packages("rEDM")

## ----GitHub installation instructions, eval = FALSE---------------------------
#  devtools::install_github("SugiharaLab/rEDM")

## ----LorenzProjection, echo = FALSE, fig.cap = "Time Series Projection from the Lorenz Attractor.", out.width = "50%", fig.align = 'center'----
knitr::include_graphics("Lorenz_Projection.png")

## ----fig_attractor_reconstruction, echo = FALSE, fig.cap = "Attractor Reconstruction from 3 Lagged Coordinates", out.width = "50%", fig.align = 'center'----
knitr::include_graphics("Lorenz_Reconstruct.png")

## ----load package-------------------------------------------------------------
library(rEDM)
str(TentMap)

## ----simplex on tentmap-------------------------------------------------------
simplex_out <- Simplex( dataFrame = TentMap, lib = "1 100", pred = "201 500",
columns = 'TentMap', target = 'TentMap', E = 3 )
simplex_out[ c(1:2, 300:301), ]

## ----tentmap simplex stats----------------------------------------------------
ComputeError( simplex_out $ Observations, simplex_out $ Predictions )

## ----Embed_Dim_TentMap, fig.cap="TentMap data prediction skill vs. embedding dimension."----
rho_E <- EmbedDimension( dataFrame = TentMap, lib = "1 100", pred = "201 500",
columns = 'TentMap', target = 'TentMap' )

## ----Prediction_interval_on_TentMap, fig.cap = "Tent map first differences simplex prediction skill as a function of forecast interval."----
rho_Tp <- PredictInterval( dataFrame = TentMap, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )

## ----Predict_nonlinear_on_TentMap, fig.cap = "Tent map first differences S-map prediction skill as a function of S-map localisation parameter."----
rho_theta <- PredictNonlinear( dataFrame = TentMapNoise, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )

## ----Simplex_TentMap----------------------------------------------------------
tentMapPredict <- Simplex( dataFrame = TentMap, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )

ComputeError( tentMapPredict $ Observations, tentMapPredict $ Predictions ) $ rho

## ----Tentmap_Noise_S-map------------------------------------------------------
smap = SMap( dataFrame = TentMapNoise, lib = "1 100", pred = "201 500",
target = 'TentMap', columns = 'TentMap', E = 2, theta = 3 )

## ----S-map_tentMap_results----------------------------------------------------
head( cbind( smap $ predictions, smap $ coefficients ), 2 )
tail( cbind( smap $ predictions, smap $ coefficients ), 2 )

## ----block_3sp_data-----------------------------------------------------------
head( block_3sp, 3 )

## ----3_Species_simplex, fig.cap="Simplex Tp=1 forecast of $x_t$."-------------
smplx_3species = Simplex( dataFrame = block_3sp, lib = "1 100",
pred = "101 190", E = 3, columns = "x_t x_t-1 z_t", target = "x_t",
embedded = TRUE )

## ----plot_3_species, fig.cap="Scatter plot of simplex forecast of $x_t$ vs. observations."----
err = ComputeError( smplx_3species $ Observations,
                    smplx_3species $ Predictions )
plot( smplx_3species $ Observations, smplx_3species $ Predictions,
pch = 19, cex = 0.5, xlab = 'Observations', ylab = 'Predictions',
main = '3 Species x_t' )
abline( a = 0, b = 1, lty = 2, col = 'blue' )
text( -1, 1, paste( capture.output( cbind( err ) ), collapse = '\n' ) )

## ----S-map_Lorenz96-----------------------------------------------------------
smap_Lorenz <- SMap( dataFrame = Lorenz5D, lib = "1 500", pred = "601 900",
E = 4, theta = 3, columns = "V1 V2 V3 V4", target = "V1", embedded = TRUE )

## ----set precision, echo = FALSE----------------------------------------------
options(digits=4)

## ----S-map_Lorenz_results, warning = FALSE------------------------------------
head( cbind( smap_Lorenz $ predictions, smap_Lorenz $ coefficients[,2:6] ), 3 )

## ----echo = FALSE-------------------------------------------------------------
par(mfrow = c(4, 1), mar = c(2, 4, 0.5, 1), oma = c(0, 0, 0, 0),
    mgp = c(2.1, 1, 0) )

## ----smap_Lorenz_plot, out.width = "70%", fig.cap = "S-map prediction and coefficients of Lorenz'96 5-D system."----
predictions = smap_Lorenz $ predictions
coefficients = smap_Lorenz $ coefficients
Time = predictions $ Time

plot( Time, predictions $ Observations, type = "l", col = "blue",
ylab = "V1", xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
lines( Time, predictions $ Predictions, lwd = 2, col = 'red' )
legend("topright", legend = c("observed", "predicted"),
fill = c( 'blue', 'red' ), bty = "n", cex = 1.3 )

plot( Time, coefficients[, 6], type = "l", col = "brown",
ylab = paste( "\U2202", "V4/", "\U2202", "V1" , sep = '' ), 
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
plot( Time, coefficients[, 5], type = "l", col = "darkgreen",
ylab = paste( "\U2202", "V3/", "\U2202", "V1" , sep = '' ),
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
plot( Time, coefficients[, 4], type = "l", col = "blue",
ylab = paste( "\U2202", "V2/", "\U2202", "V1" , sep = '' ),
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )

## ----echo = FALSE-------------------------------------------------------------
par(mfrow = c(1, 1), mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))

## ----multiview----------------------------------------------------------------
Mview = Multiview( dataFrame = block_3sp, lib = "1 100", pred = "101 190", E = 3, columns = "x_t y_t z_t", target = "x_t" )

## ----multiview results--------------------------------------------------------
Mview $ View[ which( Mview $ View $ rho > 0.91 ),  ]

## ----fig_cross_map, echo = FALSE, fig.cap = "Cross Mapping Between Reconstructions of the Lorenz Attractor"----
knitr::include_graphics("CrossMap.png")

## ----sardine_anchovy_ccm, fig.cap = "Convergent cross mapping of Newport sea surface temperature with anchovy landings."----
cmap <- CCM( dataFrame = sardine_anchovy_sst, E = 3, Tp = 0, columns = "anchovy", target = "np_sst", libSizes = "10 70 5", sample = 100, showPlot = TRUE )

## ----Thrips data--------------------------------------------------------------
head(Thrips, 2)

## ----thrips plot, echo = FALSE, fig.width = 3, fig.height = 3.5, fig.cap = "Thrips abundance and environmental variables."----
par(mfrow = c(4, 1), mar = c(2, 4, 0, 1), mgp = c(2., 0.5, 0))
time_dec <- Thrips $ Year + (Thrips $ Month)/12
plot(time_dec, Thrips $ Thrips_imaginis, type = "l", lwd = 2, col = "darkgreen", ylab = "Thrips", xlab = "")
plot(time_dec, Thrips $ maxT_degC, type = "l", lwd = 2, col = "red", ylab = "maxT (C)", xlab = "")
plot(time_dec, Thrips $ Rain_mm, type = "l", lwd = 2, col = "blue", ylab = "Rain (mm)", xlab = "")
plot(time_dec, Thrips $ Season, type = "l", lwd = 2, col = "magenta", ylab = "Season", xlab = "" )
mtext("Year", side = 1, outer = TRUE, line = 1)

## ----echo = FALSE-------------------------------------------------------------
par(mfrow = c(1, 1), mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))

## ----univariate thrips, fig.cap = "Simplex embedding dimension for Thrips abundance.", fig.width = 4, fig.height = 3.----
rho_E <- EmbedDimension( dataFrame = Thrips, columns = "Thrips_imaginis", target = "Thrips_imaginis", lib = "1 72", pred = "1 72", showPlot = TRUE )

## ----smap for thrips, fig.cap = "SMap localisation parameter for Thrips abundance.", fig.width = 4, fig.height = 3.----
E = 8
rho_theta_e3 = PredictNonlinear(dataFrame = Thrips, columns = "Thrips_imaginis", target = "Thrips_imaginis", lib = "1 73", pred = "1 73", E = E )

## ----compute ccm matrix for thrips, results='hold', cache = TRUE--------------
vars       = colnames(Thrips[3:6])
var_pairs  = combn( vars, 2 )  # Combinations of vars, 2 at a time
libSize    = paste( NROW(Thrips) - E, NROW(Thrips) - E, 10, collapse = " " )
ccm_matrix = array(NA, dim = c(length(vars), length(vars)), 
                   dimnames = list(vars, vars))

for( i in 1:ncol( var_pairs ) ) {
   ccm_out = CCM(dataFrame = Thrips,
                 columns = var_pairs[1,i], target = var_pairs[2,i],
                 libSizes = libSize, Tp = 0, E = E, sample = 100)
                 
   outVars = names( ccm_out )

   var_out = unlist( strsplit( outVars[2], ':' ) )
   ccm_matrix[ var_out[2], var_out[1] ] = ccm_out[1,2]
                 
   var_out = unlist( strsplit( outVars[3], ':' ) )
   ccm_matrix[ var_out[2], var_out[1] ] = ccm_out[1,3]
}

## ----compute corr matrix for thrips, cache = TRUE-----------------------------
corr_matrix <- array(NA, dim = c(length(vars), length(vars)), 
                     dimnames = list(vars, vars))

for(ccm_from in vars) {
    for(ccm_to in vars[vars != ccm_from]) {
        ccf_out <- ccf(Thrips[,ccm_from], Thrips[,ccm_to], 
                       type = "correlation", lag.max = 6, plot = FALSE)$acf
        corr_matrix[ccm_from, ccm_to] <- max(abs(ccf_out))
    }
}

## ----xmap vs. corr matrix for thrips------------------------------------------
ccm_matrix
corr_matrix

## ----thrips setup, echo = FALSE-----------------------------------------------
par( mfrow = c( 1, 3 ), mar = c( 3.5, 1.3, 2.3, 0.2 ), mgp = c(1.5, 0.5, 0) )

CCMPlot = function( ccm.df, E, ylim = c( 0.2, 1 ) ) {
    libSize = ccm.df $ LibSize
    V1      = names( ccm.df )[2]
    V2      = names( ccm.df )[3]

    title = paste( V1, " : ", V2, "\nE=" , E )
    
    print( title )
    print( libSize )

    plot( libSize, ccm.df[ , V1 ],
          ylim = ylim, main = title, col = "blue", type = "l", lwd = 3,
          xlab = 'Library Size', ylab = 'Prediction Skill (\U03C1)' )
    lines( libSize, ccm.df[ , V2 ], col = "red", lwd = 3 )
    abline( h = 0 )
    legend( 'topright', c( V1, V2 ), 
            fill = c( 'blue', 'red' ), bty = 'n', cex = 1.2 )
}

## ----ccm thrips, fig.cap = "Thrips cross mapped to climatic variables. Vertical axis is cross map prediction skill (rho).", fig.width = 7.5----
thrips_xmap_maxT <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                        columns = "Thrips_imaginis", target = "maxT_degC",
                        libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_maxT, E )
abline(h = corr_matrix['Thrips_imaginis', 'maxT_degC'], col = "black", lty = 2)

thrips_xmap_Rain <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                        columns = "Thrips_imaginis", target = "Rain_mm",
                        libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_Rain, E )
abline(h = corr_matrix['Thrips_imaginis', 'Rain_mm'], col = "black", lty = 2)

thrips_xmap_Season <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                          columns = "Thrips_imaginis", target = "Season",
                          libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_Season, E )
abline(h = corr_matrix['Thrips_imaginis', 'Season'], col = "black", lty = 2)

## ----thrips setdown, echo = FALSE---------------------------------------------
par(mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))

## ----seasonal surrogates thrips, cache = TRUE---------------------------------
# Create matrix with temperature and rain surrogates (1000 time series vectors)
surr_maxT = SurrogateData( Thrips $ maxT_degC, method = "seasonal", T_period = 12, num_surr = 1000, alpha = 3)
surr_rain = SurrogateData( Thrips $ Rain_mm, method = "seasonal", T_period = 12, num_surr = 1000, alpha = 3)

# Rain cannot be negative
surr_rain = apply( surr_rain, 2, function(x){ i = which(x<0); x[i] = 0; x } )

# data.frame to hold CCM rho values between Thrips abundance and variable
rho_surr <- data.frame( maxT = numeric(1000), Rain = numeric(1000) )

# data.frames with time, Thrips, and 1000 surrogate climate variables for CCM()
maxT_data = as.data.frame( cbind( seq(1:nrow(Thrips)), Thrips $ Thrips_imaginis, surr_maxT ) )
names( maxT_data ) = c( 'time', 'Thrips_imaginis', paste( 'T', as.character( seq( 1,1000 ) ), sep = '' ) )

rain_data = as.data.frame( cbind( seq(1:nrow(Thrips)), Thrips $ Thrips_imaginis, surr_rain ) )
names( rain_data ) = c( 'time', 'Thrips_imaginis', paste( 'R', as.character( seq( 1,1000 ) ), sep = '' ) )

# Cross mapping
for (i in 1:1000) {
  targetCol = paste( 'T', i, sep = '' ) # as in maxT_data
  
  ccm_out = CCM( dataFrame = maxT_data, E = E, Tp = 0,
                 columns = 'Thrips_imaginis', target = targetCol,
                 libSizes = "73 73 5", sample = 1 )
                 
  col = paste( 'Thrips_imaginis', ':', targetCol, sep = '' )
   
  rho_surr $ maxT[i] = ccm_out[1,col]
}

for (i in 1:1000) {
  targetCol = paste( 'R', i, sep = '' ) # as in rain_data
  
  ccm_out = CCM( dataFrame = rain_data, E = E, Tp = 0,
                 columns = 'Thrips_imaginis', target = targetCol,
                 libSizes = "73 73 5", sample = 1 )
                 
  col = paste( 'Thrips_imaginis', ':', targetCol, sep = '' )
   
  rho_surr $ Rain[i] = ccm_out[1,col]
}

## ----significance of randomization test---------------------------------------
1 - ecdf( rho_surr $ maxT )( ccm_matrix["maxT_degC", "Thrips_imaginis" ] )
1 - ecdf( rho_surr $ Rain )( ccm_matrix["Rain_mm",   "Thrips_imaginis" ] )

## ----reset par, include = FALSE-----------------------------------------------
par( par.orig )

## ----parameter-table, echo = FALSE--------------------------------------------
paramTable = read.csv( "ParameterTable.csv", as.is = TRUE )
knitr::kable( paramTable, caption = '' )

#------------------------------------------------------------------------
# Shared marshalling utilities for the pure-R EDM kernels.
# Ported from rEDM/R/EDM_AuxFuncs.R; behaviour preserved.
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#' Test for a length-one atomic value.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
is.scalar <- function(x) is.atomic(x) && length(x) == 1L

#------------------------------------------------------------------------
#' Assemble the Simplex/SMap prediction data.frame.
#'
#' Port of Formatting.py::FormatProjection for the projection/variance path
#' (Time, Observations, Predictions, Pred_Variance). Internal position
#' arithmetic mirrors pyEDM (0-based), converted to 1-based at R indexing.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
ProjectionLayout <- function(predList, pred_i, pred_i_all, timeVec,
                             Tp, nRows, tau) {
  TpMag <- abs(Tp)

  # 0-based locals to mirror pyEDM arithmetic
  predList0   <- lapply(predList, function(s) s - 1L)
  pred_i0     <- pred_i - 1L
  pred_i_all0 <- pred_i_all - 1L

  # --- Observations index vector obs_i (0-based) and total outSize ---
  obs_i   <- integer(0)
  outSize <- 0L
  for (seg in predList0) {
    nPred <- length(seg)
    outSize <- outSize + nPred + TpMag
    if (nPred == 0L) { next }
    if (Tp == 0) {
      appendI <- seg
    }
    else if (Tp > 0) {
      if (seg[nPred] + Tp < nRows) {
        appendI <- c(seg, (seg[nPred] + 1):(seg[nPred] + Tp))
      } else appendI <- seg
    }
    else {
      if (seg[1] + Tp > -1) {
        appendI <- c((seg[1] + Tp):(seg[1] - 1), seg)
      } else appendI <- seg
    }
    obs_i <- c(obs_i, appendI)
  }

  # --- Projection output positions predOut_i (0-based) ---
  predOut_i <- integer(0)
  predOut0  <- 0L
  for (seg in predList0) {
    nPred <- length(seg)
    if (nPred == 0L) { next }
    if (Tp == 0) {
      idx <- predOut0:(predOut0 + nPred - 1L)
      predOut_i <- c(predOut_i, idx); predOut0 <- idx[length(idx)] + 1L
    }
    else if (Tp > 0) {
      idx <- (predOut0 + Tp):(predOut0 + Tp + nPred - 1L)
      predOut_i <- c(predOut_i, idx); predOut0 <- idx[length(idx)] + 1L
    }
    else {
      idx <- predOut0:(predOut0 + nPred - 1L)
      predOut_i <- c(predOut_i, idx); predOut0 <- idx[length(idx)] + TpMag + 1L
    }
  }

  # NaN removal may have shortened pred_i vs predOut_i: remap
  if (length(pred_i0) != length(predOut_i)) {
    nMap <- length(predOut_i)
    if (tau < 0) {
      keys <- pred_i_all0[(length(pred_i_all0) - nMap + 1L):length(pred_i_all0)]
    }
    else {
      keys <- pred_i_all0[seq_len(nMap)]
    }
    mp <- predOut_i
    names(mp) <- as.character(keys)
    predOut_i <- unname(mp[as.character(pred_i0)])
  }

  # --- Observation output positions obsOut_i (0-based) ---
  if (Tp > 0) {
    if (obs_i[length(obs_i)] + Tp > nRows - 1L) {
      obsOut_i <- seq_len(length(obs_i)) - 1L
    }
    else {
      obsOut_i <- seq_len(outSize) - 1L
    }
  }
  else {
    obsOut_i <- seq_len(length(obs_i)) - 1L
      
    if (pred_i0[1] + Tp < 0) {
      obsOut_i <- obsOut_i + (TpMag - pred_i0[1])
    }
  }

  # --- Time ---
  timeOut <- rep(NA, outSize)
  if (Tp == 0 ||
      (Tp > 0 && (pred_i_all[length(pred_i_all)] + Tp) <= length(timeVec)) ||
      (Tp < 0 && (pred_i0[1] + Tp >= 0))) {
    timeOut[obsOut_i + 1L] <- timeVec[obs_i + 1L]
  }
  else {
    timeOut <- AddTime(timeVec, TpMag, outSize, obs_i, obsOut_i, Tp)
  }

  list(obs_i = obs_i, outSize = outSize, predOut_i = predOut_i,
       obsOut_i = obsOut_i, timeOut = timeOut)
}

#------------------------------------------------------------------------
#' Assemble the Simplex/SMap prediction data.frame.
#'
#' Port of Formatting.py::FormatProjection for the projection/variance path
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
FormatProjection <- function(predList, pred_i, pred_i_all, targetVec, timeVec,
                             projection, variance, Tp, nRows, tau, timeName) {

  L <- ProjectionLayout(predList, pred_i, pred_i_all, timeVec, Tp, nRows, tau)

  observationOut <- rep(NA_real_, L$outSize)
  projectionOut  <- rep(NA_real_, L$outSize)
  varianceOut    <- rep(NA_real_, L$outSize)
  observationOut[L$obsOut_i  + 1L] <- targetVec[L$obs_i + 1L]
  projectionOut [L$predOut_i + 1L] <- projection
  varianceOut   [L$predOut_i + 1L] <- variance

  dfNames = c( timeName, "Observations", "Predictions", "Pred_Variance" )
  df = data.frame(L$timeOut, observationOut, projectionOut, varianceOut,
                  stringsAsFactors = FALSE, check.names = FALSE)
  names(df) = dfNames
  df
}

#------------------------------------------------------------------------
#' Extend the time vector for forecast rows beyond the data (Tp shift).
#' Numeric time only; assumes a constant step from the last two values.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
AddTime <- function(timeVec, TpMag, outSize, obs_i, obsOut_i, Tp) {
  timeOut <- rep(NA, outSize)
  timeOut[obsOut_i + 1L] <- timeVec[obs_i + 1L]
  # Fill any trailing/leading NA by extrapolating a constant step
  known <- which(!is.na(timeOut))
  if (length(known) >= 2 && is.numeric(timeVec)) {
    step <- timeVec[2] - timeVec[1]
    for (i in seq_len(outSize)) {
      if (is.na(timeOut[i])) {
        earlier <- known[known < i]
        if (length(earlier)) {
          ref <- max(earlier)
          timeOut[i] <- timeOut[ref] + step * (i - ref)
        }
      }
    }
    # leading NA
    for (i in rev(seq_len(outSize))) {
      if (is.na(timeOut[i])) {
        later <- known[known > i]
        if (length(later)) {
          ref <- min(later)
          timeOut[i] <- timeOut[ref] - step * (ref - i)
        }
      }
    }
  }
  timeOut
}

#------------------------------------------------------------------------
#' Split a columns/target argument (space string or vector) into names.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
SplitColumns <- function(x) {
  if (length(x) == 1L && is.character(x) && grepl("\\s", x)) {
    strsplit(trimws(x), "\\s+")[[1]]
  }
  else {
    x
  }
}

#------------------------------------------------------------------------
#' Build 1-based library and prediction indices from (start end) pairs.
#'
#' Port of EDM.py::CreateIndices with the 0-offset removed.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
CreateIndices <- function(lib, pred, E, tau, Tp, embedded, nRows,
                          name = "Simplex") {
  embedShift <- abs(tau) * (E - 1)

  libPairs <- matrix(ParsePairs(lib), ncol = 2, byrow = TRUE)
  libList  <- vector("list", nrow(libPairs))

  for (r in seq_len(nrow(libPairs))) {
    start <- libPairs[r, 1]; stop <- libPairs[r, 2]

    if (name %in% c("Simplex", "SMap", "Multiview") && start >= stop) {
      stop(sprintf("CreateIndices: lib start %d exceeds lib end %d.",
                   start, stop))
    }
    if (start < 1 || stop < 1) {
      stop("CreateIndices: lib indices < 1 not allowed.")
    }
    if (!embedded) {
      if (tau < 0) start <- start + embedShift else stop <- stop - embedShift
    }
    if (Tp < 0) {
      if (!embedded) start <- max(start, start + abs(Tp) - 1)
    }
    else if (r == nrow(libPairs)) {
      stop <- stop - Tp
    }
    libList[[r]] <- start:stop
  }
  lib_i       <- unlist(libList)
  disjointLib <- length(libList) > 1

  predPairs <- matrix(ParsePairs(pred), ncol = 2, byrow = TRUE)
  predList  <- vector("list", nrow(predPairs))

  for (r in seq_len(nrow(predPairs))) {
    pStart <- predPairs[r, 1]; pStop <- predPairs[r, 2]

    if (name %in% c("Simplex", "SMap", "Multiview") && pStart >= pStop) {
      stop(sprintf("CreateIndices: pred start %d exceeds pred end %d.",
                   pStart, pStop))
    }
    if (pStart < 1 || pStop < 1) {
      stop("CreateIndices: pred indices < 1 not allowed.")
    }

    predList[[r]] <- pStart:pStop
  }
  pred_i_all <- unlist(predList)

  # Remove leading/trailing embedding-NaN rows from predList (not embedded)
  if (!embedded && embedShift > 0) {
    nanStart <- seq_len(embedShift)                       # rows 1..embedShift
    nanEnd   <- (nRows - embedShift + 1):nRows            # last embedShift rows
    for (i in seq_along(predList)) {
      seg <- predList[[i]]
      if (tau > 0) { predList[[i]] <- seg[!(seg %in% nanEnd)]   }
      else         { predList[[i]] <- seg[!(seg %in% nanStart)] }
    }
  }
  pred_i <- unlist(predList)

  if (lib_i[length(lib_i)] > nRows) {
    stop("CreateIndices: library index exceeds number of data rows.")
  }
  if (pred_i[length(pred_i)] > nRows) {
    stop("CreateIndices: prediction index exceeds number of data rows.")
  }

  libOverlap <- length(intersect(lib_i, pred_i)) > 0

  list(lib_i = lib_i, pred_i = pred_i, pred_i_all = pred_i_all,
       predList = predList, disjointLib = disjointLib,
       libOverlap = libOverlap, embedShift = embedShift)
}

#------------------------------------------------------------------------
#' Parse (start end) pairs from a space string or numeric vector.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
ParsePairs <- function(spec) {
  if (is.character(spec)) {
    v <- suppressWarnings(as.integer(strsplit(trimws(spec), "\\s+")[[1]]))
  }
  else { v <- as.integer(spec) }

  if (length(v) == 0L || anyNA(v) || length(v) %% 2L != 0L) {
    stop("CreateIndices: lib/pred must be an even set of (start end) pairs.")
  }

  v
}

#------------------------------------------------------------------------
#' Remove embedding-NaN rows from lib_i / pred_i (ignoreNan path).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
RemoveNan <- function(embedding, lib_i, pred_i, ignoreNan) {
  if (!ignoreNan) {
    return(list(lib_i = lib_i, pred_i = pred_i))
  }

  naRow <- function(idx) apply(embedding[idx, , drop = FALSE], 1,
                               function(r) any(is.na(r)))
  naLib  <- naRow(lib_i)
  naPred <- naRow(pred_i)
  list(lib_i  = lib_i[!naLib], pred_i = pred_i[!naPred])
}

#------------------------------------------------------------------------
#' Flatten a vector / list / data.frame / matrix to a delimited string.
#'
#' Preserves the rEDM convention used to pass \code{columns}, \code{target},
#' \code{lib} and \code{pred} to the kernels as space-delimited strings.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
FlattenToString <- function(x, delimiter = " ") {
  if (is.data.frame(x) || is.matrix(x)) {
    s <- ""
    for (row in seq_len(nrow(x))) {
      s <- paste(s, paste(x[row, ], collapse = delimiter),
                 collapse = delimiter)
    }
  }
  else if (is.list(x)) {
    s <- paste(unlist(x), collapse = delimiter)
  }
  else if (is.vector(x)) {
    s <- paste(x, collapse = delimiter)
  }
  else {
    s <- x
  }
  s
}

#------------------------------------------------------------------------
#' Parse a lib/pred specification into 1-based row indices.
#'
#' Accepts either a whitespace string ("1 100" or "1 100 201 300") or a
#' numeric vector (c(1, 100) or c(1, 100, 201, 300)). Values are interpreted
#' as inclusive (start, end) pairs and expanded to a 1-based integer index
#' vector. This is the R-native form: no 0-offset conversion is applied
#' (contrast pyEDM, which subtracts 1 for NumPy).
#'
#' @param spec character scalar or numeric vector of (start end) pairs.
#' @return integer vector of 1-based indices (sorted, de-duplicated within
#'   overlapping pairs preserved in order of pairs).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
ParseIndexPairs <- function(spec) {
  if (is.character(spec)) {
    toks <- strsplit(trimws(spec), "\\s+")[[1]]
    toks <- toks[nzchar(toks)]
    vals <- suppressWarnings(as.integer(toks))
  }
  else {
    vals <- as.integer(spec)
  }

  if (length(vals) == 0L || anyNA(vals)) {
    stop("ParseIndexPairs: spec must be integers as (start end) pairs, got: ",
         FlattenToString(spec))
  }
  if (length(vals) %% 2L != 0L) {
      stop( paste("ParseIndexPairs: expected an even number of values",
                  "(start end pairs), got "),
         length(vals), ": ", FlattenToString(vals) )
  }

  idx <- integer(0)
  for (p in seq(1L, length(vals), by = 2L)) {
    start <- vals[p]
    end   <- vals[p + 1L]
    if (start < 1L) {
      stop("ParseIndexPairs: 1-based indices must be >= 1; got start = ", start)
    }
    if (end < start) {
      stop("ParseIndexPairs: pair end (", end, ") < start (", start, ")")
    }
    idx <- c(idx, start:end)
  }
  idx
}

#------------------------------------------------------------------------
#' Is the object a non-empty data.frame with numeric value columns?
#'
#' Port of rEDM isValidDataFrame: columns 2..ncol must be numeric (the first
#' column may be time / date / character).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
isValidDataFrame <- function(dataFrame) {
  if (!inherits(dataFrame, "data.frame")) {
    message("isValidDataFrame(): object is not a data.frame."); return(FALSE)
  }
  if (nrow(dataFrame) == 0 || ncol(dataFrame) == 0) {
    message("isValidDataFrame(): dataFrame is empty."); return(FALSE)
  }
  if (ncol(dataFrame) >= 2) {
    cls <- vapply(dataFrame[, 2:ncol(dataFrame), drop = FALSE],
                  function(col) class(col)[1], character(1))
    if (any(cls %in% c("character", "factor"))) {
      message("isValidDataFrame(): non-numeric value column detected.")
      return(FALSE)
    }
  }
  TRUE
}

#------------------------------------------------------------------------
#' Validate a data.frame and that columns/target exist in it.
#'
#' Port of rEDM ValidateDataFrame (dataFrame path). Returns TRUE/FALSE.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
ValidateDataFrame <- function(dataFrame, columns, target, verbose = FALSE) {
  if (!isValidDataFrame(dataFrame)) return(FALSE)
  nm <- names(dataFrame)
  for (c in SplitColumns(columns)) {
    if (!c %in% nm) {
      message("ValidateDataFrame(): column '", c, "' not found.")
      return(FALSE)
    }
  }
  
  for (t in SplitColumns(target)) {
    if (!t %in% nm) {
      message("ValidateDataFrame(): target '", t, "' not found.")
      return(FALSE)
    }
  }
  if (verbose) { message("ValidateDataFrame(): dataFrame validated.") }
  TRUE
}

#------------------------------------------------------------------------
#' Stop with a clear message when input is not usable (API entry guard).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
RequireValidData <- function(name, dataFrame, columns, target) {
  if (!ValidateDataFrame(dataFrame, columns, target)) {
    stop(sprintf("%s(): invalid dataFrame, or columns/target not found.", name),
         call. = FALSE)
  }
  invisible(TRUE)
}
#------------------------------------------------------------------------
#' Generate surrogate time series for significance testing.
#'
#' Three methods:
#' \code{random_shuffle} (permutation), \code{ebisuzaki} (Fourier
#' phase-randomisation preserving the power spectrum / autocorrelation), and
#' \code{seasonal} (period-\code{T_period} spline cycle plus shuffled
#' residuals plus Gaussian noise).
#'
#' @param ts numeric time-series vector.
#' @param method one of "random_shuffle", "ebisuzaki", "seasonal".
#' @param num_surr number of surrogate series to generate.
#' @param T_period seasonal period (seasonal method).
#' @param alpha Gaussian-noise standard deviation (seasonal method).
#' @return numeric matrix with \code{length(ts)} rows and \code{num_surr} columns
#' @keywords internal
#------------------------------------------------------------------------
SurrogateData <- function(ts,
                          method = c("random_shuffle", "ebisuzaki", "seasonal"),
                          num_surr = 100, T_period = 1, alpha = 0) {
  method <- match.arg(method)

  if (method == "random_shuffle") {
    return(sapply(seq_len(num_surr), function(i) sample(ts, size = length(ts))))
  }

  if (method == "ebisuzaki") {
    if (any(!is.finite(ts))) {
      stop("SurrogateData(): input time series contained invalid values")
    }
    n  <- length(ts)
    n2 <- floor(n / 2)
    sigma <- stats::sd(ts)
    amplitudes    <- abs(stats::fft(ts))
    amplitudes[1] <- 0

    return(sapply(seq_len(num_surr), function(i) {
      if (n %% 2 == 0) {                           # even length
        thetas    <- 2 * pi * stats::runif(n2 - 1)
        angles    <- c(0, thetas, 0, -rev(thetas))
        recf      <- amplitudes * exp(complex(imaginary = angles))
        recf[n2]  <- complex(real = sqrt(2) * amplitudes[n2] *
                             cos(stats::runif(1) * 2 * pi))
      }
      else {                                       # odd length
        thetas <- 2 * pi * stats::runif(n2)
        angles <- c(0, thetas, -rev(thetas))
        recf   <- amplitudes * exp(complex(imaginary = angles))
      }
      temp <- Re(stats::fft(recf, inverse = TRUE) / n)
      temp / stats::sd(temp) * sigma                # match original variance
    }))
  }

  # seasonal
  if (any(!is.finite(ts))) {
    stop("SurrogateData(): input time series contained invalid values")
  }

  n        <- length(ts)
  I_season <- suppressWarnings(matrix(1:T_period, nrow = n, ncol = 1))
  seasonalF <- stats::smooth.spline(
           c(I_season - T_period, I_season, I_season + T_period), c(ts, ts, ts))
  
  seasonalCyc   <- stats::predict(seasonalF, I_season)$y
  seasonalResid <- ts - seasonalCyc

  sapply(seq_len(num_surr), function(i)
    seasonalCyc + sample(seasonalResid, n) + stats::rnorm(n, 0, alpha))
}

#------------------------------------------------------------------------
#' Takens time-delay embedding block (numeric matrix).
#'
#' Internal engine builder: returns the unnamed numeric embedding matrix
#' consumed by the neighbour search and projection kernels. All rows are
#' retained; the first |tau|*(E-1) partial rows contain NA.
#'
#' @param dataFrame data.frame of source data.
#' @param E embedding dimension.
#' @param tau delay (negative = past lags, pyEDM default -1).
#' @param columns character vector (or space string) of column names.
#' @return numeric matrix (N x (E * length(columns))).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
MakeBlock <- function(dataFrame, E, tau, columns) {
  if (E < 1)    stop("MakeBlock(): E must be positive.")
  if (tau == 0) stop("MakeBlock(): tau must be non-zero.")
  columns <- SplitColumns(columns)
  if (length(columns) == 0L) stop("MakeBlock(): columns required.")
  absent <- setdiff(columns, names(dataFrame))
  if (length(absent))
    stop(sprintf("MakeBlock(): column(s) not found: %s",
                 paste(absent, collapse = ", ")))

  shiftVec <- seq(0, by = -tau, length.out = E)   # tau<0 -> 0,|tau|,2|tau|,...
  blocks <- vector("list", length(columns))
  for (ci in seq_along(columns)) {
    col <- dataFrame[[columns[ci]]]
    m <- matrix(NA_real_, length(col), E)
    for (j in seq_len(E)) m[, j] <- ShiftVec(col, shiftVec[j])
    blocks[[ci]] <- m
  }
  do.call(cbind, blocks)
}

#------------------------------------------------------------------------
#' Takens time-delay embedding of one or more columns.
#'
#' Wrapper over \code{MakeBlock} that returns a named data.frame matching
#' pyEDM's \code{Embed()}: each embedding column is named
#' \code{column(t-lag)} (or \code{(t+lag)} for positive \code{tau}), where
#' \code{lag = i * |tau|} for \code{i = 0..E-1}. All rows are kept, with NA
#' in the first \code{|tau|*(E-1)} partial rows.
#'
#' @param dataFrame data.frame of source data; column 1 is time.
#' @param E embedding dimension.
#' @param tau delay (negative = past lags, pyEDM default -1).
#' @param columns column name(s): a character vector or a space-separated string.
#' @param includeTime if TRUE, prepend column 1 of \code{dataFrame} (time),
#'   keeping its original name.
#' @param pathIn directory for \code{dataFile} (used when reading from disk).
#' @param dataFile CSV file name to read as input; when set it takes
#'   precedence over \code{dataFrame}.
#' @return a data.frame of the embedding (with the time column first when
#'   \code{includeTime = TRUE}).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
Embed <- function(dataFrame = NULL, E = 0, tau = -1, columns = "",
                  includeTime = FALSE, pathIn = "./", dataFile = "") {
  dataFrame <- ResolveInput(pathIn, dataFile, dataFrame)
  columns <- SplitColumns(columns)
  block   <- MakeBlock(dataFrame, E, tau, columns)   # validates E / tau / columns
  df      <- as.data.frame(block, stringsAsFactors = FALSE)
  names(df) <- EmbedColumnNames(columns, E, tau, embedded = FALSE)
  if (isTRUE(includeTime))
    df <- cbind(dataFrame[, 1, drop = FALSE], df)
  df
}

#------------------------------------------------------------------------
#' Embedding column names, pyEDM convention.
#'
#' For column \code{c} and step \code{i = 0..E-1} the lag is \code{i*|tau|};
#' the name is \code{c(t-lag)} when \code{tau < 0} and \code{c(t+lag)} when
#' \code{tau > 0}. When \code{embedded}, \code{columns} already name the
#' embedding and are returned unchanged. Shared by \code{Embed} and the
#' S-map coefficient labelling so the two never disagree.
#'
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
EmbedColumnNames <- function(columns, E, tau, embedded = FALSE) {
  if (embedded) return(columns)
  sgn  <- if (tau < 0) "t-" else "t+"
  lags <- (0:(E - 1)) * abs(tau)
  unlist(lapply(columns, function(c) sprintf("%s(%s%d)", c, sgn, lags)))
}

#------------------------------------------------------------------------
#' Shift a vector by p (pandas .shift convention): p>0 moves values down
#' (NA at head), p<0 moves up (NA at tail).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
ShiftVec <- function(x, p) {
  n <- length(x)
  if (p == 0) return(x)
  if (p > 0)  return(c(rep(NA_real_, p), x[seq_len(n - p)]))
  p <- -p
  c(x[(p + 1):n], rep(NA_real_, p))
}

#------------------------------------------------------------------------
# Convergent Cross Mapping (CCM), pyEDM 2.5.0 model.
#
# Port of pyEDM CCM.py. Strategy: precompute delay embeddings and validity
# filtering once; for each (direction x libSize) task, run `sample` random
# subsample trials. Each trial builds a KDTree from L library vectors,
# queries all M valid points, computes CCM Simplex-weighted cross-map
# predictions, and takes the Pearson correlation.
#
# CCM has its OWN Simplex weighting (NOT the Simplex kernel): no 1e-6 floor,
# d_min_nz = where(d>0, d, 1), explicit zero-distance branch, explicit
# normalization.
#
# RNG: NumPy's PCG64 / SeedSequence subsampling cannot be bit-reproduced in
# R, so results cannot match pyEDM's fixtures exactly (see the tiered
# validation in the tests). For reproducibility, each (direction x libSize)
# task draws from its own L'Ecuyer-CMRG stream, making results independent of
# execution order and of sequential-vs-parallel dispatch.
#
#------------------------------------------------------------------------


#------------------------------------------------------------------------
#' Convergent Cross Mapping (top-level API).
#'
#' @param dataFrame data.frame of source data.
#' @param columns,target variable name(s); univariate targets use the first.
#' @param E,Tp,knn,tau,exclusionRadius standard EDM parameters (CCM Tp default 0).
#' @param libSizes explicit vector/string or (start stop increment).
#' @param sample subsamples per library size.
#' @param seed RNG seed (reproducible; not bit-comparable to pyEDM/NumPy).
#' @param embedded TRUE if columns/target already form embeddings.
#' @param validLib logical mask of admissible rows (or length 0).
#' @param includeData add per-libSize sample variance columns.
#' @param numProcess worker processes for the (direction x libSize) grid.
#' @param backend "RANN" or "brute".
#' @param pathIn file path for input dataFile.
#' @param dataFile input dataFile, .csv format.
#' @param pathOut file path for predictFile.
#' @param predictFile output file name.
#' @return data.frame(LibSize, <col:tgt>, <tgt:col> [, _var columns]).
#' @keywords internal
#------------------------------------------------------------------------
CCM <- function(dataFrame = NULL, columns, target, E, Tp = 0, knn = 0, tau = -1,
                exclusionRadius = 0, libSizes, sample = 30, seed = NULL,
                embedded = FALSE, validLib = logical(0), includeData = FALSE,
                numProcess = 4, backend = "RANN",
                pathIn = "./", dataFile = "", pathOut = "./", predictFile = "",
                parameterList = FALSE, showPlot = FALSE) {

  parameters <- as.list(environment())
  parameters[c("dataFrame")] <- NULL

  dataFrame <- ResolveInput(pathIn, dataFile, dataFrame)
  columns   <- SplitColumns(columns)
  target    <- SplitColumns(target)
  RequireValidData("CCM", dataFrame, columns, target)
  nRows     <- nrow(dataFrame)

  if (embedded) { E   <- length(columns) }
  if (knn < 1)  { knn <- E + 1L }

  libSizes <- CcmLibSizes(libSizes, E, nRows)
  shifts   <- (0:(E - 1)) * tau

  tgtVec     <- as.numeric(dataFrame[[target[1]]])
  colPredVec <- as.numeric(dataFrame[[columns[1]]])
  N          <- nRows

  if (embedded) {
    embCol   <- as.matrix(dataFrame[, columns, drop = FALSE])
    embTgt   <- as.matrix(dataFrame[, target,  drop = FALSE])
    validCol <- !apply(embCol, 1, function(r) any(is.na(r)))
    validTgt <- !apply(embTgt, 1, function(r) any(is.na(r)))
  }
  else {
    colVecs <- lapply(columns, function(c) as.numeric(dataFrame[[c]]))
    tgtVecs <- lapply(target,  function(c) as.numeric(dataFrame[[c]]))
    bc      <- CcmBuildEmbedding(colVecs, shifts)
    embCol  <- bc$emb; validCol <- bc$valid
    bt      <- CcmBuildEmbedding(tgtVecs, shifts)
    embTgt  <- bt$emb; validTgt <- bt$valid
  }

  if (length(validLib)) {
    validCol <- validCol & validLib
    validTgt <- validTgt & validLib
  }
  validCol <- validCol & CcmTpValidMask(N, tgtVec, Tp)     & !is.na(colPredVec)
  validTgt <- validTgt & CcmTpValidMask(N, colPredVec, Tp) & !is.na(tgtVec)

  idxCol <- which(validCol)
  idxTgt <- which(validTgt)

  predValsCol   <- tgtVec[idxCol + Tp]
  predValsTgt   <- colPredVec[idxTgt + Tp]
  embedValidCol <- embCol[idxCol, , drop = FALSE]
  embedValidTgt <- embTgt[idxTgt, , drop = FALSE]

  # ---- Task grid: 2 x nLib, each with its own L'Ecuyer stream ----
  nLib    <- length(libSizes)
  nTasks  <- 2L * nLib
  oldKind <- RNGkind("L'Ecuyer-CMRG")
  on.exit(RNGkind(oldKind[1]), add = TRUE)
  if (is.null(seed)) seed <- as.integer(stats::runif(1, 1, .Machine$integer.max))
  set.seed(seed)
  streams <- vector("list", nTasks)
  cur <- get(".Random.seed", envir = .GlobalEnv, inherits = FALSE)
  for (i in seq_len(nTasks)) { streams[[i]] <- cur;
                               cur <- parallel::nextRNGStream(cur) }

  tasks <- vector("list", nTasks)
  for (i in seq_len(nLib)) {
      tasks[[2L * i - 1L]] <- list(dir = "fwd", L = libSizes[i],
                                   stream = streams[[2L * i - 1L]])
      tasks[[2L * i]]      <- list(dir = "rev",
                                   L = libSizes[i], stream = streams[[2L * i]])
  }

  worker <- function(task) {
    assign(".Random.seed", task$stream, envir = .GlobalEnv)
    if (task$dir == "fwd") {
      rhos <- CcmForLibSize(embedValidCol, predValsCol, idxCol, task$L,
                            sample, knn, exclusionRadius, backend)
    }
    else {
      rhos <- CcmForLibSize(embedValidTgt, predValsTgt, idxTgt, task$L,
                            sample, knn, exclusionRadius, backend)
    }
    
    c(mean = mean(rhos, na.rm = TRUE),
      var  = if (includeData) {
                stats::var(rhos[!is.na(rhos)]) * (sum(!is.na(rhos))-1) /
                sum(!is.na(rhos)) # population var
             }
             else {
               NA_real_
             }
      ) 
  }

  results <- SweepApply(tasks, worker, numProcess)

  fwdMean <- vapply(seq_len(nLib),
                    function(i) results[[2L * i - 1L]]["mean"], numeric(1))
  revMean <- vapply(seq_len(nLib),
                    function(i) results[[2L * i]]["mean"],      numeric(1))

  cn <- paste0(columns[1], ":", target[1])
  tn <- paste0(target[1],  ":", columns[1])
  out <- data.frame(LibSize = libSizes, a = fwdMean, b = revMean,
                    check.names = FALSE, stringsAsFactors = FALSE)
  names(out) <- c("LibSize", cn, tn)

  if (includeData) {
    out[[paste0(cn, "_var")]] <- vapply(seq_len(nLib),
                                        function(i) results[[2L*i - 1L]]["var"],
                                        numeric(1))
    out[[paste0(tn, "_var")]] <- vapply(seq_len(nLib),
                                        function(i) results[[2L * i]]["var"],
                                        numeric(1))
  }
  if (showPlot) { PlotCCM(out, E) }
  out = EdmFinalize(out, out, pathOut, predictFile, parameterList, parameters)
}

#------------------------------------------------------------------------
#' Single-variable delay embedding (CCM convention; NA in invalid rows).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
CcmSingleEmbed <- function(vec, shifts) {
  N <- length(vec); E <- length(shifts)
  emb <- matrix(NA_real_, N, E)
  for (dim in seq_len(E)) {
    s <- shifts[dim]
    if (s <= 0) { emb[(1 - s):N, dim] <- vec[1:(N + s)] }
    else        { emb[1:(N - s), dim] <- vec[(s + 1):N] }
  }
  emb
}

#------------------------------------------------------------------------
#' Mixed-multivariate delay embedding: horizontal stack of per-variable
#' embeddings. Returns emb and the AND of per-variable validity masks.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
CcmBuildEmbedding <- function(vectors, shifts) {
  blocks <- lapply(vectors, CcmSingleEmbed, shifts = shifts)
  emb    <- do.call(cbind, blocks)
  valid  <- !apply(emb, 1, function(r) any(is.na(r)))
  list(emb = emb, valid = valid)
}

#------------------------------------------------------------------------
#' True where t+Tp is in bounds and vec[t+Tp] is finite.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
CcmTpValidMask <- function(N, vec, Tp) {
  shifted  <- seq_len(N) + Tp
  inBounds <- shifted >= 1 & shifted <= N
  clipped  <- pmin(pmax(shifted, 1), N)
  inBounds & !is.na(vec[clipped])
}

#------------------------------------------------------------------------
#' Pearson r over finite pairs; NA if fewer than 3 valid pairs.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
NanSafePearson <- function(pred, act) {
  ok <- !(is.na(pred) | is.na(act))

  if (sum(ok) < 3) { return(NA_real_) }

  p <- pred[ok]; a <- act[ok]
  pm <- p - mean(p); am <- a - mean(a)
  denom <- sqrt(sum(pm^2) * sum(am^2))

  if (denom == 0) { return(0) }

  sum(pm * am) / denom
}

#------------------------------------------------------------------------
#' KNN query of all points against an L-point library (unsquared Euclidean).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
CcmKnnQuery <- function(libEmbed, queryEmbed, kQuery, backend) {
  if (backend == "RANN") {
    r <- RANN::nn2(libEmbed, queryEmbed, k = kQuery,
                   treetype = "kd", searchtype = "standard", eps = 0)
    return(list(idx = r$nn.idx, dist = r$nn.dists))
  }
  M    <- nrow(queryEmbed)
  idx  <- matrix(0L, M, kQuery)
  dist <- matrix(0, M, kQuery)
  libT <- t(libEmbed)
  for (i in seq_len(M)) {
    d <- sqrt(colSums((libT - queryEmbed[i, ])^2))
    o <- order(d)[seq_len(kQuery)]
    idx[i, ]  <- o
    dist[i, ] <- d[o]
  }
  list(idx = idx, dist = dist)
}

#------------------------------------------------------------------------
#' Run S subsample CCM trials at one library size (CCM.py::_ccm_for_libsize).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
CcmForLibSize <- function(embed, predVals, timeIndices, L, S, k,
                          exclusionRadius, backend) {
  M <- nrow(embed)
  L <- min(L, M)
  k <- min(k, L - 1L)
  if (k < 1L) return(rep(NA_real_, S))

  kQuery <- k + 1L
  if (exclusionRadius > 0) { kQuery <- kQuery + 2L * exclusionRadius }
  kQuery <- min(kQuery, L)

  rhos <- numeric(S)
  rowM <- seq_len(M)

  for (s in seq_len(S)) {
    libIdx   <- sort(sample.int(M, L))               # replace = FALSE
    libEmbed <- embed[libIdx, , drop = FALSE]
    q        <- CcmKnnQuery(libEmbed, embed, kQuery, backend)
    nnGlobal <- matrix(libIdx[q$idx], nrow = M)      # local -> M-space
    nnDist   <- q$dist

    isSelf <- nnGlobal == rowM                       # column-recycled compare
    if (exclusionRadius > 0) {
      nnT    <- matrix(timeIndices[nnGlobal], nrow = M)
      isExcl <- abs(timeIndices[rowM] - nnT) <= exclusionRadius
      mask   <- isSelf | isExcl
    }
    else {
      mask <- isSelf
    }

    valid <- !mask
    cs    <- RowCumsum(valid)
    firstK       <- valid & (cs <= k)
    totalFound   <- cs[, ncol(cs)]
    insufficient <- totalFound < k

    if (all(insufficient)) { rhos[s] <- NA_real_; next }

    firstK[insufficient, ] <- FALSE

    suf  <- which(!insufficient)
    sel  <- t(vapply(suf, function(r) which(firstK[r, ])[seq_len(k)],
                     integer(k)))                    # nSuf x k column positions
    if (k == 1L) sel <- matrix(sel, ncol = 1L)

    rr   <- matrix(suf, length(suf), k)
    Dsel <- matrix(nnDist[cbind(as.vector(rr), as.vector(sel))],
                   nrow = length(suf))
    Gsel <- matrix(nnGlobal[cbind(as.vector(rr), as.vector(sel))],
                   nrow = length(suf))

    # CCM Simplex weighting (no 1e-6 floor)
    dMin   <- Dsel[, 1]
    dMinNz <- ifelse(dMin > 0, dMin, 1)
    W      <- exp(-Dsel / dMinNz)
    zero   <- dMin == 0
    if (any(zero)) {
      W[zero, ] <- (Dsel[zero, , drop = FALSE] == 0) * 1
    }
    wSum <- rowSums(W); wSum <- ifelse(wSum > 0, wSum, 1)
    W    <- W / wSum

    predictions      <- rep(NA_real_, M)
    predictions[suf] <- rowSums(W * matrix(predVals[Gsel], nrow = length(suf)))
    rhos[s] <- NanSafePearson(predictions, predVals)
  }
  rhos
}

#------------------------------------------------------------------------
#' Expand a libSizes argument (explicit list or start/stop/increment).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
CcmLibSizes <- function(libSizes, E, nRows) {
  if (is.character(libSizes)) {
    libSizes <- as.integer(strsplit(trimws(libSizes), "\\s+")[[1]])
  }
  libSizes <- as.integer(libSizes)
  
  if (length(libSizes) == 3L) {
    start <- libSizes[1]; stop <- libSizes[2]; incr <- libSizes[3]
    if (incr < stop) libSizes <- seq.int(start, stop, by = incr)
  }

  if (libSizes[length(libSizes)] > nRows) {
    stop("CCM: maximum libSize exceeds data size.")
  }

  libSizes
}

#------------------------------------------------------------------------
# Nearest-neighbour search and exclusion.
#
# Port of pyEDM/src/pyEDM/Neighbors.py::FindNeighbors (2.5.0).
# See spec: rEDM_Neighbors_spec.md.
#
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#' Find k nearest library neighbours for each prediction point.
#'
#' @param embedding      numeric matrix (N x E), row-indexed by data row.
#' @param libRows        integer vector of 1-based library data-row indices.
#' @param predRows       integer vector of 1-based prediction data-row indices.
#' @param knn            neighbours per prediction (assumed already defaulted).
#' @param exclusionRadius Theiler window in rows (>= 0).
#' @param validLib       logical vector (length N) of admissible library rows,
#'                       or length-0 for "all".
#' @param libOverlap     TRUE if libRows and predRows intersect.
#' @param xRadKnnFactor  over-query multiplier for exclusionRadius (default 5).
#' @param backend        "RANN" or "brute".
#' @param verbose        emit progress / deficiency warnings.
#'
#' @return list(neighbors, distances): each an (nPred x knn) matrix of
#'   1-based library data-row indices and unsquared Euclidean distances,
#'   ascending by distance per row, after exclusion.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
FindNeighbors <- function(embedding, libRows, predRows, knn,
                          exclusionRadius = 0, validLib = logical(0),
                          libOverlap = FALSE, xRadKnnFactor = 5,
                          backend = c("RANN", "brute"), tieBreak = FALSE,
                          verbose = FALSE) {
  backend <- match.arg(backend)
  nPred   <- length(predRows)

  #-----------------------------------------------
  # (3.1) Determine if exclusionRadius filtering needed
  #-----------------------------------------------
  exclusionRadiusKnn <- FALSE
  if (exclusionRadius > 0) {
    if (libOverlap) {
      exclusionRadiusKnn <- TRUE
    }
    else {
      excludeRow <- 0
      nLibCur    <- length(libRows)
      if (predRows[1] > libRows[nLibCur]) {
        excludeRow <- predRows[1] - libRows[nLibCur] # pred start beyond lib end
      }
      else if (libRows[1] > predRows[nPred]) {
        excludeRow <- libRows[1] - predRows[nPred]   # lib start beyond pred end
      }
      if (exclusionRadius >= excludeRow) { exclusionRadiusKnn <- TRUE }
    }
  }

  #-----------------------------------------------
  # (3.2) Filter library by validLib (order preserving)
  #-----------------------------------------------
  if (length(validLib)) {
    validLibIdx <- which(validLib)                       # 1-based data rows
    libRows     <- libRows[libRows %in% validLibIdx]
    if (length(libRows) == 0L) {
      stop("FindNeighbors: no valid library points; all excluded by validLib.")
    }
    if (length(libRows) < knn && verbose) {
      warning(sprintf("FindNeighbors: %d valid library points, but knn = %d.",
                      length(libRows), knn))
    }
  }
  nLib <- length(libRows)

  #-----------------------------------------------
  # (3.3) Determine kQuery : neighbours to request
  #-----------------------------------------------
  kQuery <- knn
  if (exclusionRadiusKnn) {
    kQuery <- min(knn * xRadKnnFactor, nLib)
  }
  else if (libOverlap) {
    kQuery <- knn + 1L
  }
  
  if (length(validLib)) {
    kQuery <- nLib                    # examine all
  }
  
  if (tieBreak && nLib > knn) {
    # tieBreak needs a lookahead column even in the plain (disjoint) case
    # so a boundary tie can be detected and completed via full scan.
    kQuery <- min(max(kQuery, knn + 1L), nLib)
  }
  kQuery <- max(1L, min(kQuery, nLib)) # never exceed nLib

  #-----------------------------------------------
  # (3.4) Query
  #-----------------------------------------------
  libData  <- embedding[libRows,  , drop = FALSE]
  predData <- embedding[predRows, , drop = FALSE]
  q        <- KNNQuery(libData, predData, kQuery, backend = backend)
  nnIdx    <- q$nnIdx  # (nPred x kQuery) 1-based into libRows, 0 = sentinel
  nnDist   <- q$nnDist

  #-----------------------------------------------
  # (3.5, 3.6) Sentinel guard + remap to data rows
  #-----------------------------------------------
  sentinel  <- nnIdx == 0L
  neighbors <- matrix(0L, nrow(nnIdx), ncol(nnIdx))
  if (any(!sentinel)) {
    neighbors[!sentinel] <- libRows[nnIdx[!sentinel]] # 1-based, no offset
  }
  distances <- nnDist
  distances[sentinel] <- Inf

  #-----------------------------------------------
  # (3.7) Exclusion mask + compaction
  #-----------------------------------------------
  needsFiltering <- libOverlap || exclusionRadiusKnn ||
                    kQuery > knn || any(sentinel) || tieBreak

  if (!needsFiltering) {
    return(list(neighbors = neighbors, distances = distances))
  }

  predCol <- matrix(predRows, nPred, kQuery)            # broadcast down columns

  if (exclusionRadiusKnn) {
    mask <- abs(predCol - neighbors) <= exclusionRadius # subsumes self-match
  }
  else if (libOverlap) {
    mask <- predCol == neighbors                        # self-match only
  }
  else {
    mask <- matrix(FALSE, nPred, kQuery)                # validLib trim only
  }
  mask <- mask | sentinel                               # always exclude sentinel

  #-----------------------------------------------
  # (3.7t) Deterministic tie-braking (Simplex only).
  # Backend-independent: ordering rules:
  #    1. distance
  #    2. |predRow-libRow| (proximity to prediction)
  #    3. libRow           (preceeding proximal)
  # applied to eligible candidates, with full-scan completion for rows whose
  # knn-th distance reaches the over-query boundary. Shared with pyEDM so the
  # two packages agree on degenerate (tied) data.
  #-----------------------------------------------
  if (tieBreak) {
    return(TieBreakSelect(embedding, libRows, predRows, neighbors, distances,
                          mask, knn, exclusionRadius, exclusionRadiusKnn,
                          libOverlap, kQuery, verbose))
  }

  valid   <- !mask
  cs      <- RowCumsum(valid)
  firstK  <- valid & (cs <= knn)

  # Deficiency: fewer than knn valid neighbours in a row. Such rows are NOT
  # back-filled with excluded / self neighbours (that reinstates the
  # self-match at distance 0 and leaks the target; issue #74 / pyEDM PR #74).
  # Their surplus slots are left as Inf / 0 padding by the compaction below.
  validCounts <- cs[, kQuery]
  if (verbose) {
    nDeficient <- sum(validCounts < knn & validCounts > 0)
    nEmpty     <- sum(validCounts == 0)
    if (nDeficient > 0) {
      warning(sprintf(
        paste("FindNeighbors: fewer than knn=%d neighbours outside",
              "exclusionRadius=%d for %d prediction(s); those rows use",
              "fewer neighbours. Consider reducing knn or exclusionRadius."),
        knn, exclusionRadius, nDeficient))
    }
    if (nEmpty > 0) {
      warning(sprintf(
        paste("FindNeighbors: %d prediction(s) have no valid neighbours",
              "outside exclusionRadius=%d; those predictions are NA."),
        nEmpty, exclusionRadius))
    }
  }

  # Compact: leftmost selected columns per row (stable), take first kOut.
  # Rows with fewer than kOut selected keep their Inf / 0 padding in the
  # surplus slots; every consumer treats a non-finite distance as
  # "no neighbour" (issue #74).
  kOut         <- min(knn, kQuery)
  outNeighbors <- matrix(0L,  nPred, kOut)
  outDist      <- matrix(Inf, nPred, kOut)

  for (i in seq_len(nPred)) {
    sel <- which(firstK[i, ])
    if (length(sel) > kOut) { sel <- sel[seq_len(kOut)] }
    m <- length(sel)
    if (m) {
      outNeighbors[i, seq_len(m)] <- neighbors[i, sel]
      outDist[i,      seq_len(m)] <- distances[i, sel]
    }
  }

  list(neighbors = outNeighbors, distances = outDist)
}


#------------------------------------------------------------------------
#' Deterministic tie-broken knn selection (Simplex only).
#'
#' Applied only when tieBreak = TRUE (the Simplex path). Selects knn
#' neighbours per prediction row by the ordering key
#'   (distance asc, |predRow - libRow| asc, libRow asc)
#' on original 1-based data-row indices. A full-library scan (FullScanRow)
#' completes only rows whose knn-th distance reaches the over-query
#' boundary (a possible straddling tie) or that are deficient; all other
#' rows are resolved from the queried candidates. Backend-independent by
#' construction, so RANN and brute agree, and match pyEDM.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
TieBreakSelect <- function(embedding, libRows, predRows, neighbors, distances,
                           mask, knn, exclusionRadius, exclusionRadiusKnn,
                           libOverlap, kQuery, verbose) {
  nPred        <- length(predRows)
  nLib         <- length(libRows)
  kOut         <- min(knn, kQuery)
  outNeighbors <- matrix(0L,  nPred, kOut)
  outDist      <- matrix(Inf, nPred, kOut)
  canComplete  <- kQuery < nLib
  embLib       <- NULL
  warnedDef    <- FALSE

  for (i in seq_len(nPred)) {
    p        <- predRows[i]
    elig     <- which(!mask[i, ])
    needScan <- FALSE
    nbrO     <- integer(0)
    dstO     <- numeric(0)

    if (length(elig) >= knn) {
      nbr  <- neighbors[i, elig]
      dst  <- distances[i, elig]
      ord  <- order(dst, abs(p - nbr), nbr)     # distance, proximity, index
      nbrO <- nbr[ord]
      dstO <- dst[ord]
      if (canComplete) {
        finite <- distances[i, is.finite(distances[i, ])]
        if (length(finite) && dstO[knn] >= max(finite)) needScan <- TRUE
      }
    }
    else {
      needScan <- canComplete                   # too few eligible in query
    }

    if (needScan) {
      if (is.null(embLib)) embLib <- embedding[libRows, , drop = FALSE]
      # Exclusion-respecting scan only : never ignore the exclusion radius
      # to complete a deficient row (that reinstates the self-match at
      # distance 0 and leaks the target; issue #74). Deficient rows keep
      # their Inf padding.
      fs <- FullScanRow(embLib, libRows, embedding[p, ], p, knn,
                        exclusionRadius, exclusionRadiusKnn, libOverlap)
      if (is.null(fs) || length(fs$nbr) < knn) warnedDef <- TRUE
      nbrO <- if (is.null(fs)) integer(0) else fs$nbr
      dstO <- if (is.null(fs)) numeric(0) else fs$dst
    }

    m <- min(kOut, length(nbrO))
    if (m) {
      outNeighbors[i, seq_len(m)] <- nbrO[seq_len(m)]
      outDist[i,      seq_len(m)] <- dstO[seq_len(m)]
    }
  }

  if (warnedDef && verbose) {
    warning(sprintf(
      paste("FindNeighbors: fewer than knn=%d neighbours outside",
            "exclusionRadius=%d for some prediction(s); those rows use",
            "fewer neighbours. Consider reducing knn or exclusionRadius."),
      knn, exclusionRadius))
  }

  list(neighbors = outNeighbors, distances = outDist)
}

#------------------------------------------------------------------------
#' Exact full-library scan for one prediction row, ordered by the key.
#'
#' Returns list(nbr, dst) of the knn nearest library rows under
#' (distance, |p - libRow|, libRow), or NULL when no rows remain.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
FullScanRow <- function(embLib, libRows, predVec, p, knn,
                        exclusionRadius, exclusionRadiusKnn, libOverlap) {
  diff <- sweep(embLib, 2L, predVec, "-")
  d    <- sqrt(rowSums(diff * diff))
  if (exclusionRadiusKnn) {
    keep <- abs(p - libRows) > exclusionRadius   # subsumes self-match
  }
  else if (libOverlap) {
    keep <- libRows != p                         # self-match only
  }
  else {
    keep <- rep(TRUE, length(libRows))
  }
  nbr <- libRows[keep]
  dd  <- d[keep]
  if (length(nbr) == 0L) { return(NULL) }

  ord <- order(dd, abs(p - nbr), nbr)
  sel <- ord[seq_len(min(knn, length(ord)))]
  list(nbr = nbr[sel], dst = dd[sel])
}

#------------------------------------------------------------------------
#' Exact k nearest-neighbour query (backend abstraction).
#'
#' Returns 1-based indices into \code{libData} and unsquared Euclidean
#' distances, ascending per row. When \code{k} exceeds the number of
#' library rows the surplus columns are filled with the "no neighbour"
#' sentinel (index 0, distance Inf), matching RANN's convention.
#'
#' @param libData  numeric matrix (nLib x E), the library embedding.
#' @param predData numeric matrix (nPred x E), the query embedding.
#' @param k        neighbours to request.
#' @param backend  "RANN" (production) or "brute" (exact base-R fallback).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
KNNQuery <- function(libData, predData, k, backend = c("RANN", "brute")) {
  backend <- match.arg(backend)
  nLib    <- nrow(libData)
  nPred   <- nrow(predData)
  kEff    <- min(k, nLib)

  nnIdx  <- matrix(0L,  nPred, k)   # sentinel index 0
  nnDist <- matrix(Inf, nPred, k)   # sentinel distance Inf

  if (backend == "RANN") {
    if (!requireNamespace("RANN", quietly = TRUE)) {
      stop("KNNQuery: package 'RANN' not available; ",
           "install RANN or call with backend = 'brute'.")
    }
    res <- RANN::nn2(data = libData, query = predData, k = kEff,
                     treetype = "kd", searchtype = "standard", eps = 0)
    nnIdx[,  seq_len(kEff)] <- res$nn.idx
    nnDist[, seq_len(kEff)] <- res$nn.dists
  }
  else {
    # Vectorised exact brute force: d^2 = |p|^2 + |l|^2 - 2 p.l
    libSq  <- rowSums(libData  * libData)          # nLib
    predSq <- rowSums(predData * predData)          # nPred
    cross  <- predData %*% t(libData)               # nPred x nLib
    d2     <- outer(predSq, libSq, "+") - 2 * cross # nPred x nLib
    d2[d2 < 0] <- 0                                 # guard FP negatives
    d  <- sqrt(d2)
    for (i in seq_len(nPred)) {
      o <- order(d[i, ])[seq_len(kEff)]            # stable; ties by index
      nnIdx[i,  seq_len(kEff)] <- o
      nnDist[i, seq_len(kEff)] <- d[i, o]
    }
  }

  list(nnIdx = nnIdx, nnDist = nnDist)
}

#------------------------------------------------------------------------
#' Row-wise cumulative sum of a logical / numeric matrix (vectorised).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
RowCumsum <- function(m) {
  k <- ncol(m)
  upper <- upper.tri(matrix(0, k, k), diag = TRUE) * 1 # k x k, U[a,b]=1 if a<=b
  (m + 0) %*% upper
}

#------------------------------------------------------------------------
# Multiview embedding (Ye & Sugihara 2016).
#
# Port of pyEDM Multiview (Multiview.py, API.Multiview, PoolFunc). Build the
# multivariate delay embedding, form all D-column combinations ("views"),
# rank each view by in-/out-of-sample Simplex rho, average the predictions of
# the top sqrt(nCombo) views, and report the ranked view table.
#
# Built entirely on the validated Simplex path (embedded, noTime), so results
# follow Simplex exactly; only view selection and averaging are new.
#
# Naming: functions/constants uppercase-first, variables lowercase-first.
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#' Multiview embedding forecast.
#'
#' @param dataFrame data.frame of source data.
#' @param columns library variable name(s) (vector or space string).
#' @param target target variable name.
#' @param lib,pred (start end) index pairs.
#' @param D view dimension (number of embedding columns combined; default
#'   = number of input columns).
#' @param E per-variable embedding dimension.
#' @param Tp,knn,tau,exclusionRadius standard EDM parameters.
#' @param multiview number of top views to average (default floor(sqrt(nCombo))).
#' @param trainLib rank on the library (in-sample) rather than pred.
#' @param excludeTarget drop the target variable from candidate views.
#' @param numProcess worker processes for the ranking grid.
#' @param backend "RANN" or "brute".
#' @return list(Predictions = data.frame(Time, Observations, Predictions),
#'   View = data.frame(Columns, rho, MAE, CAE, RMSE)).
#' @keywords internal
#------------------------------------------------------------------------
Multiview <- function(dataFrame = NULL, columns, target, lib = numeric(0),
                      pred = numeric(0), D = 0, E = 1, Tp = 1, knn = 0,
                      tau = -1, multiview = 0, exclusionRadius = 0,
                      trainLib = TRUE, excludeTarget = FALSE, ignoreNan = TRUE,
                      numProcess = 4, backend = "RANN",
                      pathIn = "./", dataFile = "", pathOut = "./", predictFile = "",
                      parameterList = FALSE, showPlot = FALSE) {
    
  parameters <- as.list(environment())
  parameters[c("dataFrame")] <- NULL

  dataFrame <- ResolveInput(pathIn, dataFile, dataFrame)
  columns   <- SplitColumns(columns)
  target    <- SplitColumns(target)
  target1   <- paste0(target[1], "(t-0)")             # embedding-column label
  N         <- nrow(dataFrame)

  if (trainLib && !length(lib) && !length(pred)) {
    lib  <- c(1, floor(N / 2))
    pred <- c(floor(N / 2) + 1, N)
  }
  if (D == 0) D <- length(columns)
  comboCols <- if (excludeTarget) setdiff(columns, target) else columns

  # Multivariate delay embedding as a named, time-less data.frame
  emb      <- MakeBlock(dataFrame, E, tau, comboCols)
  embNames <- EmbedColumnNames(comboCols, E, tau, embedded = FALSE)
  embDf    <- as.data.frame(emb, stringsAsFactors = FALSE)
  names(embDf) <- embNames

  combos <- utils::combn(embNames, D, simplify = FALSE)
  if (multiview < 1)              multiview <- floor(sqrt(length(combos)))
  if (multiview > length(combos)) multiview <- length(combos)

  RunView <- function(combo, predRange)
    Simplex(embDf, columns = combo, target = target1, lib = lib,
            pred = predRange, E = D, Tp = Tp, tau = tau,
            exclusionRadius = exclusionRadius, embedded = TRUE,
            noTime = TRUE, ignoreNan = ignoreNan, backend = backend,
            .tieBreak = FALSE)   # Multiview ranking keeps 2.0.2 neighbour order

  # ---- Rank views by Simplex rho ----
  rankPred <- if (trainLib) lib else pred
  rhoVec <- unlist(SweepApply(combos, function(combo) {
    df <- RunView(combo, rankPred)
    ComputeError(df$Observations, df$Predictions)$rho
  }, numProcess))

  # rev(order()) matches numpy argsort()[::-1]: NA last ascending -> first here
  topRank   <- rev(order(rhoVec))[seq_len(multiview)]
  topCombos <- combos[topRank]

  # ---- Project top views and average ----
  projList <- SweepApply(topCombos, function(combo) RunView(combo, pred),
                         numProcess)

  predMat <- vapply(projList, function(d) d$Predictions,
                    numeric(nrow(projList[[1]])))
  if (is.null(dim(predMat))) predMat <- matrix(predMat, ncol = length(projList))
  multiviewPredict <- rowMeans(predMat)

  first <- projList[[1]]
  predictions <- data.frame(Time = first$Time,
                            Observations = first$Observations,
                            Predictions = multiviewPredict,
                            stringsAsFactors = FALSE, check.names = FALSE)

  # ---- View ranking table ----
  stats <- lapply(projList, function(d) ComputeError(d$Observations, d$Predictions))
  view  <- data.frame(
    Columns = vapply(topCombos,
                     function(c) paste0("(",paste0("'", c, "'", collapse = ", "), ")"),
                     character(1)),
    rho  = vapply(stats, `[[`, numeric(1), "rho"),
    MAE  = vapply(stats, `[[`, numeric(1), "MAE"),
    CAE  = vapply(stats, `[[`, numeric(1), "CAE"),
    RMSE = vapply(stats, `[[`, numeric(1), "RMSE"),
    stringsAsFactors = FALSE, check.names = FALSE)

  if (showPlot) PlotObsPred(predictions, "Multiview  ", D, Tp)
  if (showPlot) PlotObsPred(predictions, "", D, Tp)
  result = list(Predictions = predictions, View = view)
  result = EdmFinalize( result, predictions, pathOut, predictFile,
                        parameterList, parameters )
}

library( rEDM )

#-------------------------------------------------------------------
# EDM Embed wrapper
# Create time-delay embedding with time column for EDM. 
# Useful to create mixed multivariate embeddings for SMap and
# embeddings with time-advanced vectors.  
# Rename V(t-0), V(t+0) to V. Add Time column.
# If columns is NULL, embedd all except the first (time) column.
# If plusminus create time-advanced & time-delayed columns.
#-------------------------------------------------------------------
Embedding = function(
  dataFrame = NULL,
  dataFile  = NULL,        
  outFile   = NULL,
  plusminus = FALSE,
  columns   = NULL,
  E         = 2,
  tau       = -1,
  verbose   = FALSE
) {

  if ( is.null( dataFrame ) & is.null( dataFile ) ) {
    stop( 'dataFrame and dataFile are empty, specify one.' )
  }
  if ( tau > 0 & plusminus ) {
    # Convert to negative
    tau = -tau
  }

  if ( is.null( dataFrame ) ) {
    # Load from dataFile
    data = read.csv( dataFile )
  }
  else {
    data = dataFrame
  }

  # Presume time is first column
  timeName   = colnames( data )[1]
  timeSeries = data[ , timeName ]

  # If no columns specified, use all except first
  if ( is.null( columns ) ) {
    columns = colnames( data )[ 2 : ncol( data ) ]
  }

  if ( verbose ) {
    print( paste( "Time column: ", timeName ) )
    print( "Embed columns: " ); print( columns )
  }

  # Create embeddings of columns
  # There will be redundancies vis V1(t-0), V1(t+0)
  if ( plusminus ) {
    embed_minus = Embed( dataFrame = data, E = E, tau = tau, columns = columns )
    embed_plus  = Embed( dataFrame = data, E = E, tau = abs( tau ),
                         columns = columns )
    embed = cbind( timeSeries, embed_minus, embed_plus, stringsAsFactors=FALSE )

    # TRUE / FALSE vector
    cols_tplus0 = grepl( '(t+0)', colnames( embed ), fixed = TRUE )
    # Remove *(t+0) : redunant with *(t-0)
    embed = embed[ , !cols_tplus0 ]
  }
  else {
    embed_ = Embed( dataFrame = data, E = E, tau = tau, columns = columns )
    embed  = cbind( timeSeries, embed_, stringsAsFactors = FALSE )
  }

  # Rename *(t-0) to original column names
  columnNames = colnames( embed )
  for ( i in 1:length( columnNames ) ) {
    if ( grepl( '(t-0)', columnNames[i], fixed = TRUE ) ) {
      columnNames[i] = sub( '(t-0)', '', columnNames[i], fixed = TRUE )
    }
  }

  # Rename *(t+0) to original column names
  for ( i in 1:length( columnNames ) ) {
    if ( grepl( '(t+0)', columnNames[i], fixed = TRUE ) ) {
      columnNames[i] = sub( '(t+0)', '', columnNames[i], fixed = TRUE )
    }
  }

  # Rename first column to original time column name
  columnNames[ 1 ]  = timeName
  colnames( embed ) = columnNames

  if ( verbose ) {
    print( head( embed, 4 ) )
    print( tail( embed, 4 ) )
  }

  if ( ! is.null( outFile ) ) {
    write.csv( embed, file = outFile, row.names = FALSE )
  }

  return( embed )
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(echo = TRUE, warning = FALSE, tidy = TRUE,
                      fig.pos = 'h', fig.align = 'center' )
                      #fig.width = 4.5, fig.height = 4 )
knitr::opts_knit$set(global.par = TRUE, progress = FALSE)
options(digits = 2)
par.orig = par()
par(mar = c(3, 3, 2, 1), mgp = c(2.5, 1, 0))

## ----CRAN_installation_instructions, eval = FALSE-----------------------------
# install.packages("rEDM")

## ----GitHub_installation_instructions, eval = FALSE---------------------------
# devtools::install_github("SugiharaLab/rEDM")

## ----LorenzProjection, echo = FALSE, fig.cap = "Time Series Projection from the Lorenz Attractor.", out.width = "50%", fig.align = 'center'----
knitr::include_graphics("Lorenz_Projection.png")

## ----fig_attractor_reconstruction, echo = FALSE, fig.cap = "Attractor Reconstruction from 3 Lagged Coordinates", out.width = "50%", fig.align = 'center'----
knitr::include_graphics("Lorenz_Reconstruct.png")

## ----load_package-------------------------------------------------------------
library(rEDM)
paste('rEDM version:', packageVersion('rEDM'))

## ----examine_TentMap_dataset--------------------------------------------------
str(TentMap)

## ----simplex_tentmap----------------------------------------------------------
simplex_out <- Simplex( dataFrame = TentMap, lib = "1 100", pred = "900 999",
columns = 'TentMap', target = 'TentMap', E = 3 )
simplex_out[ c(1:3, 99:101), ]

## ----tentmap_simplex_stats----------------------------------------------------
ComputeError( simplex_out $ Observations, simplex_out $ Predictions )

## ----Embed_Dim_TentMap, fig.cap="TentMap data prediction skill vs. embedding dimension.", fig.width=5, fig.height=4.5----
rho_E <- EmbedDimension( dataFrame = TentMap, lib = "1 100", pred = "201 500",
columns = 'TentMap', target = 'TentMap', showPlot = TRUE )
rho_E

## ----Prediction_interval_on_TentMap, fig.cap = "Tent map first differences simplex prediction skill as a function of forecast interval."----
rho_Tp <- PredictInterval( dataFrame = TentMap, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )
rho_Tp

## ----Predict_nonlinear_on_TentMap, fig.cap = "Tent map first differences S-map prediction skill as a function of S-map localisation parameter."----
rho_theta <- PredictNonlinear( dataFrame = TentMapNoise, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )
rho_theta[c(1:5,11:15),]

## ----Simplex_TentMap----------------------------------------------------------
tentMapPredict <- Simplex( dataFrame = TentMap, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )

ComputeError( tentMapPredict $ Observations, tentMapPredict $ Predictions ) $ rho

## ----Tentmap_Noise_S-map------------------------------------------------------
smap = SMap( dataFrame = TentMapNoise, lib = "1 100", pred = "201 500",
target = 'TentMap', columns = 'TentMap', E = 2, theta = 3 )

## ----S-map_tentMap_results----------------------------------------------------
head( cbind( smap $ predictions, smap $ coefficients ), 2 )
tail( cbind( smap $ predictions, smap $ coefficients ), 2 )

## ----block_3sp_data-----------------------------------------------------------
head( block_3sp, 3 )

## ----3_Species_simplex, fig.cap="Simplex Tp=1 forecast of $x_t$."-------------
smplx_3species = Simplex( dataFrame = block_3sp, lib = "1 100",
pred = "101 190", E = 3, columns = "x_t x_t-1 z_t", target = "x_t",
embedded = TRUE )

## ----plot_3_species, fig.cap="Scatter plot of simplex forecast of $x_t$ vs. observations.", fig.width=5, fig.height=4.5----
err = ComputeError( smplx_3species $ Observations,
                    smplx_3species $ Predictions )
plot( smplx_3species $ Observations, smplx_3species $ Predictions,
pch = 19, cex = 0.5, xlab = 'Observations', ylab = 'Predictions',
main = '3 Species x_t' )
abline( a = 0, b = 1, lty = 2, col = 'blue' )
text( -1, 1, paste( capture.output( cbind( err ) ), collapse = '\n' ) )

## ----S-map_Lorenz96-----------------------------------------------------------
smap_Lorenz <- SMap( dataFrame = Lorenz5D, lib = "1 500", pred = "601 900",
E = 4, theta = 3, columns = "V1 V2 V3 V4", target = "V1", embedded = TRUE )

## ----set_precision, echo = FALSE----------------------------------------------
options(digits=4)

## ----S-map_Lorenz_results, warning = FALSE------------------------------------
head( cbind( smap_Lorenz $ predictions, smap_Lorenz $ coefficients[,2:6] ), 3 )

## ----echo = FALSE-------------------------------------------------------------
par(mfrow = c(4, 1), mar = c(2, 4, 0.5, 1), oma = c(0, 0, 0, 0),
    mgp = c(2.1, 1, 0) )

## ----smap_Lorenz_plot, out.width = "70%", fig.height = 4.5, fig.cap = "S-map prediction and coefficients of Lorenz'96 5-D system."----
predictions = smap_Lorenz $ predictions
coefficients = smap_Lorenz $ coefficients
Time = predictions $ Time

plot( Time, predictions $ Observations, type = "l", col = "blue",
ylab = "V1", xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
lines( Time, predictions $ Predictions, lwd = 2, col = 'red' )
legend("topright", legend = c("observed", "predicted"),
fill = c( 'blue', 'red' ), bty = "n", cex = 1.3 )

plot( Time, coefficients[, 6], type = "l", col = "brown",
ylab = expression(partialdiff * V4 / partialdiff * V1),
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
plot( Time, coefficients[, 5], type = "l", col = "darkgreen",
ylab = expression(partialdiff * V3 / partialdiff * V1),
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
plot( Time, coefficients[, 4], type = "l", col = "blue",
ylab = expression(partialdiff * V2 / partialdiff * V1),
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )

## ----echo = FALSE-------------------------------------------------------------
par(mfrow = c(1, 1), mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))

## ----multiview----------------------------------------------------------------
Mview = Multiview( dataFrame = block_3sp, lib = "1 100", pred = "101 190", E = 3, columns = "x_t y_t z_t", target = "x_t" )

## ----multiview results--------------------------------------------------------
Mview $ View[ which( Mview $ View $ rho > 0.91 ),  ]

## ----fig_cross_map, echo = FALSE, fig.cap = "Cross Mapping Between Reconstructions of the Lorenz Attractor"----
knitr::include_graphics("CrossMap.png")

## ----sardine_anchovy_ccm, fig.cap = "Convergent cross mapping of Newport sea surface temperature with anchovy landings.", fig.width=5, fig.height=4.5----
cmap <- CCM( dataFrame = sardine_anchovy_sst, E = 3, Tp = 0, columns = "anchovy", target = "np_sst", libSizes = "10 70 5", sample = 100, showPlot = TRUE )

## ----Thrips data--------------------------------------------------------------
head(Thrips, 2)

## ----thrips_plot, echo = FALSE, fig.width = 6, fig.height = 5, fig.cap = "Thrips abundance and environmental variables."----
par(mfrow = c(4, 1), mar = c(2, 4, 0, 1), mgp = c(2., 0.5, 0))
time_dec <- Thrips $ Year + (Thrips $ Month)/12
plot(time_dec, Thrips $ Thrips_imaginis, type = "l", lwd = 2, col = "darkgreen", ylab = "Thrips", xlab = "")
plot(time_dec, Thrips $ maxT_degC, type = "l", lwd = 2, col = "red", ylab = "maxT (C)", xlab = "")
plot(time_dec, Thrips $ Rain_mm, type = "l", lwd = 2, col = "blue", ylab = "Rain (mm)", xlab = "")
plot(time_dec, Thrips $ Season, type = "l", lwd = 2, col = "magenta", ylab = "Season", xlab = "" )
mtext("Year", side = 1, outer = TRUE, line = 1)

## ----echo = FALSE-------------------------------------------------------------
par(mfrow = c(1, 1), mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))

## ----univariate_thrips, fig.cap = "Simplex embedding dimension for Thrips abundance.", fig.width=5, fig.height=4.5----
rho_E <- EmbedDimension( dataFrame = Thrips, columns = "Thrips_imaginis", target = "Thrips_imaginis", lib = "1 72", pred = "1 72", showPlot = TRUE )

## ----smap_thrips, fig.cap = "SMap localisation parameter for Thrips abundance.", fig.width = 5, fig.height = 4.5----
E = 8
rho_theta_e3 = PredictNonlinear(dataFrame = Thrips, columns = "Thrips_imaginis", target = "Thrips_imaginis", lib = "1 73", pred = "1 73", E = E, showPlot = TRUE )

## ----compute_ccm_matrix_thrips, results='hold', cache = TRUE------------------
vars       = colnames(Thrips[3:6])
var_pairs  = combn( vars, 2 )  # Combinations of vars, 2 at a time
libSize    = paste( NROW(Thrips) - E, NROW(Thrips) - E, 10, collapse = " " )
ccm_matrix = array(NA, dim = c(length(vars), length(vars)), 
                   dimnames = list(vars, vars))

for( i in 1:ncol( var_pairs ) ) {
   ccm_out = CCM(dataFrame = Thrips,
                 columns = var_pairs[1,i], target = var_pairs[2,i],
                 libSizes = libSize, Tp = 0, E = E, sample = 100)
                 
   outVars = names( ccm_out )

   var_out = unlist( strsplit( outVars[2], ':' ) )
   ccm_matrix[ var_out[2], var_out[1] ] = ccm_out[1,2]
                 
   var_out = unlist( strsplit( outVars[3], ':' ) )
   ccm_matrix[ var_out[2], var_out[1] ] = ccm_out[1,3]
}

## ----compute_corr_matrix_thrips, cache = TRUE---------------------------------
corr_matrix <- array(NA, dim = c(length(vars), length(vars)), 
                     dimnames = list(vars, vars))

for(ccm_from in vars) {
    for(ccm_to in vars[vars != ccm_from]) {
        ccf_out <- ccf(Thrips[,ccm_from], Thrips[,ccm_to], 
                       type = "correlation", lag.max = 6, plot = FALSE)$acf
        corr_matrix[ccm_from, ccm_to] <- max(abs(ccf_out))
    }
}

## ----xmap_vs_corr_matrix_thrips-----------------------------------------------
ccm_matrix
corr_matrix

## ----thrips_setup, echo = FALSE-----------------------------------------------
par( mfrow = c( 1, 3 ), mar = c( 3.5, 1.3, 2.3, 0.2 ), mgp = c(1.5, 0.5, 0) )

CCMPlot = function( ccm.df, E, ylim = c( 0.2, 1 ) ) {
    libSize = ccm.df $ LibSize
    V1      = names( ccm.df )[2]
    V2      = names( ccm.df )[3]

    title = paste( V1, " : ", V2, "\nE=" , E )
    
    print( title )
    print( libSize )

    plot( libSize, ccm.df[ , V1 ],
          ylim = ylim, main = title, col = "blue", type = "l", lwd = 3,
          xlab = 'Library Size', ylab = expression(Prediction~Skill~(rho)) )
    lines( libSize, ccm.df[ , V2 ], col = "red", lwd = 3 )
    abline( h = 0 )
    legend( 'topright', c( V1, V2 ), 
            fill = c( 'blue', 'red' ), bty = 'n', cex = 1.2 )
}

## ----ccm_thrips, fig.cap = "Thrips cross mapped to climatic variables. Vertical axis is cross map prediction skill (rho).", fig.width = 6.5, fig.height=5----
thrips_xmap_maxT <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                        columns = "Thrips_imaginis", target = "maxT_degC",
                        libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_maxT, E )
abline(h = corr_matrix['Thrips_imaginis', 'maxT_degC'], col = "black", lty = 2)

thrips_xmap_Rain <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                        columns = "Thrips_imaginis", target = "Rain_mm",
                        libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_Rain, E )
abline(h = corr_matrix['Thrips_imaginis', 'Rain_mm'], col = "black", lty = 2)

thrips_xmap_Season <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                          columns = "Thrips_imaginis", target = "Season",
                          libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_Season, E )
abline(h = corr_matrix['Thrips_imaginis', 'Season'], col = "black", lty = 2)

## ----thrips_setdown, echo = FALSE---------------------------------------------
par(mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))

## ----seasonal_surrogates_thrips, cache = TRUE---------------------------------
# Create matrix with temperature and rain surrogates (1000 time series vectors)
surr_maxT = SurrogateData( Thrips $ maxT_degC, method = "seasonal", T_period = 12, num_surr = 1000, alpha = 3)
surr_rain = SurrogateData( Thrips $ Rain_mm, method = "seasonal", T_period = 12, num_surr = 1000, alpha = 3)

# Rain cannot be negative
surr_rain = apply( surr_rain, 2, function(x){ i = which(x<0); x[i] = 0; x } )

# data.frame to hold CCM rho values between Thrips abundance and variable
rho_surr <- data.frame( maxT = numeric(1000), Rain = numeric(1000) )

# data.frames with time, Thrips, and 1000 surrogate climate variables for CCM()
maxT_data = as.data.frame( cbind( seq(1:nrow(Thrips)), Thrips $ Thrips_imaginis, surr_maxT ) )
names( maxT_data ) = c( 'time', 'Thrips_imaginis', paste( 'T', as.character( seq( 1,1000 ) ), sep = '' ) )

rain_data = as.data.frame( cbind( seq(1:nrow(Thrips)), Thrips $ Thrips_imaginis, surr_rain ) )
names( rain_data ) = c( 'time', 'Thrips_imaginis', paste( 'R', as.character( seq( 1,1000 ) ), sep = '' ) )

# Cross mapping
for (i in 1:1000) {
  targetCol = paste( 'T', i, sep = '' ) # as in maxT_data
  
  # Note this is a single sample CCM at full library size
  ccm_out = CCM( dataFrame = maxT_data, E = E, Tp = 0,
                 columns = 'Thrips_imaginis', target = targetCol,
                 libSizes = "73", sample = 1 )
                 
  col = paste( 'Thrips_imaginis', ':', targetCol, sep = '' )
   
  rho_surr $ maxT[i] = ccm_out[1,col]
}

for (i in 1:1000) {
  targetCol = paste( 'R', i, sep = '' ) # as in rain_data
  
  # Note this is a single sample CCM at full library size
  ccm_out = CCM( dataFrame = rain_data, E = E, Tp = 0,
                 columns = 'Thrips_imaginis', target = targetCol,
                 libSizes = "73", sample = 1 )
                 
  col = paste( 'Thrips_imaginis', ':', targetCol, sep = '' )
   
  rho_surr $ Rain[i] = ccm_out[1,col]
}

## ----significance_randomization_test------------------------------------------
1 - ecdf( rho_surr $ maxT )( ccm_matrix["maxT_degC", "Thrips_imaginis" ] )
1 - ecdf( rho_surr $ Rain )( ccm_matrix["Rain_mm",   "Thrips_imaginis" ] )

## ----reset_par, include = FALSE-----------------------------------------------
par( par.orig )

#------------------------------------------------------------------------
# ComputeError : Pearson rho, MAE, CAE, RMSE.
#
# Faithful port of pyEDM AuxFunc.ComputeError, including its quirks:
#   - MAE is round(max(obs - pred)) : the SIGNED maximum error (pyEDM labels
#     it MAE but does not take the mean or absolute value). Preserved for
#     bit-compatibility with pyEDM outputs.
#   - CAE is the summed absolute error.
#   - Fewer than 5 finite pairs -> all NaN.
#   - NaN removed from pred first, then obs (Pearson needs finite pairs).
#
# Naming: functions/constants uppercase-first, variables lowercase-first.
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#' Prediction skill statistics.
#'
#' @param obs numeric vector of observations.
#' @param pred numeric vector of predictions (aligned with obs).
#' @param digits rounding applied to each statistic (pyEDM default 6).
#' @return named list(rho, MAE, CAE, RMSE).
#' @keywords internal
#------------------------------------------------------------------------
ComputeError <- function(obs, pred, digits = 6) {
  notNan <- is.finite(pred)
  if (any(!notNan)) { pred <- pred[notNan]; obs <- obs[notNan] }
  notNan <- is.finite(obs)
  if (any(!notNan)) { pred <- pred[notNan]; obs <- obs[notNan] }

  if (length(pred) < 5)
    return(list(rho = NA_real_, MAE = NA_real_, CAE = NA_real_, RMSE = NA_real_))

  err  <- obs - pred
  rho  <- round(suppressWarnings(stats::cor(obs, pred)), digits)   # Pearson
  MAE  <- round(max(err), digits)                                  # signed max (pyEDM)
  CAE  <- round(sum(abs(err)), digits)
  RMSE <- round(sqrt(mean(err^2)), digits)

  list(rho = rho, MAE = MAE, CAE = CAE, RMSE = RMSE)
}

#------------------------------------------------------------------------
# Phase 3 : S-map (sequential locally weighted global linear map).
#
# Port of pyEDM SMap (SMap.py, EDM.py). See spec: rEDM_SMap_solver_spec.md.
#
# Solver replicates numpy.linalg.lstsq(rcond = None): SVD least squares with
# singular-value cutoff eps * max(M, N) * s_max (eps = .Machine$double.eps).
# Theta weighting uses the row MEAN distance (not Simplex 1e-6 floor).
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#' S-map projection (top-level API).
#'
#' @inheritParams Simplex
#' @param theta local weighting parameter (0 = global linear map).
#' @return list(predictions, coefficients, singularValues) data.frames.
#' @keywords internal
#------------------------------------------------------------------------
SMap <- function(dataFrame = NULL, columns, target, lib, pred,
                 E = NULL, Tp = 1, knn = 0, tau = -1, theta = 0,
                 exclusionRadius = 0, embedded = FALSE, validLib = logical(0),
                 noTime = FALSE, ignoreNan = TRUE, backend = "RANN",
                 pathIn = "./", dataFile = "", pathOut = "./", predictFile = "",
                 parameterList = FALSE, includeState = FALSE,
                 verbose = FALSE, showPlot = FALSE) {

  parameters <- as.list(environment())
  parameters[c("dataFrame", "includeState", "pathIn", "dataFile",
               "pathOut", "predictFile", "parameterList", "showPlot",
               "verbose", "backend")] <- NULL

  dataFrame <- ResolveInput(pathIn, dataFile, dataFrame)
  columns   <- SplitColumns(columns)
  target    <- SplitColumns(target)
  nRows     <- nrow(dataFrame)

  E      <- RequireE("SMap", E, embedded, columns) # required unless embedded
  idx    <- CreateIndices(lib, pred, E, tau, Tp, embedded, nRows, "SMap")
  lib_i  <- idx$lib_i
  pred_i <- idx$pred_i

  originalLibLen <- length(lib_i)
  if (knn < 1) { knn <- originalLibLen - 1L } # SMap default: full library
  originalKnn <- knn

  targetVec <- dataFrame[[target[1]]]
  timeVec   <- if (noTime) seq_len(nRows) else dataFrame[[1]]

  embedding <- if (embedded) {
                 as.matrix(dataFrame[, columns, drop = FALSE])
               }
               else {
                 MakeBlock(dataFrame, E, tau, columns)
               }

  # Drop rows with NaN in the embedding from the lib and pred index sets.
  rn <- RemoveNan(embedding, lib_i, pred_i, ignoreNan)

  # S-map's default knn is the full library (originalLibLen - 1). If NaN
  # removal shrank the library, that default is now too large, adjust
  # to the reduced library. A user-specified knn is left untouched.
  libraryShrank     <- length(rn$lib_i) < originalLibLen
  knnWasFullLibrary <- originalKnn == (originalLibLen - 1L)

  if (isTRUE(ignoreNan) && libraryShrank && knnWasFullLibrary) {
    # re-apply "whole library" to the smaller library
    knn <- length(rn$lib_i) - 1L
  }

  lib_i  <- rn$lib_i
  pred_i <- rn$pred_i

  nb <- FindNeighbors(embedding, lib_i, pred_i, knn,
                      exclusionRadius = exclusionRadius, validLib = validLib,
                      libOverlap = idx$libOverlap, backend = backend,
                      tieBreak = FALSE, verbose = verbose) # SMap: no tie-break

  # Flag if nan are present in targetVec for solver filtering
  targetVecNan <- any(is.na(targetVec))

  pr <- SMapProject(nb$neighbors, nb$distances, embedding, targetVec,
                    pred_i, E, Tp, theta, knn, targetVecNan)

  timeName = "Time"
  if (!noTime && !is.null(names(dataFrame))) {
    timeName <- names(dataFrame)[1]
  }

  predictions <- FormatProjection(idx$predList, pred_i, idx$pred_i_all,
                                  targetVec, timeVec, pr$projection,
                                  pr$variance, Tp, nRows, tau, timeName)

  # Coefficients & singular-value frames share the projection layout
  L <- ProjectionLayout(idx$predList, pred_i, idx$pred_i_all,
                        timeVec, Tp, nRows, tau)
  nDim     <- E + 1
  embNames <- EmbedColumnNames(if (embedded) columns
                               else columns, E, tau, embedded)
  coefNames <- paste0("\u2202", target[1], "/\u2202", embNames)

  coefOut <- matrix(NA_real_, L$outSize, nDim)
  svOut   <- matrix(NA_real_, L$outSize, nDim)
  coefOut[L$predOut_i + 1L, ] <- pr$coefficients
  svOut[L$predOut_i + 1L, ]   <- pr$singularValues

  coefDf <- data.frame(Time = L$timeOut, coefOut, check.names = FALSE,
                       stringsAsFactors = FALSE)
  names(coefDf) <- c("Time", "C0", coefNames)

  svDf <- data.frame(Time = L$timeOut, svOut, check.names = FALSE,
                     stringsAsFactors = FALSE)
  names(svDf) <- c("Time", paste0("C", 0:(nDim - 1)))

  if (showPlot) {
    PlotObsPred(predictions, "", E, Tp)
  }

  result <- list(predictions = predictions, coefficients = coefDf,
                 singularValues = svDf)

  if (showPlot) {
    PlotSmap(result, "", E, Tp)
  }

  internal <- if (isTRUE(includeState)) {
                  list(knn_neighbors = nb$neighbors,
                       knn_distances = nb$distances,
                       lib_i         = lib_i,
                       pred_i        = pred_i,
                       targetVec     = targetVec,
                       embedding     = embedding)
              }
              else { NULL }
    
    result = EdmFinalize(result, result $ predictions,
                         pathOut, predictFile, parameterList, parameters,
                         includeState = includeState, internal = internal)
}

#------------------------------------------------------------------------
#' Minimum-norm least squares via truncated SVD.
#'
#' @param A design matrix (m x n).
#' @param b response vector (length m).
#' @return list(C = coefficients length n, SV = singular values length min(m,n)).
#' @details matches numpy.linalg.lstsq( rcond=(eps * max(m, n)) )
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
#
# Reference is pyEDM's numpy.linalg.lstsq, a wrapper over LAPACK gelsd() 
# (SVD-based, singular-value truncation, minimum-norm solution). R has
# no base equivalent that both solves and returns singular values, so we
#   First: compute SVD with svd() -- a LAPACK dgesdd call -- and
#   Next:  reconstruct the least-squares solution by hand.
# Steps after svd() below are the extra work gelsd does internally.
#
# SMap is deliberately ill-conditioned at large theta, where SVD's
# minimum-norm truncation is what matches gelsd/pyEDM. The cutoff
# reproduces numpy.linalg.lstsq's default rcond, which is what makes the
# coefficient and singularValue fixtures agree bit-for-bit. Do not swap the
# solver or drop the cutoff without re-baselining those fixtures.
#
# Returns: C  = fit coefficients (pyEDM 'coefficients')
#          SV = singular values of A (pyEDM 'singularValues'), raw from LAPACK
#------------------------------------------------------------------------
SMapSolve <- function(A, b) {
  # (1) LAPACK dgesdd: A = u diag(d) t(v). Decomposition only, no solve.
  s <- svd(A)
  d <- s$d    # singular values, descending

  # (2) rcond-style threshold ~ numpy lstsq default
  cutoff <- .Machine$double.eps * max(nrow(A), ncol(A)) * d[1]

  # (3) truncated reciprocal: zero below cutoff -> minimum-norm, rank-safe
  dInv <- ifelse(d > cutoff, 1 / d, 0)

  # (4) reconstruct x = V diag(dInv) t(U) b  (the solve gelsd does internally)
  C <- s$v %*% (dInv * crossprod(s$u, b))

  list(C = as.numeric(C), SV = d)
}

#------------------------------------------------------------------------
#' S-map projection kernel (SMap.py::Project).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
SMapProject <- function(knnNeighbors, knnDistances, embedding, targetVec,
                        pred_i, E, Tp, theta, knn, targetVecNan) {
  nPred <- length(pred_i)
  nDim  <- E + 1

  projection     <- rep(NA_real_, nPred)
  variance       <- rep(NA_real_, nPred)
  coefficients   <- matrix(NA_real_, nPred, nDim)
  singularValues <- matrix(NA_real_, nPred, nDim)

  # Mean neighbour distance per row over FINITE (valid) neighbours only;
  # Inf padding slots are excluded. Rows with no finite neighbour give NaN,
  # which flags an empty (skipped) row below (issue #74).
  finiteD     <- is.finite(knnDistances)
  finiteCnt   <- rowSums(finiteD)
  distSum     <- rowSums(ifelse(finiteD, knnDistances, 0))
  distRowMean <- ifelse(finiteCnt > 0, distSum / finiteCnt, NaN)
  if (theta == 0) {
    W <- matrix(1, nrow(knnDistances), ncol(knnDistances))
  }
  else {
    W <- exp(-(theta / distRowMean) * knnDistances)   # row-scaled
  }

  neighborsTp <- knnNeighbors + Tp
  n <- length(targetVec)
  B <- matrix(NA_real_, nPred, ncol(knnNeighbors))
  inRange <- neighborsTp >= 1 & neighborsTp <= n
  B[inRange] <- targetVec[neighborsTp[inRange]]

  wB <- W * B

  for (row in seq_len(nPred)) {
    # Empty row : no finite-distance neighbour survived exclusion / validLib.
    # Leave projection / coefficients / variance at their NA defaults.
    if (is.nan(distRowMean[row])) { next }

    # Valid slots : finite distance (exclude Inf padding) and, when the
    # target has NA, finite target. Gating on distance finiteness (not the
    # weight) makes padding inert for all theta, so theta == 0 needs no
    # special case (issue #74).
    valid <- is.finite(knnDistances[row, ])
    if (targetVecNan) { valid <- valid & is.finite(B[row, ]) }
    if (!any(valid))  { next }

    # Build A over valid slots only : knnNeighbors padding carries the 0
    # sentinel index, which must never index the embedding.
    libRows <- knnNeighbors[row, valid]
    m       <- sum(valid)
    A       <- matrix(NA_real_, m, nDim)
    A[, 1]  <- W[row, valid]
      
    for (j in 2:nDim) {
      A[, j] <- W[row, valid] * embedding[libRows, j - 1]
    }
    wBrow <- wB[row, valid]

    sol <- SMapSolve(A, wBrow)
    C   <- sol$C
    coefficients[row, ]   <- C
    singularValues[row, ] <- c(sol$SV, rep(NA_real_, nDim - length(sol$SV)))

    proj <- if (is.na(C[1])) { 0 }
            else             { C[1] }

    for (e in 2:nDim) {
      proj <- proj + C[e] * embedding[pred_i[row], e - 1]
    }

    projection[row] <- proj

    wRow          <- W[row, valid]
    deltaSqr      <- (B[row, valid] - proj)^2
    variance[row] <- sum(wRow * deltaSqr) / sum(wRow)
  }

  list(projection = projection, variance = variance,
       coefficients = coefficients, singularValues = singularValues)
}

#------------------------------------------------------------------------
# Simplex projection
#
# Port of pyEDM Simplex (Simplex.py, EDM.py, Formatting.py, API.Embed).
# See spec: rEDM_Simplex_spec.md.
#
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#' Simplex projection (top-level API).
#'
#' @param dataFrame data.frame; first column is time unless noTime = TRUE.
#' @param columns,target column name(s) (vector or space string).
#' @param lib,pred (start end) index pairs (vector or space string).
#' @param E,Tp,knn,tau,exclusionRadius standard EDM parameters.
#' @param embedded TRUE if columns already form the embedding.
#' @param validLib logical mask of admissible library rows (or length 0).
#' @param noTime TRUE to synthesise a 1..N time index.
#' @param ignoreNan remove embedding-NaN rows from lib/pred.
#' @param backend "RANN" or "brute".
#' @param pathIn file path for input dataFile.
#' @param dataFile input dataFile, .csv format.
#' @param pathOut file path for predictFile.
#' @param predictFile output file name.
#' @param parameterList append list of parameters to return.
#' @return data.frame(Time, Observations, Predictions, Pred_Variance).
#' @keywords internal
#------------------------------------------------------------------------
Simplex <- function(dataFrame = NULL, columns, target, lib, pred,
                    E = NULL, Tp = 1, knn = 0, tau = -1, exclusionRadius = 0,
                    embedded = FALSE, validLib = logical(0), noTime = FALSE,
                    ignoreNan = TRUE, backend = "RANN", pathIn = "./",
                    dataFile = "", pathOut = "./", predictFile = "",
                    parameterList = FALSE, includeState = FALSE,
                    verbose = FALSE, showPlot = FALSE,
                    .tieBreak = TRUE) {

  parameters <- as.list(environment())
  parameters[c("dataFrame", "includeState", "pathIn", "dataFile",
               "pathOut", "predictFile", "parameterList", "showPlot",
               "verbose", "backend")] <- NULL

  dataFrame <- ResolveInput(pathIn, dataFile, dataFrame)
  columns   <- SplitColumns(columns)
  target    <- SplitColumns(target)
  RequireValidData("Simplex", dataFrame, columns, target)
  nRows     <- nrow(dataFrame)


  E <- RequireE("Simplex", E, embedded, columns)  # required unless embedded
  if (knn < 1) { knn <- E + 1 }   # Simplex default

  idx <- CreateIndices(lib, pred, E, tau, Tp, embedded, nRows, "Simplex")

  targetVec <- dataFrame[[target[1]]]
  timeVec   <- if (noTime) { seq_len(nRows) }
               else        { dataFrame[[1]] }

  embedding <- if (embedded) {
                 as.matrix(dataFrame[, columns, drop = FALSE])
               }
               else {
                 MakeBlock(dataFrame, E, tau, columns)
               }

  rn     <- RemoveNan(embedding, idx$lib_i, idx$pred_i, ignoreNan)
  lib_i  <- rn$lib_i
  pred_i <- rn$pred_i

  nb <- FindNeighbors(embedding, lib_i, pred_i, knn,
                      exclusionRadius = exclusionRadius, validLib = validLib,
                      libOverlap = idx$libOverlap, backend = backend,
                      tieBreak = .tieBreak, verbose = verbose)

  pr <- SimplexProject(nb$neighbors, nb$distances, targetVec, Tp)

  timeName = "Time"
  if (!noTime && !is.null(names(dataFrame))) {
    timeName <- names(dataFrame)[1]
  }

  out <- FormatProjection(idx$predList, pred_i, idx$pred_i_all, targetVec,
                          timeVec, pr$projection, pr$variance, Tp, nRows,
                          tau, timeName)

  if (showPlot) { PlotObsPred(out, "", E, Tp) }

  internal <- if (isTRUE(includeState)) {
                  list(knn_neighbors = nb$neighbors,
                       knn_distances = nb$distances,
                       lib_i = lib_i, pred_i = pred_i,
                       targetVec = targetVec,
                       embedding = embedding) }
              else { NULL }

  out = EdmFinalize(out, out, pathOut, predictFile, parameterList, parameters,
                    includeState = includeState, internal = internal)
}

#------------------------------------------------------------------------
#' Simplex projection kernel (Simplex.py::Project).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
SimplexProject <- function(knnNeighbors, knnDistances, targetVec, Tp) {
  # Finite (valid) slots : Inf-distance slots are exclusion padding (issue #74)
  finiteD <- is.finite(knnDistances)

  # Minimum distance per row over FINITE neighbours only. Empty rows (no
  # finite neighbour) get a placeholder and resolve to NA below.
  dForMin      <- ifelse(finiteD, knnDistances, Inf)
  minDistances <- apply(dForMin, 1L, min)
  empty        <- !is.finite(minDistances)
  minDistances <- pmax(ifelse(empty, 1, minDistances), 1e-6)

  scaledDistances   <- knnDistances / minDistances
  weights           <- exp(-scaledDistances)
  weights[!finiteD] <- 0            # padding slots carry zero weight
  weightRowSum      <- rowSums(weights)

  neighborsTp <- knnNeighbors + Tp
  n <- length(targetVec)
  libTargets  <- matrix(NA_real_, nrow(knnNeighbors), ncol(knnNeighbors))
  inRange     <- neighborsTp >= 1 & neighborsTp <= n & knnNeighbors >= 1
  libTargets[inRange]  <- targetVec[neighborsTp[inRange]]
  libTargets[!finiteD] <- 0         # zero padding targets (avoid 0 * NA)

  # Guard division for rows whose neighbours are all excluded
  wRowSumSafe <- ifelse(weightRowSum > 0, weightRowSum, 1)
  projection  <- rowSums(weights * libTargets) / wRowSumSafe
  delta       <- libTargets - projection
  variance    <- rowSums(weights * delta^2) / wRowSumSafe

  # Empty rows (no valid neighbour) -> NA projection & variance
  projection[empty] <- NA_real_
  variance[empty]   <- NA_real_
  list(projection = projection, variance = variance)
}

#------------------------------------------------------------------------
# Legacy rEDM v1 compatibility: file-based I/O and parameterList.
#
# Self-contained helpers used by the public API functions. Adding the file
# parameters (pathIn, dataFile, pathOut, predictFile) and parameterList at
# their defaults leaves every function's numeric behaviour unchanged.
#
# Naming: functions/constants uppercase-first, variables lowercase-first.
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#' Resolve the input data.frame from a file or an in-memory frame.
#'
#' If \code{dataFile} is non-empty, read \code{file.path(pathIn, dataFile)}
#' (with \code{check.names = FALSE} so headers such as \code{x_t-1} or
#' \code{Var 5 1} survive); this takes precedence over \code{dataFrame}.
#' Otherwise return \code{dataFrame} unchanged.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
ResolveInput <- function(pathIn, dataFile, dataFrame) {
  if (nzchar(dataFile))
    return(utils::read.csv(file.path(pathIn, dataFile),
                           check.names = FALSE, stringsAsFactors = FALSE))
  dataFrame
}

#------------------------------------------------------------------------
#' Write the primary output frame and shape the return value.
#'
#' Optionally writes \code{primaryFrame} to \code{file.path(pathOut,
#' predictFile)}, then applies the \code{parameterList} contract:
#'   - frame result  -> list(predictions = result, parameters = parameters)
#'   - list result   -> result with a $parameters element appended
#' When \code{parameterList} is FALSE the result is returned unchanged.
#'
#' @param result the function's normal return object (data.frame or list).
#' @param primaryFrame the frame written to disk (result itself, or the
#'   predictions element of a list result).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
EdmFinalize <- function(result, primaryFrame, pathOut, predictFile,
                        parameterList, parameters,
                        includeState = FALSE, internal = NULL) {
  if (nzchar(predictFile))
    utils::write.csv(primaryFrame, file.path(pathOut, predictFile),
                     row.names = FALSE)
  if (isTRUE(parameterList) || isTRUE(includeState)) {
    if (is.data.frame(result))
      result <- list(predictions = result)
    if (isTRUE(parameterList)) result$parameters <- parameters
    if (isTRUE(includeState))  result$internal   <- internal
    return(result)
  }
  result
}

#------------------------------------------------------------------------
#' Resolve / require the embedding dimension E.
#'
#' pyEDM makes E a required argument for Simplex, SMap, PredictInterval and
#' PredictNonlinear. When embedded, E is inferred as length(columns) and any
#' user value is ignored; otherwise E must be supplied (non-NULL, >= 1).
#' @keywords internal
#------------------------------------------------------------------------
RequireE <- function(fn, E, embedded, columns) {
  if (isTRUE(embedded)) return(length(columns))
  if (is.null(E) || length(E) == 0L || E < 1L)
    stop(sprintf("%s(): E is required (integer >= 1).", fn), call. = FALSE)
  E
}

#------------------------------------------------------------------------
# Plotting helpers for showPlot = TRUE (base graphics).
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#' Observed vs predicted time series.
#' @param df data.frame with Time, Observations, Predictions.
#' @param dataName optional label used in the title.
#' @param E,Tp embedding dimension and forecast interval, shown in the title.
#' @return Invisibly \code{NULL}.
#' @keywords internal
#------------------------------------------------------------------------
PlotObsPred <- function(df, dataName = "", E = NULL, Tp = NULL) {
  err   <- ComputeError(df$Observations, df$Predictions)
  title <- sprintf("%s E=%d Tp=%d  rho=%.2f  RMSE=%.2f",
                   dataName, E, Tp,
                   ifelse(is.na(err$rho),  NA_real_, err$rho ),
                   ifelse(is.na(err$RMSE), NA_real_, err$RMSE))
  time <- PlotTimeAxis(df[[1]], nrow(df))
  yl <- range(c(df$Observations, df$Predictions), na.rm = TRUE)

  plot(time, df$Observations, type = "l", lwd = 3, ylim = yl,
       col = "blue", xlab = names(df)[1], ylab = "", main = title,
       cex.axis = 1.3, cex.lab = 1.3)

  graphics::lines(time, df$Predictions, col = "red", lwd = 3)
  graphics::legend("topright", legend = c("Observations", "Predictions"),
                   fill = c("blue", "red"), bty = "n", cex = 1.2)
  invisible(NULL)
}

#------------------------------------------------------------------------
#' S-map observed/predicted plus one panel per coefficient.
#' @param SmapList list returned by \code{SMap} (predictions, coefficients).
#' @param dataName optional label used in the title.
#' @param E,Tp embedding dimension and forecast interval, shown in the title.
#' @return Invisibly \code{NULL}.
#' @keywords internal
#------------------------------------------------------------------------
PlotSmap <- function(SmapList, dataName = "", E = NULL, Tp = NULL) {
  if (!all(c("predictions", "coefficients") %in% names(SmapList))) {
    message("PlotSmap(): need predictions and coefficients.")
    return(invisible(NULL))
  }
  p        <- SmapList$predictions
  co       <- SmapList$coefficients
  time     <- PlotTimeAxis(p[[1]], nrow(p))
  numCoeff <- ncol(co) - 1L
  oldPar   <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(oldPar), add = TRUE)
  
  graphics::par(mfrow = c(numCoeff + 1L, 1L), mar = c(3.5, 4, 0.5, 1),
                mgp = c( 1.5, 0.5, 0 ), cex.axis = 1.3, cex.lab = 1.3)
  yl <- range(c(p$Observations, p$Predictions), na.rm = TRUE)
  
  # Observations & Predictions
  title = sprintf("S-Map %s E=%d Tp=%d", dataName, E, Tp)
  
  plot(time, p$Observations, type = "l", ylim = yl, xlab = names(p)[1],
       ylab = "", col = "blue", lwd = 3, cex.axis = 1.3, cex.lab = 1.3,
       main = title )
  
  graphics::lines(time, p$Predictions, col = "red", lwd = 3)

  graphics::legend("topright", legend = c("Observations", "Predictions"),
                   fill = c("blue", "red"), bty = "n", cex = 1.5)

  # Coefficients
  coefName <- names(co)
  title = paste( dataName, 'S-Map Coefficients', '\nE=' , E, ' Tp=', Tp )
  for (j in 2:ncol(co)) {
    # coefName[j] is "∂<target>/∂<emb>"; recover the two operands
    ops <- strsplit(sub("\u2202", "", coefName[j], fixed = TRUE),
                    "/\u2202", fixed = TRUE)[[1]]
    # replace the explicit ∂ with plotmath partialdiff for any device
    ylab = bquote(paste(partialdiff, .(ops[1]), "/", partialdiff, .(ops[2])))
    plot(time, co[,j], type = "l", lwd = 3, col = "blue",
         xlab = names(p)[1], ylab = ylab)
    graphics::mtext( title, side = 3, line = -1.5, cex = 1.2 )
  }

  invisible(NULL)
}

#------------------------------------------------------------------------
#' rho versus a swept parameter (EmbedDimension / PredictInterval / ...).
#' @keywords internal
#------------------------------------------------------------------------
PlotSweep <- function(df, xlab, title = NULL) {
  plot(df[[1]], df$rho, type = "l", lwd = 3, main = title,
       xlab = xlab, ylab = expression("Prediction Skill (" * rho * ")"))

  invisible(NULL)
}

#------------------------------------------------------------------------
#' CCM cross-map skill versus library size, both directions.
#' @param df CCM result data.frame (LibSize plus two rho columns).
#' @param E embedding dimension, shown in the title.
#' @return Invisibly \code{NULL}.
#' @keywords internal
#------------------------------------------------------------------------
PlotCCM <- function(df, E = NULL) {
  yl <- range(c(df[[2]], df[[3]]), na.rm = TRUE)

  title = if (!is.null(E)) paste0(names(df)[2], " : ", names(df)[3], "  E=", E) else ""
  plot(df$LibSize, df[[2]], type = "l", col = "blue", ylim = yl, lwd = 3, 
       main = title,
       xlab = "Library Size",
       ylab = expression("Cross Map Skill (" * rho * ")"))
  
  graphics::lines(df$LibSize, df[[3]], col = "red", lwd = 3)
  graphics::abline(h = 0)
  graphics::legend("topright", legend = names(df)[2:3],
                   fill = c("blue", "red"), bty = "n", cex = 1.2)
  invisible(NULL)
}

#------------------------------------------------------------------------
#' Coerce a first column into a plottable time axis (numeric, Date, or index).
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
PlotTimeAxis <- function(col, n) {
  if (is.numeric(col)) return(col)
  t <- try(as.Date(col), silent = TRUE)
  if (inherits(t, "try-error") || all(is.na(t)))
    t <- try(as.POSIXct(col), silent = TRUE)
  if (inherits(t, "try-error") || all(is.na(t))) t <- seq_len(n)
  t
}

#------------------------------------------------------------------------
# Diagnostic parameter sweeps.
#
# Port of pyEDM EmbedDimension / PredictInterval / PredictExclusionRadius /
# PredictNonlinear (API.py + PoolFunc.py). Each sweeps one parameter over a
# list, runs Simplex (or SMap for PredictNonlinear) at each value, and
# collects ComputeError()$rho.
#
# The sweeps are RNG-free and order-preserving, so parallel and sequential
# evaluation give identical results. Column naming follows the rEDM
# convention (E, Tp, ExclusionRadius, Theta); pyEDM uses lower-case theta /
# 'Exclusion radius', which the validation harness reconciles.
#
# Naming: functions/constants uppercase-first, variables lowercase-first.
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#' EmbedDimension : prediction skill vs embedding dimension E in 1..maxE.
#'
#' @inheritParams Simplex
#' @param maxE largest embedding dimension to evaluate.
#' @param numProcess worker processes for the sweep.
#' @return data.frame(E, rho).
#' @keywords internal
#------------------------------------------------------------------------
EmbedDimension <- function(dataFrame = NULL, columns, target, lib, pred,
                           maxE = 10, Tp = 1, tau = -1, exclusionRadius = 0,
                           embedded = FALSE, validLib = logical(0),
                           noTime = FALSE, ignoreNan = TRUE, numProcess = 4,
                           pathIn = "./", dataFile = "", pathOut = "./",
                           predictFile = "", backend = "RANN",
                           parameterList = FALSE, showPlot = FALSE) {

  parameters <- as.list(environment())
  parameters[c("dataFrame")] <- NULL

  dataFrame <- ResolveInput(pathIn, dataFile, dataFrame)
  Evals     <- seq_len(maxE)
  worker <- function(E) SweepRho(
    Simplex(dataFrame, columns, target, lib, pred, E = E, Tp = Tp, tau = tau,
            exclusionRadius = exclusionRadius, embedded = embedded,
            validLib = validLib, noTime = noTime, ignoreNan = ignoreNan,
            backend = backend))

  rho <- unlist(SweepApply(Evals, worker, numProcess))
  out <- data.frame(E = Evals, rho = rho)

  if (showPlot) {
    title = paste( columns, ":", target, "\nTp=", Tp)
    PlotSweep(out, "E", title)
  }
  out = EdmFinalize( out, out, pathOut, predictFile, parameterList, parameters )
}

#------------------------------------------------------------------------
#' PredictInterval : prediction skill vs forecast interval Tp in 1..maxTp.
#'
#' @inheritParams Simplex
#' @param maxTp largest forecast interval to evaluate.
#' @return data.frame(Tp, rho).
#' @keywords internal
#------------------------------------------------------------------------
PredictInterval <- function(dataFrame = NULL, columns, target, lib, pred,
                            maxTp = 10, E = NULL, tau = -1, exclusionRadius = 0,
                            embedded = FALSE, validLib = logical(0),
                            noTime = FALSE, ignoreNan = TRUE, numProcess = 4,
                            backend = "RANN", pathIn = "./", dataFile = "",
                            pathOut = "./", predictFile = "", 
                            parameterList = FALSE, showPlot = FALSE) {

  parameters <- as.list(environment())
  parameters[c("dataFrame")] <- NULL

  dataFrame <- ResolveInput(pathIn, dataFile, dataFrame)
  E         <- RequireE("PredictInterval", E, embedded, SplitColumns(columns))
  Evals     <- seq_len(maxTp)

  worker <- function(Tp) SweepRho(
    Simplex(dataFrame, columns, target, lib, pred, E = E, Tp = Tp, tau = tau,
            exclusionRadius = exclusionRadius, embedded = embedded,
            validLib = validLib, noTime = noTime, ignoreNan = ignoreNan,
            backend = backend))

  rho <- unlist(SweepApply(Evals, worker, numProcess))
  out <- data.frame(Tp = Evals, rho = rho)

  if (showPlot) {
    title = paste( columns, ":", target, "\nE=", E) 
    PlotSweep(out, "Tp", title)
  }

  out = EdmFinalize( out, out, pathOut, predictFile, parameterList, parameters )
}

#------------------------------------------------------------------------
#' PredictExclusionRadius : prediction skill vs Theiler exclusion radius.
#'
#' @inheritParams Simplex
#' @param exclusionRadius numeric vector of radii (default pyEDM sweep list).
#' @return data.frame(ExclusionRadius, rho).
#' @keywords internal
#------------------------------------------------------------------------
PredictExclusionRadius <- function(dataFrame = NULL, columns, target, lib, pred,
                                   exclusionRadius = NULL, E = 1, Tp = 1,
                                   tau = -1, embedded = FALSE,
                                   validLib = logical(0), noTime = FALSE,
                                   ignoreNan = TRUE, numProcess = 4,
                                   backend = "RANN", pathIn = "./", 
                                   dataFile = "", pathOut = "./",
                                   predictFile = "", 
                                   parameterList = FALSE, showPlot = FALSE) {

  parameters <- as.list(environment())
  parameters[c("dataFrame")] <- NULL

  dataFrame <- ResolveInput(pathIn, dataFile, dataFrame)
  if (is.null(exclusionRadius)) {
    exclusionRadius <- c(0,1,2,3,4,5,6,7,8,9,10,12,14,16,18,20,25,30)
  }
  worker <- function(er) SweepRho(
    Simplex(dataFrame, columns, target, lib, pred, E = E, Tp = Tp, tau = tau,
            exclusionRadius = er, embedded = embedded, validLib = validLib,
            noTime = noTime, ignoreNan = ignoreNan, backend = backend))

  rho <- unlist(SweepApply(exclusionRadius, worker, numProcess))
  out <- data.frame(ExclusionRadius = exclusionRadius, rho = rho)
    
  if (showPlot) {
      title = paste( columns, ":", target, "\nE=", E, ' Tp=', Tp)
      PlotSweep(out, "Exclusion Radius", title)
  }
  out = EdmFinalize( out, out, pathOut, predictFile, parameterList, parameters )
}

#------------------------------------------------------------------------
#' PredictNonlinear : S-map prediction skill vs localisation theta.
#'
#' @inheritParams SMap
#' @param theta numeric vector of theta values (default pyEDM sweep list).
#' @return data.frame(Theta, rho).
#' @keywords internal
#------------------------------------------------------------------------
PredictNonlinear <- function(dataFrame = NULL, columns, target, lib, pred,
                             theta = NULL, E = NULL, Tp = 1, knn = 0, tau = -1,
                             exclusionRadius = 0, embedded = FALSE,
                             validLib = logical(0), noTime = FALSE,
                             ignoreNan = TRUE, numProcess = 4,
                             backend = "RANN", pathIn = "./", dataFile = "",
                             pathOut = "./", predictFile = "",
                             parameterList = FALSE, showPlot = FALSE,
                             verbose = FALSE) {

  parameters <- as.list(environment())
  parameters[c("dataFrame")] <- NULL

  dataFrame <- ResolveInput(pathIn, dataFile, dataFrame)
  E         <- RequireE("PredictNonlinear", E, embedded, SplitColumns(columns))
    
  if (is.null(theta)) {
    theta <- c(0.01, 0.1, 0.3, 0.5, 0.75, 1, 1.5, 2, 3, 4, 5, 6, 7, 8, 9)
  }

  worker <- function(th) SweepRho(
    SMap(dataFrame, columns, target, lib, pred, E = E, Tp = Tp, knn = knn,
         tau = tau, theta = th, exclusionRadius = exclusionRadius,
         embedded = embedded, validLib = validLib, noTime = noTime,
         ignoreNan = ignoreNan, backend = backend)$predictions)

  rho <- unlist(SweepApply(theta, worker, numProcess))
  out <- data.frame(Theta = theta, rho = rho)

  if (showPlot) {
    title = paste( columns, ":", target, "\nE=", E, ' Tp=', Tp) 
    PlotSweep(out, expression(paste("S-map Localization",~(theta))), title)
  }
  out = EdmFinalize( out, out, pathOut, predictFile, parameterList, parameters )
}

#------------------------------------------------------------------------
#' Map a worker over a sweep list, in parallel where supported.
#'
#' Uses forked parallel::mclapply on unix when numProcess > 1; falls back to
#' sequential lapply otherwise. Results are deterministic either way.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
SweepApply <- function(values, worker, numProcess = 4) {
  useParallel <- numProcess>1 && .Platform$OS.type == "unix" && length(values)>1
  if (useParallel) {
    # Respect the R CMD check core limit (_R_CHECK_LIMIT_CORES_ caps at 2);
    # otherwise use the machine core count.
    limited  <- isTRUE(as.logical(Sys.getenv("_R_CHECK_LIMIT_CORES_", "false")))
    maxCores <- if (limited) 2L else parallel::detectCores()
      
    if (is.na(maxCores) || maxCores < 1L) { maxCores <- 1L }

    cores <- max(1L, min(numProcess, length(values), maxCores))

    if (cores > 1L) {
      out <- tryCatch(
        parallel::mclapply(values, worker, mc.cores = cores,
                           mc.preschedule = TRUE),
        error = function(e) NULL)
      
      if (!is.null(out) &&
          !any(vapply(out, function(x) inherits(x, "try-error"), logical(1)))) {
        return(out)
      }
    }
  }
  lapply(values, worker) # sequential fallback
}

#------------------------------------------------------------------------
#' rho from a Simplex/SMap prediction frame.
#' @keywords internal
#' @noRd
#------------------------------------------------------------------------
SweepRho <- function(predictions) {
  ComputeError(predictions$Observations, predictions$Predictions)$rho
}

---
title: "Empirical Dynamic Modeling"
author:
- George Sugihara
- Ethan Deyle
- Joseph Park
- Erik Saberski
- Cameron Smith
- Hao Ye
date: "`r Sys.Date()`"
output: html_document
#output:
#  pdf_document:
#    latex_engine: xelatex
#  html_document:
#    rmarkdown::html_vignette
vignette: >
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteIndexEntry{"Empirical Dynamic Modeling"}
    %\VignetteEncoding{UTF-8}
bibliography: References.bib
header-includes: 
- \usepackage{graphicx}
- \usepackage{float}
- \usepackage[utf8]{inputenc}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(echo = TRUE, warning = FALSE, tidy = TRUE,
                      fig.pos = 'h', fig.align = 'center',
                      fig.width = 4, fig.height = 3.3 )
knitr::opts_knit$set(global.par = TRUE, progress = FALSE)
options(digits = 2)
par.orig = par()
par(mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))
```

<!------------------------------------------------------------------------->
# Abstract
<!------------------------------------------------------------------------->

Empirical dynamic modeling (EDM) is an emerging non-parametric framework for modeling nonlinear dynamic systems. EDM is based on the mathematical theory of reconstructing attractor manifolds from time series data [@Takens_1981]. The **rEDM** package collects several EDM methods, including simplex projection [@Sugihara_1990], S-map [@Sugihara_1994], multivariate embeddings [@Dixon_1999], convergent cross mapping [@Sugihara_2012], and multiview embedding [@Ye_2016]. Here, we introduce the basic underlying theory, and describe the functionality of **rEDM** using examples from both model simulations and real data.

<!------------------------------------------------------------------------->
# Installation
<!------------------------------------------------------------------------->

The **rEDM** package can be obtained in two main ways. The standard version of the package can be obtained through CRAN (the Comprehensive R Archive Network): https://cran.r-project.org/package=rEDM:

```{r CRAN installation instructions, eval = FALSE}
install.packages("rEDM")
```

Also available on GitHub at [SugiharaLab](https://github.com/SugiharaLab/rEDM), and can be installed using R **devtools**.

```{r GitHub installation instructions, eval = FALSE}
devtools::install_github("SugiharaLab/rEDM")
```

<!------------------------------------------------------------------------->
# Introduction
<!------------------------------------------------------------------------->

Many scientific fields use models as approximations of reality and for various purposes, for example, testing hypotheses regarding mechanisms or processes, explaining past observations, and predicting future outcomes. In many cases these models are based on hypothesized parametric equations; however, explicit equations can be impractical when the underlying mechanisms are unknown or are too complex to be characterized with existing datasets. Empirical models, which infer patterns and associations from the data (instead of using hypothesized equations), represent an alternative and highly flexible approach. Here, we review the theoretical background for empirical dynamic modeling (EDM) and the functionality of the **rEDM** package, which are intended for nonlinear dynamic systems that can prove problematic for traditional modeling approaches.

The basic goal underlying EDM is to reconstruct the behavior of dynamic systems using time series data. This approach is based on mathematical theory developed initially by [@Takens_1981], and expanded by others [@Sauer_1991; @Casdagli_1991; @Deyle_2011]. Because these methods operate with minimal assumptions, they are particularly suitable for studying systems that exhibit non-equilibrium dynamics and nonlinear state-dependent behavior (i.e. where interactions change over time and as a function of the system state).

<!------------------------------------------------------------------------->
# Empirical Dynamic Modeling
<!------------------------------------------------------------------------->

<!------------------------------------------------------------------------->
## Time Series as Observations of a Dynamic System
<!------------------------------------------------------------------------->

The essential concept is that time series can be viewed as projections of the behavior of a dynamic system. First, the system state can be described as a point in a high-dimensional space. The axes of this space can be thought of as fundamental state variables; in an ecosystem, these variables might correspond to population abundances, resources, or environmental conditions. Second, the system state changes through time following a set of deterministic rules. In other words, the behavior of the system is not completely stochastic.

Consequently, it is possible to project the system state onto one of the coordinate axes and obtain the value of the corresponding state variable. Sequential projections over time will thus produce a time series for that variable. For example, in figure 1 the states of the canonical Lorenz Attractor [@Lorenz_1963] are projected onto the $x$-axis, creating a time series of variable $x$.

<!------------
   Figure 1 
------------->
```{r, LorenzProjection, echo = FALSE, fig.cap = "Time Series Projection from the Lorenz Attractor.", out.width = "50%", fig.align = 'center'}
knitr::include_graphics("Lorenz_Projection.png")
```

Although different time series observed from a system can represent independent state variables, in general, each time series is an *observation function* of the system state that may convolve several different state variables.

<!------------------------------------------------------------------------->
## Attractor Reconstruction / Takens' Theorem
<!------------------------------------------------------------------------->

The goal of EDM is to reconstruct the system dynamics from time series data. As seen above, a time series can be thought of as sequential projections of the motion on an attractor; in other words, information about the behavior is encoded in the temporal ordering of the time series. Takens' Theorem [@Takens_1981] states that mathematically valid and property preserving reconstructions of the attractor can be created using lags of a single time series, then substituting those lagged time series for unknown or unobserved variables. In other words, instead of representing the system state using a complete set of state variables, we can instead use an `E`-dimensional lagged-coordinate embedding:

$$ \vec{x}_t = \langle x_t, x_{t-\tau}, \dots, x_{t-(E-1)\tau} \rangle $$

<!------------
   Figure 2
------------->
```{r fig_attractor_reconstruction, echo = FALSE, fig.cap = "Attractor Reconstruction from 3 Lagged Coordinates", out.width = "50%", fig.align = 'center'}
knitr::include_graphics("Lorenz_Reconstruct.png")
```

If sufficient lags are used, the reconstruction preserves essential mathematical properties of the original system: reconstructed states will map one-to-one to actual system states, and nearby points in the reconstruction will correspond to similar system states. Figure 2 shows a reconstruction of the Lorenz attractor where the reconstructed system state is comprised of 3 lags of variable $x$. Here, the visual similarity between the reconstruction and the original Lorenz Attractor is quite clear.

As a consequence of the fact that dynamical properties of the original system can be recovered from a single time series, there are multiple applications. For example, empirical models can be used for forecasting [@Sugihara_1990], to understand nonlinear behavior [@Sugihara_1994], or to uncover mechanism [@Dixon_1999]. Moreover, recent work describes how EDM can be used to identify causal interactions, by testing whether two time series are observed from the same system [@Sugihara_2012]. In the next section, we demonstrate how **rEDM** can be used to accomplish these various tasks.


<!------------------------------------------------------------------------->
# Demonstration of EDM
<!------------------------------------------------------------------------->

<!------------------------------------------------------------------------->
## Nearest Neighbor Forecasting using Simplex Projection
<!------------------------------------------------------------------------->

As mentioned previously, the reconstruction will map one-to-one to the original attractor manifold if enough lags are used (i.e. if the reconstruction has a sufficiently large embedding dimension). If the embedding dimension is too small, then reconstructed states can overlap and appear to be the same even though they actually correspond to different states. These "singularities" will result in poor forecast performance because the system behavior cannot be uniquely determined in the reconstruction. As a consequence, we can use prediction skill as an indicator for identifying the optimal embedding dimension. In the following example we demonstrate the `Simplex()` projection nearest neighbor forecasting method [@Sugihara_1990], and its' extension `EmbedDimension()` that automates evaluation of an optimal embedding dimension.

<!------------------------------------------------------------------------->
### Example
<!------------------------------------------------------------------------->

In this example, time series come from a simulation of the tent map that exhibits chaotic behavior. The tent map is a discrete-time dynamic system, where a sequence, $x_t$, on the interval $[0, 1]$ is iterated according to:

\begin{equation*}
  x_{t+1} = \begin{cases}
            2x_t     & x_t <   \frac{1}{2}\\
            2(1-x_t) & x_t \ge \frac{1}{2}
            \end{cases}
\end{equation*}

In **rEDM**, a sample time series of first-differenced values can be found in dataset `TentMap`.

We begin by loading the **rEDM** package and examining the `TentMap` data:
```{r load package}
library(rEDM)
str(TentMap)
```

We can see that the data consists of a data.frame with two columns: `Time` and `TentMap`. All rEDM input data files or data.frames are assumed to have a time vector in the first column.  Data files are expected to be in .csv format with the first line a header of column names, data.frames are also expected to have column names.

The `Simplex` function has 5 required parameters:

  |          |              |
--|----------|--------------|----------------------------------------
1.| `columns`|  `TentMap`   | name of column(s) of embedding library
2.| `target` |  `TentMap`   | name of column for prediction
3.| `lib`    |  `"1 100"`   | start stop indices of embedding library
4.| `pred`   |  `"201 500"` | start stop indices of predictions
5.| `E`      |  `3`         | embedding dimension

`columns` specifies the timeseries vector(s) that form the library, `target` is the column on which predictions will be made. `lib` defines row indices of the "training" portion of data, `pred` corresponds to row indices of the "test" portion, and `E` defines the embedding dimension.

*Note that if any overlap in the lib and pred is found, it will enable leave-one-out cross-validation. If `verbose = TRUE`, a warning message will be raised.*

In this univariate case, we specify the "TentMap" column of the data frame for both `columns` and `target`, and select the first 100 points (indices 1 to 100) in the time series to constitute the "library set", and a separate 300 point span (indices 201 to 500) as the "prediction set".  

Default parameters of `knn` (k-nearest neighbors) and `Tp` (time-to-prediction) are assumed. The default `knn = 0` sets the number of nearest neighbors to `E + 1`, and default `Tp` is 1 timestep (observation row).  With these parameters we demonstrate the `Simplex()` function:

```{r simplex on tentmap}
simplex_out <- Simplex( dataFrame = TentMap, lib = "1 100", pred = "201 500",
columns = 'TentMap', target = 'TentMap', E = 3 )
simplex_out[ c(1:2, 300:301), ]
```

Note that the returned data.frame has 1 `NaN` as the first `Predictions` point since `Tp = 1`, and, the last `Observations` will likewise be `NaN` with the time vector adjusted to accommodate `Tp` rows beyond the data as needed.

Computation of Pearson correlation, MAE and RMSE errors between the forecast `Observations` and `Predictions` can be performed with the `ComputeError()` function.

```{r tentmap simplex stats}
ComputeError( simplex_out $ Observations, simplex_out $ Predictions )
```

<!------------------------------------------------------------------------->
## Optimal embedding dimension
<!------------------------------------------------------------------------->

As noted earlier, identification of the optimal embedding dimension to best "unfold" the dynamics can be assessed with simplex prediction skill. **rEDM** provides the `EmbedDimension()` function to automate this task. `EmbedDimension()` parallelises function calls to `Simplex()`, which automatically sets values of `E` from 1 to `maxE`=10.  Continuing with the previous example, we invoke `EmbedDimension()`:

```{r Embed_Dim_TentMap, fig.cap="TentMap data prediction skill vs. embedding dimension." }
rho_E <- EmbedDimension( dataFrame = TentMap, lib = "1 100", pred = "201 500",
columns = 'TentMap', target = 'TentMap' )
```

The output is a data.frame with columns `E` and `rho` detailing the embedding dimension and Pearson correlation coefficient between the simplex projected forecast at `Tp = 1` timesteps ahead, and the observed data over the `pred` indices. Here, we observe that forecast skill peaks at `E = 2`, indicating that the dynamics of our data are unfolded best in 2 dimensions. *Note that this optimal value does not have to correspond to the dimensionality of the original system.*  The forecast skill will be affected by factors such as observational noise, process error, and time series length, and so it is more useful to think of the embedding dimension as a practical measure that is dependent on properties of the data.

<!------------------------------------------------------------------------->
## Prediction Decay
<!------------------------------------------------------------------------->

An important property of many natural systems is that nearby trajectories eventually diverge over time (i.e. "deterministic chaos" -- the "butterfly effect"). In essence, this means that while short-term prediction is often possible, information about the predictive state of the system is diluted over time, hindering long-term forecasting. We can demonstrate this effect by examining how prediction skill changes as we increase the `Tp` argument, the "time to prediction", defining the number of time steps into the future at which forecasts are made.  **rEDM** provides the `PredictInterval()` function to automate this task.

<!------------------------------------------------------------------------->
### Example
<!------------------------------------------------------------------------->

Using the same data with the `PredictInterval()` function, we supply the embedding dimension parameter with the value determined previously (`E = 2`):

```{r Prediction_interval_on_TentMap, fig.cap = "Tent map first differences simplex prediction skill as a function of forecast interval." }
rho_Tp <- PredictInterval( dataFrame = TentMap, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )
```

As above, the returned object is a data.frame with forecast skill `rho` and time to prediction `Tp`.  As expected (because the parameters chosen for the tent map fall in the region for chaotic behavior), the decline in forecast skill (`rho` $\rightarrow 0$) as the forecast interval `Tp` increases, indicates that the system may be chaotic.


<!------------------------------------------------------------------------->
## Identifying Nonlinearity
<!------------------------------------------------------------------------->

One concern is that time series may show predictability even if they are purely stochastic, they behave similarly to autocorrelated red noise. Fortunately, we can distinguish between red noise and nonlinear deterministic behavior by using S-maps as described in [@Sugihara_1994].

In contrast to the nearest-neighbor interpolation of simplex projection, the S-map forecasting method [@Sugihara_1994] fits local linear maps to describe the dynamics. In addition to the standard set of parameters for a lagged-coordinate reconstruction as in simplex, S-maps contain a nonlinear localisation parameter, $\theta$, that determines the degree to which points are weighted when fitting the local linear map. For example, when $\theta = 0$, all points are equally weighted, such that the local linear map is identical for different points in the reconstructed state-space. As such, the S-map will be identical to a global linear map (i.e. an autoregressive model). When values of $\theta$ are greater than $0$, nearby points in the state space receive larger weight, and the local linear map can vary in state-space to accommodate nonlinear behavior. 

Consequently, if the time series are sampled from autoregressive red noise, then the linear model ($\theta = 0$) should produce better forecasts, because the global linear map (which will, in effect, be fitted to more data points) will reduce the effects of observation error compared to local linear maps. In contrast, if forecast skill increases for $\theta > 0$, then the results are suggestive of nonlinear dynamics wherein better forecasts are achieved when the local linear map can change depending on the location in state-space: it is a better description of state-dependent behavior.

<!------------------------------------------------------------------------->
### Example
<!------------------------------------------------------------------------->

The `PredictNonlinear()` function provides an evaluation of S-map forecast skill as a function of the localisation parameter `theta`.  If unspecified, `theta` values will range from 0.01 to 9.

Typically, when using `S-map` to test for nonlinear behavior, we want to use all available points in the reconstruction of the local linear map, not just `knn` nearest neighbors as in simplex projection.  With all points available, S-map uses the `theta` parameter to control the weighting assigned to individual points, thereby localising the dynamics to capture nonlinear behavior.  When `knn = 0`, the default, `SMap()` will use all available points.

Here we use an embedding dimension of `E = 2` and the same parameters as in the previous examples, however, we specify the "TentMapNoise" data that adds Gaussian noise to the TentMap data as one would normally encounter with noisy observational data.

```{r Predict_nonlinear_on_TentMap, fig.cap = "Tent map first differences S-map prediction skill as a function of S-map localisation parameter."}
rho_theta <- PredictNonlinear( dataFrame = TentMapNoise, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )
```

The result is a data.frame with columns `Theta` and `rho`. Here, we see that forecast skill substantially improves as `theta` increases, indicating the presence of nonlinear dynamics. We also observe a degradation in forecast skill at high values of `theta` as the local linear map overfits to insufficient nearest neighbors.

<!------------------------------------------------------------------------->
## `Simplex()` and `SMap()` functions
<!------------------------------------------------------------------------->

The functions `EmbedDimension()`, `PredictInterval()` and `PredictNonlinear()` are multithreaded wrapper functions for the `Simplex()` and `SMap()` algorithms. `EmbedDimension()` and `PredictInterval()` parallelise calls to `Simplex()` to evaluate forecast skill as a function of embedding dimension and prediction interval respectively.  `PredictNonlinear()` parallelises calls to `SMap()` to assess predictive skill as a function of the nearest neighbor localisation parameter. However, one can equivalently call the underlying `Simplex()` and `SMap()` functions directly.

<!------------------------------------------------------------------------->
### `Simplex()`
<!------------------------------------------------------------------------->

For example, evaluation of the simplex prediction at an optimal embedding dimension of `E = 2` can be performed as:

```{r Simplex_TentMap }
tentMapPredict <- Simplex( dataFrame = TentMap, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )

ComputeError( tentMapPredict $ Observations, tentMapPredict $ Predictions ) $ rho
```

<!------------------------------------------------------------------------->
### `SMap()`
<!------------------------------------------------------------------------->

An individual S-map evaluation corresponding to the optimal `PredictNonlinear()` result from above is:

```{r Tentmap_Noise_S-map } 
smap = SMap( dataFrame = TentMapNoise, lib = "1 100", pred = "201 500",
target = 'TentMap', columns = 'TentMap', E = 2, theta = 3 )
```

`SMap()` returns a named list with `predictions` and  `coefficients` data.frames, with `NaN` inserted appropriately where no predictions or observations are available. 

```{r S-map_tentMap_results}
head( cbind( smap $ predictions, smap $ coefficients ), 2 )
tail( cbind( smap $ predictions, smap $ coefficients ), 2 )
```

<!------------------------------------------------------------------------->
## Generalized Takens Theorem
<!------------------------------------------------------------------------->

A practical reality is that sampled observations of complex dynamics are usually composed of finite, noisy data.  Additionally, the presence of stochastic, non-deterministic drivers means that "multivariate" reconstructions can often be a better description than "univariate" reconstructions.  This means that in addition to creating an attractor from lags of one time series, it can be advantageous to combine different time series to create the phase-space embedding, provided they are all observed from the same system [@Sauer_1991; @Deyle_2011]. 

In **rEDM**, the `Simplex()` and `SMap()` functions allow multivariate reconstructions from any set of observation vectors. A multivariate reconstruction is defined by specifying which columns to use as coordinates in the `columns` argument, and which column is to be forecast in the `target` argument.  By default, **rEDM** will create Takens time-delay embeddings from univariate or multivariate data, however, this can be prevented by setting the `embedded` parameter `TRUE`. In this case, the input data are assumed to already constitute a valid multidimensional embedding, and no time-delay embedding is performed. 

<!------------------------------------------------------------------------->
### Example
<!------------------------------------------------------------------------->

We begin by examining an example dataset from a coupled 3-species model system.

```{r block_3sp_data}
head( block_3sp, 3 )
```

Here, `block_3sp` is a 10-column data.frame with 9 data columns. This data has already been time-delay embedded to dimension `E = 3` with a time delay of `tau = -1`. We use simplex forecasting based on a multivariate embedding of the three data vectors `x_t x_t-1 z_t` with `embedded = TRUE` and `E = 3`:

```{r 3_Species_simplex, fig.cap="Simplex Tp=1 forecast of $x_t$."}
smplx_3species = Simplex( dataFrame = block_3sp, lib = "1 100",
pred = "101 190", E = 3, columns = "x_t x_t-1 z_t", target = "x_t",
embedded = TRUE )
```

A plot of the predictions vs. observations can be examined with:

```{r plot_3_species, fig.cap="Scatter plot of simplex forecast of $x_t$ vs. observations."}
err = ComputeError( smplx_3species $ Observations,
                    smplx_3species $ Predictions )
plot( smplx_3species $ Observations, smplx_3species $ Predictions,
pch = 19, cex = 0.5, xlab = 'Observations', ylab = 'Predictions',
main = '3 Species x_t' )
abline( a = 0, b = 1, lty = 2, col = 'blue' )
text( -1, 1, paste( capture.output( cbind( err ) ), collapse = '\n' ) )
```

<!------------------------------------------------------------------------->
## S-map Coefficients
<!------------------------------------------------------------------------->

As described in [@Deyle_2016], S-map coefficients from the appropriate multivariate embedding can be interpreted as dynamic, time-varying interaction strengths. We demonstrate this with a chaotic timeseries described in [@Lorenz_1996], defined for N variables k=1, ... N, as 

\begin{equation*}
  \frac{ dx_{k} }{ dt } = -X_{k-2} X_{k-1} + X_{k-1} X_{k+1} - X_k + F
\end{equation*}

The `Lorenz5D` data.frame contains a N=5 dimensional system with F=8 from [@Lorenz_1996].  Here, we use `SMap()` to compute a 4-dimensional forecast at `Tp`=1:

```{r S-map_Lorenz96 }
smap_Lorenz <- SMap( dataFrame = Lorenz5D, lib = "1 500", pred = "601 900",
E = 4, theta = 3, columns = "V1 V2 V3 V4", target = "V1", embedded = TRUE )
```

As noted earlier, `SMap()` returns a named list with two data frames, `predictions` and `coefficients`:

```{r set precision, echo = FALSE}
options(digits=4)
```

```{r S-map_Lorenz_results, warning = FALSE}
head( cbind( smap_Lorenz $ predictions, smap_Lorenz $ coefficients[,2:6] ), 3 )
```

Here, we plot the time series for the observed (blue) and predicted (red) values of `V1` in the top panel; and the inferred interactions (S-map coefficients) for the influence of `V4`, `V3` and `V2` on future values of `V1` in the lower panels.

```{r, echo = FALSE}
par(mfrow = c(4, 1), mar = c(2, 4, 0.5, 1), oma = c(0, 0, 0, 0),
    mgp = c(2.1, 1, 0) )
```

```{r smap_Lorenz_plot, out.width = "70%", fig.cap = "S-map prediction and coefficients of Lorenz'96 5-D system."}
predictions = smap_Lorenz $ predictions
coefficients = smap_Lorenz $ coefficients
Time = predictions $ Time

plot( Time, predictions $ Observations, type = "l", col = "blue",
ylab = "V1", xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
lines( Time, predictions $ Predictions, lwd = 2, col = 'red' )
legend("topright", legend = c("observed", "predicted"),
fill = c( 'blue', 'red' ), bty = "n", cex = 1.3 )

plot( Time, coefficients[, 6], type = "l", col = "brown",
ylab = paste( "\U2202", "V4/", "\U2202", "V1" , sep = '' ), 
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
plot( Time, coefficients[, 5], type = "l", col = "darkgreen",
ylab = paste( "\U2202", "V3/", "\U2202", "V1" , sep = '' ),
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
plot( Time, coefficients[, 4], type = "l", col = "blue",
ylab = paste( "\U2202", "V2/", "\U2202", "V1" , sep = '' ),
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
```

```{r, echo = FALSE}
par(mfrow = c(1, 1), mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))
```

\newpage
<!------------------------------------------------------------------------->
## Multiview Embedding
<!------------------------------------------------------------------------->

The generality of Takens Theorem means that in situations with multivariate time series, there can often be many different, valid attractor reconstructions. As described in [@Ye_2016], combining these different models can result in improved forecasts.

Here, we demonstrate this idea using the `Multiview()` function with the 3-species data used above. `Multiview()` operates by constructing all possible embeddings of dimension `E` with lag up to `E-1`.  These embeddings are ranked by forecast skill (`rho`) over the `lib` portion of the data.  The individual forecasts for the top `multiview` embeddings are then averaged together.  If `multiview` is not specified it is set to sqrt(C) where C is the number of E-dimensional combinations created from all data vectors.

```{r multiview}
Mview = Multiview( dataFrame = block_3sp, lib = "1 100", pred = "101 190", E = 3, columns = "x_t y_t z_t", target = "x_t" )
```

`Multiview()` returns a named list with two data.frames: `View`, and `Predictions`.  `View` lists the various combinations of data embedding vectors used for the forecasts along with their prediction statistics.  `Predictions` returns the final averaged multiview projections.

```{r multiview results}
Mview $ View[ which( Mview $ View $ rho > 0.91 ),  ]
```

\newpage
<!------------------------------------------------------------------------->
## Causality Inference and Cross Mapping
<!------------------------------------------------------------------------->

One of the corollaries to the Generalized Takens Theorem is that it should be possible to cross predict or cross map between variables that are observed from the same system. Consider two variables, $x$ and $y$ that interact in a dynamic system. Then the univariate reconstructions based on $x$ or $y$ alone should uniquely identify the system state and and thus the corresponding value of the other variable.

<!------------
   Figure 10
------------->
```{r fig_cross_map, echo = FALSE, fig.cap = "Cross Mapping Between Reconstructions of the Lorenz Attractor"}
knitr::include_graphics("CrossMap.png")
```
In the case of unidirectional causality, e.g. $x$ causes $y$, the causal variable ($x$) leaves a signature on the affected variable ($y$). Consequently, the reconstructed states based on $y$ can be used to cross predict the values of $x$ (because the reconstruction based on $y$ must be complete, it must include information about the value of $x$). Note that this cross prediction is in the *opposite* direction of the causal effect. At the same time, cross prediction from $x$ to $y$ will fail, because the time series of $x$ behaves independently of $y$, so a univariate reconstruction using only lags of $x$ is necessarily incomplete.

Although $x$ has incomplete information for predicting $y$, it does affect the values of $y$, and therefore will likely to have nonzero predictive skill. However, this cross mapping will be limited to the statistical association between $x$ and $y$ and will generally not improve as longer time series are used for reconstruction. In contrast, the cross prediction of $x$ from $y$ will generally improve. This convergence is therefore a crucial property for inferring causality. For practical reasons, the sensitivity of detecting causality this way is improved if, instead of predicting the future value of another variable, we estimate the concurrent value of another variable. We refer to this modified method as cross mapping, because we are not "predicting" the future.

For a more detailed description of using cross mapping to infer causation, see [@Sugihara_2012].

<!------------------------------------------------------------------------->
## Convergent Cross Mapping (CCM)
<!------------------------------------------------------------------------->

In **rEDM**, convergent cross mapping is implemented as the `CCM()` function, which provides a wrapper to compute cross map skill for different subsamples of the data. In the following example, we reproduce the analysis from [@Sugihara_2012] to identify causality between anchovy landings in California and Newport Pier sea-surface temperature. For this example, a previously identified value of `3` for the embedding dimension will be used.

To quantify convergence, we compute the cross map skill over many random subsamples of the time series. The `libSizes` argument specifies the size of the library set, and `sample` specifies the number of subsamples generated at each library size. `random` and `replacement` specify how the subsamples will be generated.  The default is random sampling without replacement.

```{r sardine_anchovy_ccm, fig.cap = "Convergent cross mapping of Newport sea surface temperature with anchovy landings."}
cmap <- CCM( dataFrame = sardine_anchovy_sst, E = 3, Tp = 0, columns = "anchovy", target = "np_sst", libSizes = "10 70 5", sample = 100, showPlot = TRUE )
```

The output is a data.frame with statistics for each model run (in this case, 100 models at each library size) as a function of library size.  Recalling that cross mapping indicates causal influence in the reverse direction (target to source), we see that the cross mapping `anchovy:np_sst` converges at a positive value of $\rho$, indicating that Newport sea surface temperature influences anchovy landings.  Because average cross map skill less than 0 means there is no prediction skill, (predictions should not be anticorrelated with observations), we infer from the `np_sst:anchovy` cross mapping that anchovy landings do not effect sea surface temperatures. 


<!------------------------------------------------------------------------->
# Real Data Example
<!------------------------------------------------------------------------->

<!------------------------------------------------------------------------->
## Apple-Blossom Thrips
<!------------------------------------------------------------------------->

In this example, we use EDM to re-examine the classic apple-blossom thrips (*Thrips imaginis*) time series from the Wait Institute in Australia [@Davidson_1948; @Davidson_1948a]. Seasonal outbreaks of *Thrips imaginis* were observed to vary greatly in magnitude from year to year, but large outbreaks tended to coincide across large spatial domains. This lead to the hypothesis that regional-scale climatic factors were responsible for controlling the size of the seasonal outbreaks (what might now be called the Moran effect [@Moran_1953]).

```{r Thrips data}
head(Thrips, 2)
```

The data column `Thrips_imaginis` contains counts of *Thrips imaginis* obtained from the Global Population Dynamics Database (GPDD) [@NERC-Centre-for-Population-Biology_2010]. `maxT_degC` is the mean maximum daily temperature ($^\circ$C) taken over each month and `Rain_mm` is the monthly rainfall (mm), both from the Waite Institute. The final column `Season` is a simple annual sinusoid that peaks in December (the Austral summer) to emulate an indicator of season.

First, we plot the data. Note that all the time-series variables, particularly the mean maximum daily temperature, show marked seasonality.

```{r thrips plot, echo = FALSE, fig.width = 3, fig.height = 3.5, fig.cap = "Thrips abundance and environmental variables."}
par(mfrow = c(4, 1), mar = c(2, 4, 0, 1), mgp = c(2., 0.5, 0))
time_dec <- Thrips $ Year + (Thrips $ Month)/12
plot(time_dec, Thrips $ Thrips_imaginis, type = "l", lwd = 2, col = "darkgreen", ylab = "Thrips", xlab = "")
plot(time_dec, Thrips $ maxT_degC, type = "l", lwd = 2, col = "red", ylab = "maxT (C)", xlab = "")
plot(time_dec, Thrips $ Rain_mm, type = "l", lwd = 2, col = "blue", ylab = "Rain (mm)", xlab = "")
plot(time_dec, Thrips $ Season, type = "l", lwd = 2, col = "magenta", ylab = "Season", xlab = "" )
mtext("Year", side = 1, outer = TRUE, line = 1)
```

```{r, echo = FALSE }
par(mfrow = c(1, 1), mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))
```

<!------------------------------------------------------------------------->
### Univariate Analysis
<!------------------------------------------------------------------------->

We first examine the dependence of simplex predictability on the embedding dimension.

```{r univariate thrips, fig.cap = "Simplex embedding dimension for Thrips abundance.", fig.width = 4, fig.height = 3. }
rho_E <- EmbedDimension( dataFrame = Thrips, columns = "Thrips_imaginis", target = "Thrips_imaginis", lib = "1 72", pred = "1 72", showPlot = TRUE )
```

While there is an initial peak in the simplex prediction at `E = 3`, the global maximum is at `E = 8`. This suggests that both `E = 3` and `E = 8` are practical embedding dimensions, although `E = 8` is preferrable with a higher predictive skill. 

To test for nonlinearity we use the S-map `PredictNonlinear()` function.

```{r smap for thrips, fig.cap = "SMap localisation parameter for Thrips abundance.", fig.width = 4, fig.height = 3.}
E = 8
rho_theta_e3 = PredictNonlinear(dataFrame = Thrips, columns = "Thrips_imaginis", target = "Thrips_imaginis", lib = "1 73", pred = "1 73", E = E )
```

The S-map results demonstrate clear nonlinearity in the Thrips time series, as nonlinear models `theta > 0` give substantially better predictions than the linear model `theta = 0`. This suggests that *Thrips*, despite the strong seasonal dynamics, do not simply track the environment passively, but have some intrinsic dynamics. To look more closely at the issue of seasonal drivers, however, we turn to convergent cross-mapping (CCM).

<!------------------------------------------------------------------------->
### Seasonal Drivers
<!------------------------------------------------------------------------->

Recall that there is a two-part criterion for CCM to be a rigorous test of causality:

1. The cross map prediction skill is statistically significant when using the full time series as the library.
2. Cross map prediction demonstrates convergence, i.e. prediction skill increases as more of the time series is used for the library and the reconstructed attractor becomes more dense.

For an initial summary, we first compute the cross map skill (measured with Pearsons $\rho$) for each variable pair. Note that `CCM()` computes the cross map in both "directions". 

```{r compute ccm matrix for thrips, results='hold', cache = TRUE}
vars       = colnames(Thrips[3:6])
var_pairs  = combn( vars, 2 )  # Combinations of vars, 2 at a time
libSize    = paste( NROW(Thrips) - E, NROW(Thrips) - E, 10, collapse = " " )
ccm_matrix = array(NA, dim = c(length(vars), length(vars)), 
                   dimnames = list(vars, vars))

for( i in 1:ncol( var_pairs ) ) {
   ccm_out = CCM(dataFrame = Thrips,
                 columns = var_pairs[1,i], target = var_pairs[2,i],
                 libSizes = libSize, Tp = 0, E = E, sample = 100)
                 
   outVars = names( ccm_out )

   var_out = unlist( strsplit( outVars[2], ':' ) )
   ccm_matrix[ var_out[2], var_out[1] ] = ccm_out[1,2]
                 
   var_out = unlist( strsplit( outVars[3], ':' ) )
   ccm_matrix[ var_out[2], var_out[1] ] = ccm_out[1,3]
}
```

We note that `ccm_matrix` rows are the second of the `CCM()` returned variables, while columns are the first variable.  As outlined earlier, influences are quantified from the `target` to the `columns` variable so that here, rows are considered the 'target`, influencing variable, and columns the `columns` influenced variable. 

For comparison we also compute the lagged cross-correlation, allowing lags of up to $\pm 6$ months.

```{r compute corr matrix for thrips, cache = TRUE}
corr_matrix <- array(NA, dim = c(length(vars), length(vars)), 
                     dimnames = list(vars, vars))

for(ccm_from in vars) {
    for(ccm_to in vars[vars != ccm_from]) {
        ccf_out <- ccf(Thrips[,ccm_from], Thrips[,ccm_to], 
                       type = "correlation", lag.max = 6, plot = FALSE)$acf
        corr_matrix[ccm_from, ccm_to] <- max(abs(ccf_out))
    }
}
```

We compare the two matrices.

```{r xmap vs. corr matrix for thrips}
ccm_matrix
corr_matrix
```

We can see that the cross map strengths are not symmetric.  In general CCM(X1 : X2) != CCM(X2 : X1). We also notice that the cross map and correlation between temperature and the seasonal indicator are high, with the cross map results suggesting that the seasonal variable can almost perfectly recover the temperature, $\rho = `r format(ccm_matrix["Season", "maxT_degC"], 2)`$. This makes interpretation more complicated, because we have to consider the possibility that cross mapping is simply identifying the shared seasonality between two time series. In other words, cross mapping between temperature and any variable with a seasonal cycle, might suggest an interaction even if there is no actual causal mechanism.

<!------------------------------------------------------------------------->
#### Convergent Cross-Mapping
<!------------------------------------------------------------------------->

With this in mind, we examine convergence in cross-map predictability, i.e. we compute `rho` as a function of library size `L`. The magnitude of the cross-correlation between *Thrips* and the cross mapped variable is shown as a black dashed line for comparison. 

```{r thrips setup, echo = FALSE}
par( mfrow = c( 1, 3 ), mar = c( 3.5, 1.3, 2.3, 0.2 ), mgp = c(1.5, 0.5, 0) )

CCMPlot = function( ccm.df, E, ylim = c( 0.2, 1 ) ) {
    libSize = ccm.df $ LibSize
    V1      = names( ccm.df )[2]
    V2      = names( ccm.df )[3]

    title = paste( V1, " : ", V2, "\nE=" , E )
    
    print( title )
    print( libSize )

    plot( libSize, ccm.df[ , V1 ],
          ylim = ylim, main = title, col = "blue", type = "l", lwd = 3,
          xlab = 'Library Size', ylab = 'Prediction Skill (\U03C1)' )
    lines( libSize, ccm.df[ , V2 ], col = "red", lwd = 3 )
    abline( h = 0 )
    legend( 'topright', c( V1, V2 ), 
            fill = c( 'blue', 'red' ), bty = 'n', cex = 1.2 )
}
```

```{r ccm thrips, fig.cap = "Thrips cross mapped to climatic variables. Vertical axis is cross map prediction skill (rho).", fig.width = 7.5}
thrips_xmap_maxT <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                        columns = "Thrips_imaginis", target = "maxT_degC",
                        libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_maxT, E )
abline(h = corr_matrix['Thrips_imaginis', 'maxT_degC'], col = "black", lty = 2)

thrips_xmap_Rain <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                        columns = "Thrips_imaginis", target = "Rain_mm",
                        libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_Rain, E )
abline(h = corr_matrix['Thrips_imaginis', 'Rain_mm'], col = "black", lty = 2)

thrips_xmap_Season <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                          columns = "Thrips_imaginis", target = "Season",
                          libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_Season, E )
abline(h = corr_matrix['Thrips_imaginis', 'Season'], col = "black", lty = 2)
```

```{r thrips setdown, echo = FALSE}
par(mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))
```

The results show evidence of convergence for *Thrips* cross mapping to temperature and season variables, with the $\rho$ at maximum library size `L` significantly exceeding linear correlation. The rain variable does not indicate a substantially different cross map interaction, appearing confounded as to causal influence.

In addition, we are still left with the conundrum that temperature and to a lesser extent, rainfall, are easily predicted from the seasonal cycle, and so we cannot immediately ignore the possibility that the cross map results are an artifact of shared seasonal forcing.

To reframe, we wish to reject the null hypothesis that the level of cross mapping we obtain for `maxT_degC` and `Rain_mm` can be solely explained by shared seasonality. This hypothesis can be tested using randomization tests based on surrogate data. The idea here is to generate surrogate time series with the same level of shared seasonality. Cross mapping between the real time series and these surrogates thus generates a null distribution for $\rho$, against which the actual cross map $\rho$ value can be compared.

<!------------------------------------------------------------------------->
#### Seasonal Surrogate Test
<!------------------------------------------------------------------------->

```{r seasonal surrogates thrips, cache = TRUE}
# Create matrix with temperature and rain surrogates (1000 time series vectors)
surr_maxT = SurrogateData( Thrips $ maxT_degC, method = "seasonal", T_period = 12, num_surr = 1000, alpha = 3)
surr_rain = SurrogateData( Thrips $ Rain_mm, method = "seasonal", T_period = 12, num_surr = 1000, alpha = 3)

# Rain cannot be negative
surr_rain = apply( surr_rain, 2, function(x){ i = which(x<0); x[i] = 0; x } )

# data.frame to hold CCM rho values between Thrips abundance and variable
rho_surr <- data.frame( maxT = numeric(1000), Rain = numeric(1000) )

# data.frames with time, Thrips, and 1000 surrogate climate variables for CCM()
maxT_data = as.data.frame( cbind( seq(1:nrow(Thrips)), Thrips $ Thrips_imaginis, surr_maxT ) )
names( maxT_data ) = c( 'time', 'Thrips_imaginis', paste( 'T', as.character( seq( 1,1000 ) ), sep = '' ) )

rain_data = as.data.frame( cbind( seq(1:nrow(Thrips)), Thrips $ Thrips_imaginis, surr_rain ) )
names( rain_data ) = c( 'time', 'Thrips_imaginis', paste( 'R', as.character( seq( 1,1000 ) ), sep = '' ) )

# Cross mapping
for (i in 1:1000) {
  targetCol = paste( 'T', i, sep = '' ) # as in maxT_data
  
  ccm_out = CCM( dataFrame = maxT_data, E = E, Tp = 0,
                 columns = 'Thrips_imaginis', target = targetCol,
                 libSizes = "73 73 5", sample = 1 )
                 
  col = paste( 'Thrips_imaginis', ':', targetCol, sep = '' )
   
  rho_surr $ maxT[i] = ccm_out[1,col]
}

for (i in 1:1000) {
  targetCol = paste( 'R', i, sep = '' ) # as in rain_data
  
  ccm_out = CCM( dataFrame = rain_data, E = E, Tp = 0,
                 columns = 'Thrips_imaginis', target = targetCol,
                 libSizes = "73 73 5", sample = 1 )
                 
  col = paste( 'Thrips_imaginis', ':', targetCol, sep = '' )
   
  rho_surr $ Rain[i] = ccm_out[1,col]
}
```

We now have a null distribution, and can estimate a $p$-value for rejecting the null hypothesis of mutual seasonality.

```{r significance of randomization test}
1 - ecdf( rho_surr $ maxT )( ccm_matrix["maxT_degC", "Thrips_imaginis" ] )
1 - ecdf( rho_surr $ Rain )( ccm_matrix["Rain_mm",   "Thrips_imaginis" ] )
```

In the case of temperature, the CCM influence we estimated (`r ccm_matrix["maxT_degC", "Thrips_imaginis" ]`) is higher than the linear correlation (`r corr_matrix["maxT_degC", "Thrips_imaginis" ]`), and is highly significant in relation to a surrogate null distribution.  Regarding rainfall, the CCM influence (`r ccm_matrix["Rain_mm", "Thrips_imaginis" ]`) is higher than the linear correlate (`r corr_matrix["Rain_mm", "Thrips_imaginis" ]`), but not significant at the 95th percentile of the surrogate null distribution.  We note that the original Thrips data collections were at a much higher frequency than those available through the GPDD, and that monthly accumulated rainfall may be inadequate to resolve lifecycle influences on a species with a lifecycle of approximately one month. With more highly resolved data, it may well be possible to establish significance.

```{r reset par, include = FALSE}
par( par.orig )
```

<!------------------------------------------------------------------------->
# Package Core
<!------------------------------------------------------------------------->
The **rEDM** package is implemented as a wrapper to the [`cppEDM`](https://github.com/SugiharaLab/cppEDM) library. All EDM algorithms are executed in the core `cppEDM` library and interfaced through the [`Rcpp`](https://cran.r-project.org/package=Rcpp) package.

<!------------------------------------------------------------------------->
## Data Input
<!------------------------------------------------------------------------->
Data can be input as an R `data.frame`, or read from a `.csv` file. In either case, the first column must define a time vector, and all columns are expected to be named. The time vector can be a string encoding a Date, or Datetime format.  All subsequent columns are expected to be numeric.

<!------------------------------------------------------------------------->
### S-Map coefficients and embedded data
<!------------------------------------------------------------------------->
`SMap()` should be called with a DataFrame that has columns explicity corresponding to dimensions `E`.  This means that if a multivariate data set is used, it should not be called with an embedding from `Embed()` since `Embed()` will add lagged coordinates for each variable.  These extra columns will then not correspond to the intended dimensions in the matrix inversion and prediction reconstruction and subsequent S-map coefficients. In this case, use the `embedded = TRUE` parameter with the multivariate data so that columns selected correspond to the proper dimension.

<!------------------------------------------------------------------------->
# Parameters
<!------------------------------------------------------------------------->
Since **rEDM** is a wrapper for the [`cppEDM`](https://github.com/SugiharaLab/cppEDM) library, parameters largely correspond to function parameters of [`cppEDM`](https://github.com/SugiharaLab/cppEDM).  Primary parameters are tabulated here.

```{r parameter-table, echo = FALSE}
paramTable = read.csv( "ParameterTable.csv", as.is = TRUE )
knitr::kable( paramTable, caption = '' )
```

<!------------------------------------------------------------------------->
# Acknowledgements
<!------------------------------------------------------------------------->
**rEDM** is an evolving implementation of EDM code. Previous versions were developed by George Sugihara, Alan Trombla, Richard Penner, Victor Wong, Martin Casdagli, Jerome Cartagena, Mohsen Azarbayejani, Ava Pierce, Jennifer Trezzo, and Hao Ye.

<!------------------------------------------------------------------------->
# References
<!------------------------------------------------------------------------->

---
title: "Empirical Dynamic Modeling"
author:
- Joseph Park
- George Sugihara
- Ethan Deyle
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette: default # 1st entry -> R CMD build
  pdf_document:
    latex_engine: xelatex # Select RStudio knitr dropdown -> pdf
vignette: >
  %\VignetteIndexEntry{Empirical Dynamic Modeling}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
bibliography: References.bib
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(echo = TRUE, warning = FALSE, tidy = TRUE,
                      fig.pos = 'h', fig.align = 'center' )
                      #fig.width = 4.5, fig.height = 4 )
knitr::opts_knit$set(global.par = TRUE, progress = FALSE)
options(digits = 2)
par.orig = par()
par(mar = c(3, 3, 2, 1), mgp = c(2.5, 1, 0))
```

<!------------------------------------------------------------------------->
# Abstract
<!------------------------------------------------------------------------->

Empirical dynamic modeling (EDM) is an emerging non-parametric framework for modeling nonlinear dynamic systems. EDM is based on the mathematical theory of reconstructing attractor manifolds from time series data [@Takens_1981]. The **rEDM** package collects several EDM methods, including simplex projection [@Sugihara_1990], S-map [@Sugihara_1994], multivariate embeddings [@Dixon_1999], convergent cross mapping [@Sugihara_2012], and multiview embedding [@Ye_2016]. Here, we introduce the basic underlying theory, and describe the functionality of **rEDM** using examples from both model simulations and real data.

<!------------------------------------------------------------------------->
# Installation
<!------------------------------------------------------------------------->

The **rEDM** package can be obtained in two main ways. The standard version of the package can be obtained through CRAN (the Comprehensive R Archive Network): https://cran.r-project.org/package=rEDM:

```{r CRAN_installation_instructions, eval = FALSE}
install.packages("rEDM")
```

Also available on GitHub at [SugiharaLab](https://github.com/SugiharaLab/rEDM), and can be installed using R **devtools**.

```{r GitHub_installation_instructions, eval = FALSE}
devtools::install_github("SugiharaLab/rEDM")
```

<!------------------------------------------------------------------------->
# Introduction
<!------------------------------------------------------------------------->

Many scientific fields use models as approximations of reality and for various purposes, for example, testing hypotheses regarding mechanisms or processes, explaining past observations, and predicting future outcomes. In many cases these models are based on hypothesized parametric equations; however, explicit equations can be impractical when the underlying mechanisms are unknown or are too complex to be characterized with existing datasets. Empirical models, which infer patterns and associations from the data (instead of using hypothesized equations), represent an alternative and highly flexible approach. Here, we review the theoretical background for empirical dynamic modeling (EDM) and the functionality of the **rEDM** package, which are intended for nonlinear dynamic systems that can prove problematic for traditional modeling approaches.

The basic goal underlying EDM is to reconstruct the behavior of dynamic systems using time series data. This approach is based on mathematical theory developed initially by [@Takens_1981], and expanded by others [@Sauer_1991; @Casdagli_1991; @Deyle_2011]. Because these methods operate with minimal assumptions, they are particularly suitable for studying systems that exhibit non-equilibrium dynamics and nonlinear state-dependent behavior (i.e. where interactions change over time and as a function of the system state).

<!------------------------------------------------------------------------->
# Empirical Dynamic Modeling
<!------------------------------------------------------------------------->

<!------------------------------------------------------------------------->
## Time Series as Observations of a Dynamic System
<!------------------------------------------------------------------------->

The essential concept is that time series can be viewed as projections of the behavior of a dynamic system. First, the system state can be described as a point in a high-dimensional space. The axes of this space can be thought of as fundamental state variables; in an ecosystem, these variables might correspond to population abundances, resources, or environmental conditions. Second, the system state changes through time following a set of deterministic rules. In other words, the behavior of the system is not completely stochastic.

Consequently, it is possible to project the system state onto one of the coordinate axes and obtain the value of the corresponding state variable. Sequential projections over time will thus produce a time series for that variable. For example, in figure 1 the states of the canonical Lorenz Attractor [@Lorenz_1963] are projected onto the $x$-axis, creating a time series of variable $x$.

<!------------
   Figure 1 
------------->
```{r, LorenzProjection, echo = FALSE, fig.cap = "Time Series Projection from the Lorenz Attractor.", out.width = "50%", fig.align = 'center'}
knitr::include_graphics("Lorenz_Projection.png")
```

Although different time series observed from a system can represent independent state variables, in general, each time series is an *observation function* of the system state that may convolve several different state variables.

<!------------------------------------------------------------------------->
## Attractor Reconstruction / Takens' Theorem
<!------------------------------------------------------------------------->

The goal of EDM is to reconstruct the system dynamics from time series data. As seen above, a time series can be thought of as sequential projections of the motion on an attractor; in other words, information about the behavior is encoded in the temporal ordering of the time series. Takens' Theorem [@Takens_1981] states that mathematically valid and property preserving reconstructions of the attractor can be created using lags of a single time series, then substituting those lagged time series for unknown or unobserved variables. In other words, instead of representing the system state using a complete set of state variables, we can instead use an `E`-dimensional lagged-coordinate embedding:

$$ \vec{x}_t = \langle x_t, x_{t-\tau}, \dots, x_{t-(E-1)\tau} \rangle $$

<!------------
   Figure 2
------------->
```{r fig_attractor_reconstruction, echo = FALSE, fig.cap = "Attractor Reconstruction from 3 Lagged Coordinates", out.width = "50%", fig.align = 'center'}
knitr::include_graphics("Lorenz_Reconstruct.png")
```

If sufficient lags are used, the reconstruction preserves essential mathematical properties of the original system: reconstructed states will map one-to-one to actual system states, and nearby points in the reconstruction will correspond to similar system states. Figure 2 shows a reconstruction of the Lorenz attractor where the reconstructed system state is comprised of 3 lags of variable $x$. Here, the visual similarity between the reconstruction and the original Lorenz Attractor is quite clear.

As a consequence of the fact that dynamical properties of the original system can be recovered from a single time series, there are multiple applications. For example, empirical models can be used for forecasting [@Sugihara_1990], to understand nonlinear behavior [@Sugihara_1994], or to uncover mechanism [@Dixon_1999]. Moreover, recent work describes how EDM can be used to identify causal interactions, by testing whether two time series are observed from the same system [@Sugihara_2012]. In the next section, we demonstrate how **rEDM** can be used to accomplish these various tasks.


<!------------------------------------------------------------------------->
# Demonstration of EDM
<!------------------------------------------------------------------------->

<!------------------------------------------------------------------------->
## Nearest Neighbor Forecasting using Simplex Projection
<!------------------------------------------------------------------------->

As mentioned previously, the reconstruction will map one-to-one to the original attractor manifold if enough lags are used (i.e. if the reconstruction has a sufficiently large embedding dimension). If the embedding dimension is too small, then reconstructed states can overlap and appear to be the same even though they actually correspond to different states. These "singularities" will result in poor forecast performance because the system behavior cannot be uniquely determined in the reconstruction. As a consequence, we can use prediction skill as an indicator for identifying the optimal embedding dimension. In the following example we demonstrate the `Simplex()` projection nearest neighbor forecasting method [@Sugihara_1990], and its' extension `EmbedDimension()` that automates evaluation of an optimal embedding dimension.

<!------------------------------------------------------------------------->
### Example
<!------------------------------------------------------------------------->

In this example, time series come from a simulation of the tent map that exhibits chaotic behavior. The tent map is a discrete-time dynamic system, where a sequence, $x_t$, on the interval $[0, 1]$ is iterated according to:

\begin{equation*}
  x_{t+1} = \begin{cases}
            2x_t     & x_t <   \frac{1}{2}\\
            2(1-x_t) & x_t \ge \frac{1}{2}
            \end{cases}
\end{equation*}

In **rEDM**, a sample time series of first-differenced values can be found in dataset `TentMap`.

We begin by loading the **rEDM** package and examining the `TentMap` data:
```{r load_package}
library(rEDM)
paste('rEDM version:', packageVersion('rEDM'))
```

Now examine the *TentMap* dataset:
```{r examine_TentMap_dataset}
str(TentMap)
```

The data is a data.frame with two columns: `Time` and `TentMap`. All rEDM input data files or data.frames are assumed to have a time vector in the first column.  Data files are expected to be in .csv format with the first line a header of column names, data.frames are also expected to have column names. If no time column is present in the first column set `noTime = TRUE`.

The `Simplex` function has 5 required parameters:

  |          |              |
--|----------|--------------|----------------------------------------
1.| `columns`|  `TentMap`   | name of column(s) of embedding library
2.| `target` |  `TentMap`   | name of column for prediction
3.| `lib`    |  `"1 100"`   | start stop indices of embedding library
4.| `pred`   |  `"950 999"` | start stop indices of predictions
5.| `E`      |  `3`         | embedding dimension

`columns` specifies the timeseries vector(s) that form the library, `target` is the column on which predictions will be made. `lib` defines row indices of the "training" portion of data, `pred` corresponds to row indices of the "test" portion, and `E` defines the embedding dimension.

*Note that if any overlap in the lib and pred is found, it will enable leave-one-out cross-validation.*

In this univariate case, we specify the "TentMap" column of the data frame for both `columns` and `target`, and select the first 100 points (indices 1 to 100) in the time series to constitute the "library set", and a separate 100 point span (indices 900 to 999) as the "prediction set".  

Default parameters of `knn` (k-nearest neighbors) and `Tp` (time-to-prediction) are assumed. The default `knn = 0` sets the number of nearest neighbors to `E + 1`, and default `Tp` is 1 timestep (observation row).  With these parameters we demonstrate the `Simplex()` function:

```{r simplex_tentmap}
simplex_out <- Simplex( dataFrame = TentMap, lib = "1 100", pred = "900 999",
columns = 'TentMap', target = 'TentMap', E = 3 )
simplex_out[ c(1:3, 99:101), ]
```

Note the returned data.frame has 1 `NA` as the first `Predictions` point since `Tp = 1`, and, the last `Observations` will likewise be `NA` with the time vector adjusted to accommodate `Tp` rows beyond the data as needed.

Computation of Pearson correlation, MAE and RMSE errors between the forecast `Observations` and `Predictions` can be performed with the `ComputeError()` function.

```{r tentmap_simplex_stats}
ComputeError( simplex_out $ Observations, simplex_out $ Predictions )
```

<!------------------------------------------------------------------------->
## Optimal embedding dimension
<!------------------------------------------------------------------------->

As noted earlier, identification of the optimal embedding dimension to best "unfold" the dynamics can be assessed with simplex prediction skill. **rEDM** provides the `EmbedDimension()` function to automate this task. `EmbedDimension()` parallelises function calls to `Simplex()`, which automatically sets values of `E` from 1 to `maxE`=10.  Continuing with the previous example, we invoke `EmbedDimension()`:

```{r Embed_Dim_TentMap, fig.cap="TentMap data prediction skill vs. embedding dimension.", fig.width=5, fig.height=4.5}
rho_E <- EmbedDimension( dataFrame = TentMap, lib = "1 100", pred = "201 500",
columns = 'TentMap', target = 'TentMap', showPlot = TRUE )
rho_E
```

The output is a data.frame with columns `E` and `rho` : the embedding dimension and Pearson correlation coefficient between the simplex projections at `Tp = 1` timesteps ahead and the observed data. Here, we observe that forecast skill peaks at `E = 2`, indicating that the dynamics of our data are unfolded best in 2 dimensions. *Note that this optimal value does not have to correspond to the dimensionality of the original system.*  The forecast skill will be affected by factors such as observational noise, process error, time series length, and so it is more useful to think of the embedding dimension as a practical measure that is dependent on properties of the data.

<!------------------------------------------------------------------------->
## Prediction Decay
<!------------------------------------------------------------------------->

An important property of many natural systems is that nearby trajectories eventually diverge over time (i.e. "deterministic chaos" -- the "butterfly effect"). In essence, this means that while short-term prediction is often possible, information about the predictive state of the system is diluted over time, hindering long-term forecasting. We can demonstrate this effect by examining how prediction skill changes as we increase the `Tp` argument, the "time to prediction", defining the number of time steps into the future at which forecasts are made.  **rEDM** provides the `PredictInterval()` function to automate this task.

<!------------------------------------------------------------------------->
### Example
<!------------------------------------------------------------------------->

Using the same data with the `PredictInterval()` function, we supply the embedding dimension parameter with the value determined previously (`E = 2`):

```{r Prediction_interval_on_TentMap, fig.cap = "Tent map first differences simplex prediction skill as a function of forecast interval." }
rho_Tp <- PredictInterval( dataFrame = TentMap, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )
rho_Tp
```

As above, the returned object is a data.frame with forecast skill `rho` and time to prediction `Tp`.  As expected because the parameters chosen for the tent map fall in the region of chaotic behavior the decline in forecast skill as the forecast interval `Tp` increases (`rho` $\rightarrow 0$) indicates the system may be chaotic.

<!------------------------------------------------------------------------->
## Identifying Nonlinearity
<!------------------------------------------------------------------------->

One concern is that time series may show predictability even if they are purely stochastic, they behave similarly to autocorrelated red noise. Fortunately, we can distinguish between red noise and nonlinear deterministic behavior by using Sequential locally weighted global linear maps: *S-maps* as described in [@Sugihara_1994].

In contrast to the nearest-neighbor interpolation of simplex projection, the S-map forecasting method fits local linear maps to describe the dynamics. In addition to the standard set of parameters for a lagged-coordinate reconstruction as in simplex, S-maps contain a nonlinear localization parameter, $\theta$, that determines the degree to which points are weighted when fitting the local linear map. For example, when $\theta = 0$, all points are equally weighted, such that the local linear map is identical for different points in the reconstructed state-space. As such, the S-map will be identical to a global linear map (i.e. an autoregressive model). When values of $\theta$ are greater than $0$, nearby points in the state space receive larger weight, and the local linear map can vary in state-space to accommodate nonlinear behavior. 

Consequently, if the time series are sampled from autoregressive red noise, then the linear model ($\theta = 0$) should produce better forecasts, because the global linear map (which will, in effect, be fitted to more data points) will reduce the effects of observation error compared to local linear maps. In contrast, if forecast skill increases for $\theta > 0$, then the results are suggestive of nonlinear dynamics wherein better forecasts are achieved when the local linear map can change depending on the location in state-space: it is a better description of state-dependent behavior.

<!------------------------------------------------------------------------->
### Example
<!------------------------------------------------------------------------->

The `PredictNonlinear()` function provides an evaluation of S-map forecast skill as a function of the localisation parameter `theta`.  If unspecified, `theta` values will range from 0.01 to 9.

Typically, when using `S-map` to test for nonlinear behavior, we want to use all available points in the reconstruction of the local linear map, not just `knn` nearest neighbors as in simplex projection.  With all points available, S-map uses the `theta` parameter to control the weighting assigned to individual points, thereby localising the dynamics to capture nonlinear behavior.  When `knn = 0`, the default, `SMap()` will use all available points.

Here we use an embedding dimension of `E = 2` and the same parameters as in the previous examples, however, we specify the "TentMapNoise" data that adds Gaussian noise to the TentMap data as one would normally encounter with noisy observational data.

```{r Predict_nonlinear_on_TentMap, fig.cap = "Tent map first differences S-map prediction skill as a function of S-map localisation parameter."}
rho_theta <- PredictNonlinear( dataFrame = TentMapNoise, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )
rho_theta[c(1:5,11:15),]
```

The result is a data.frame with columns `Theta` and `rho`. Here, we see that forecast skill improves as `theta` increases, indicating the presence of nonlinear dynamics. We also observe degradation in forecast skill at high values of `theta` as the local linear map overfits to insufficient nearest neighbors.

<!------------------------------------------------------------------------->
## `Simplex()` and `SMap()` functions
<!------------------------------------------------------------------------->

The functions `EmbedDimension()`, `PredictInterval()` and `PredictNonlinear()` are parallelized wrapper functions for the `Simplex()` and `SMap()` algorithms. `EmbedDimension()` and `PredictInterval()` parallelize calls to `Simplex()` to evaluate forecast skill as a function of embedding dimension and prediction interval respectively.  `PredictNonlinear()` parallelizes calls to `SMap()` to assess predictive skill as a function of the nearest neighbor localization parameter. However, one can equivalently call the underlying `Simplex()` and `SMap()` functions directly.

<!------------------------------------------------------------------------->
### `Simplex()`
<!------------------------------------------------------------------------->

For example, evaluation of the simplex prediction at an optimal embedding dimension of `E = 2` can be performed as:

```{r Simplex_TentMap }
tentMapPredict <- Simplex( dataFrame = TentMap, lib = "1 100",
pred = "201 500", target = 'TentMap', columns = 'TentMap', E = 2 )

ComputeError( tentMapPredict $ Observations, tentMapPredict $ Predictions ) $ rho
```

<!------------------------------------------------------------------------->
### `SMap()`
<!------------------------------------------------------------------------->

An individual S-map evaluation corresponding to the optimal `PredictNonlinear()` result from above is:

```{r Tentmap_Noise_S-map } 
smap = SMap( dataFrame = TentMapNoise, lib = "1 100", pred = "201 500",
target = 'TentMap', columns = 'TentMap', E = 2, theta = 3 )
```

`SMap()` returns a named list with `predictions` and  `coefficients` data.frames, with `NaN` inserted appropriately where no predictions or observations are available. 

```{r S-map_tentMap_results}
head( cbind( smap $ predictions, smap $ coefficients ), 2 )
tail( cbind( smap $ predictions, smap $ coefficients ), 2 )
```

<!------------------------------------------------------------------------->
## Generalized Takens Theorem
<!------------------------------------------------------------------------->

A practical reality is that sampled observations of complex dynamics are usually composed of finite, noisy data.  Additionally, the presence of stochastic, non-deterministic drivers means that "multivariate" reconstructions can often be a better description than "univariate" reconstructions.  This means that in addition to creating an attractor from lags of one time series, it can be advantageous to combine different time series to create the phase-space embedding, provided they are all observed from the same system [@Sauer_1991; @Deyle_2011]. 

In **rEDM**, the `Simplex()` and `SMap()` functions allow multivariate reconstructions from any set of observation vectors. A multivariate reconstruction is defined by specifying which columns to use as coordinates in the `columns` argument, and which column is to be forecast in the `target` argument.  By default, **rEDM** will create Takens time-delay embeddings from univariate or multivariate data, however, this can be prevented by setting the `embedded` parameter `TRUE`. In this case, the input data are assumed to already constitute a valid multidimensional embedding, and no time-delay embedding is performed. 

<!------------------------------------------------------------------------->
### Example
<!------------------------------------------------------------------------->

We begin by examining an example dataset from a coupled 3-species model system.

```{r block_3sp_data}
head( block_3sp, 3 )
```

Here, `block_3sp` is a 10-column data.frame with 9 data columns. This data has already been time-delay embedded to dimension `E = 3` with a time delay of `tau = -1`. We use simplex forecasting based on a multivariate embedding of the three data vectors `x_t x_t-1 z_t` with `embedded = TRUE` and `E = 3`:

```{r 3_Species_simplex, fig.cap="Simplex Tp=1 forecast of $x_t$."}
smplx_3species = Simplex( dataFrame = block_3sp, lib = "1 100",
pred = "101 190", E = 3, columns = "x_t x_t-1 z_t", target = "x_t",
embedded = TRUE )
```

A plot of the predictions vs. observations can be examined with:

```{r plot_3_species, fig.cap="Scatter plot of simplex forecast of $x_t$ vs. observations.", fig.width=5, fig.height=4.5}
err = ComputeError( smplx_3species $ Observations,
                    smplx_3species $ Predictions )
plot( smplx_3species $ Observations, smplx_3species $ Predictions,
pch = 19, cex = 0.5, xlab = 'Observations', ylab = 'Predictions',
main = '3 Species x_t' )
abline( a = 0, b = 1, lty = 2, col = 'blue' )
text( -1, 1, paste( capture.output( cbind( err ) ), collapse = '\n' ) )
```

<!------------------------------------------------------------------------->
## S-map Coefficients
<!------------------------------------------------------------------------->

As described in [@Deyle_2016], S-map coefficients from the appropriate multivariate embedding can be interpreted as dynamic, time-varying interaction strengths. We demonstrate this with a chaotic timeseries described in [@Lorenz_1996], defined for N variables k=1, ... N, as 

\begin{equation*}
  \frac{ dx_{k} }{ dt } = -X_{k-2} X_{k-1} + X_{k-1} X_{k+1} - X_k + F
\end{equation*}

The `Lorenz5D` data.frame contains a N=5 dimensional system with F=8 from [@Lorenz_1996].  Here, we use `SMap()` to compute a 4-dimensional forecast at `Tp`=1:

```{r S-map_Lorenz96 }
smap_Lorenz <- SMap( dataFrame = Lorenz5D, lib = "1 500", pred = "601 900",
E = 4, theta = 3, columns = "V1 V2 V3 V4", target = "V1", embedded = TRUE )
```

As noted earlier, `SMap()` returns a named list with two data frames, `predictions` and `coefficients`:

```{r set_precision, echo = FALSE}
options(digits=4)
```

```{r S-map_Lorenz_results, warning = FALSE}
head( cbind( smap_Lorenz $ predictions, smap_Lorenz $ coefficients[,2:6] ), 3 )
```

Here, we plot the time series for the observed (blue) and predicted (red) values of `V1` in the top panel; and the inferred interactions (S-map coefficients) for the influence of `V4`, `V3` and `V2` on future values of `V1` in the lower panels.

```{r, echo = FALSE}
par(mfrow = c(4, 1), mar = c(2, 4, 0.5, 1), oma = c(0, 0, 0, 0),
    mgp = c(2.1, 1, 0) )
```

```{r smap_Lorenz_plot, out.width = "70%", fig.height = 4.5, fig.cap = "S-map prediction and coefficients of Lorenz'96 5-D system."}
predictions = smap_Lorenz $ predictions
coefficients = smap_Lorenz $ coefficients
Time = predictions $ Time

plot( Time, predictions $ Observations, type = "l", col = "blue",
ylab = "V1", xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
lines( Time, predictions $ Predictions, lwd = 2, col = 'red' )
legend("topright", legend = c("observed", "predicted"),
fill = c( 'blue', 'red' ), bty = "n", cex = 1.3 )

plot( Time, coefficients[, 6], type = "l", col = "brown",
ylab = expression(partialdiff * V4 / partialdiff * V1),
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
plot( Time, coefficients[, 5], type = "l", col = "darkgreen",
ylab = expression(partialdiff * V3 / partialdiff * V1),
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
plot( Time, coefficients[, 4], type = "l", col = "blue",
ylab = expression(partialdiff * V2 / partialdiff * V1),
xlab = "", lwd = 2, cex.lab = 1.3, cex.axis = 1.3 )
```

```{r, echo = FALSE}
par(mfrow = c(1, 1), mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))
```

\newpage
<!------------------------------------------------------------------------->
## Multiview Embedding
<!------------------------------------------------------------------------->

The generality of Takens Theorem means that in situations with multivariate time series, there can often be many different, valid attractor reconstructions. As described in [@Ye_2016], combining these different models can result in improved forecasts.

Here, we demonstrate this idea using the `Multiview()` function with the 3-species data used above. `Multiview()` operates by constructing all possible embeddings of dimension `E` with lag up to `E-1`.  These embeddings are ranked by forecast skill (`rho`) over the `lib` portion of the data.  The individual forecasts for the top `multiview` embeddings are then averaged together.  If `multiview` is not specified it is set to sqrt(C) where C is the number of E-dimensional combinations created from all data vectors.

```{r multiview}
Mview = Multiview( dataFrame = block_3sp, lib = "1 100", pred = "101 190", E = 3, columns = "x_t y_t z_t", target = "x_t" )
```

`Multiview()` returns a named list with two data.frames: `View`, and `Predictions`.  `View` lists the various combinations of data embedding vectors used for the forecasts along with their prediction statistics.  `Predictions` returns the final averaged multiview projections.

```{r multiview results}
Mview $ View[ which( Mview $ View $ rho > 0.91 ),  ]
```

\newpage
<!------------------------------------------------------------------------->
## Causality Inference and Cross Mapping
<!------------------------------------------------------------------------->

One of the corollaries to the Generalized Takens Theorem is that it should be possible to cross predict or cross map between variables that are observed from the same system. Consider two variables, $x$ and $y$ that interact in a dynamic system. Then the univariate reconstructions based on $x$ or $y$ alone should uniquely identify the system state and and thus the corresponding value of the other variable.

<!------------
   Figure 10
------------->
```{r fig_cross_map, echo = FALSE, fig.cap = "Cross Mapping Between Reconstructions of the Lorenz Attractor"}
knitr::include_graphics("CrossMap.png")
```
In the case of unidirectional causality, e.g. $x$ causes $y$, the causal variable ($x$) leaves a signature on the affected variable ($y$). Consequently, the reconstructed states based on $y$ can be used to cross predict the values of $x$ (because the reconstruction based on $y$ must be complete, it must include information about the value of $x$). Note that this cross prediction is in the *opposite* direction of the causal effect. At the same time, cross prediction from $x$ to $y$ will fail, because the time series of $x$ behaves independently of $y$, so a univariate reconstruction using only lags of $x$ is necessarily incomplete.

Although $x$ has incomplete information for predicting $y$, it does affect the values of $y$, and therefore will likely to have nonzero predictive skill. However, this cross mapping will be limited to the statistical association between $x$ and $y$ and will generally not improve as longer time series are used for reconstruction. In contrast, the cross prediction of $x$ from $y$ will generally improve. This convergence is therefore a crucial property for inferring causality. For practical reasons, the sensitivity of detecting causality this way is improved if, instead of predicting the future value of another variable, we estimate the concurrent value of another variable. We refer to this modified method as cross mapping, because we are not "predicting" the future.

For a more detailed description of using cross mapping to infer causation, see [@Sugihara_2012].

<!------------------------------------------------------------------------->
## Convergent Cross Mapping (CCM)
<!------------------------------------------------------------------------->

In **rEDM**, convergent cross mapping is implemented as the `CCM()` function, which provides a wrapper to compute cross map skill for different subsamples of the data. In the following example, we reproduce the analysis from [@Sugihara_2012] to identify causality between anchovy landings in California and Newport Pier sea-surface temperature. For this example, a previously identified value of `3` for the embedding dimension will be used.

To quantify convergence, we compute the cross map skill over many random subsamples of the time series. The `libSizes` argument specifies the size of the library set, and `sample` specifies the number of subsamples generated at each library size. `random` and `replacement` specify how the subsamples will be generated.  The default is random sampling without replacement.

```{r sardine_anchovy_ccm, fig.cap = "Convergent cross mapping of Newport sea surface temperature with anchovy landings.", fig.width=5, fig.height=4.5}
cmap <- CCM( dataFrame = sardine_anchovy_sst, E = 3, Tp = 0, columns = "anchovy", target = "np_sst", libSizes = "10 70 5", sample = 100, showPlot = TRUE )
```

The output is a data.frame with statistics for each model run (in this case, 100 models at each library size) as a function of library size.  Recalling that cross mapping indicates causal influence in the reverse direction (target to source), we see that the cross mapping `anchovy:np_sst` converges at a positive value of $\rho$, indicating that Newport sea surface temperature influences anchovy landings.  Because average cross map skill less than 0 means there is no prediction skill, (predictions should not be anticorrelated with observations), we infer from the `np_sst:anchovy` cross mapping that anchovy landings do not effect sea surface temperatures. 


<!------------------------------------------------------------------------->
# Real Data Example
<!------------------------------------------------------------------------->

<!------------------------------------------------------------------------->
## Apple-Blossom Thrips
<!------------------------------------------------------------------------->

In this example, we use EDM to re-examine the classic apple-blossom thrips (*Thrips imaginis*) time series from the Wait Institute in Australia [@Davidson_1948; @Davidson_1948a]. Seasonal outbreaks of *Thrips imaginis* were observed to vary greatly in magnitude from year to year, but large outbreaks tended to coincide across large spatial domains. This lead to the hypothesis that regional-scale climatic factors were responsible for controlling the size of the seasonal outbreaks (what might now be called the Moran effect [@Moran_1953]).

```{r Thrips data}
head(Thrips, 2)
```

The data column `Thrips_imaginis` contains counts of *Thrips imaginis* obtained from the Global Population Dynamics Database (GPDD) [@NERC-Centre-for-Population-Biology_2010]. `maxT_degC` is the mean maximum daily temperature ($^\circ$C) taken over each month and `Rain_mm` is the monthly rainfall (mm), both from the Waite Institute. The final column `Season` is a simple annual sinusoid that peaks in December (the Austral summer) to emulate an indicator of season.

First, we plot the data. Note that all the time-series variables, particularly the mean maximum daily temperature, show marked seasonality.

```{r thrips_plot, echo = FALSE, fig.width = 6, fig.height = 5, fig.cap = "Thrips abundance and environmental variables."}
par(mfrow = c(4, 1), mar = c(2, 4, 0, 1), mgp = c(2., 0.5, 0))
time_dec <- Thrips $ Year + (Thrips $ Month)/12
plot(time_dec, Thrips $ Thrips_imaginis, type = "l", lwd = 2, col = "darkgreen", ylab = "Thrips", xlab = "")
plot(time_dec, Thrips $ maxT_degC, type = "l", lwd = 2, col = "red", ylab = "maxT (C)", xlab = "")
plot(time_dec, Thrips $ Rain_mm, type = "l", lwd = 2, col = "blue", ylab = "Rain (mm)", xlab = "")
plot(time_dec, Thrips $ Season, type = "l", lwd = 2, col = "magenta", ylab = "Season", xlab = "" )
mtext("Year", side = 1, outer = TRUE, line = 1)
```

```{r, echo = FALSE }
par(mfrow = c(1, 1), mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))
```

<!------------------------------------------------------------------------->
### Univariate Analysis
<!------------------------------------------------------------------------->

We first examine the dependence of simplex predictability on the embedding dimension.

```{r univariate_thrips, fig.cap = "Simplex embedding dimension for Thrips abundance.", fig.width=5, fig.height=4.5}
rho_E <- EmbedDimension( dataFrame = Thrips, columns = "Thrips_imaginis", target = "Thrips_imaginis", lib = "1 72", pred = "1 72", showPlot = TRUE )
```

While there is an initial peak in the simplex prediction at `E = 3`, the global maximum is at `E = 8`. This suggests that both `E = 3` and `E = 8` are practical embedding dimensions, although `E = 8` is preferrable with a higher predictive skill. 

To test for nonlinearity we use the S-map `PredictNonlinear()` function.

```{r smap_thrips, fig.cap = "SMap localisation parameter for Thrips abundance.", fig.width = 5, fig.height = 4.5}
E = 8
rho_theta_e3 = PredictNonlinear(dataFrame = Thrips, columns = "Thrips_imaginis", target = "Thrips_imaginis", lib = "1 73", pred = "1 73", E = E, showPlot = TRUE )
```

The S-map results demonstrate clear nonlinearity in the Thrips time series, as nonlinear models `theta > 0` give substantially better predictions than the linear model `theta = 0`. This suggests that *Thrips*, despite the strong seasonal dynamics, do not simply track the environment passively, but have some intrinsic dynamics. To look more closely at the issue of seasonal drivers, however, we turn to convergent cross-mapping (CCM).

<!------------------------------------------------------------------------->
### Seasonal Drivers
<!------------------------------------------------------------------------->

Recall that there is a two-part criterion for CCM to be a rigorous test of causality:

1. The cross map prediction skill is statistically significant when using the full time series as the library.
2. Cross map prediction demonstrates convergence, i.e. prediction skill increases as more of the time series is used for the library and the reconstructed attractor becomes more dense.

For an initial summary, we first compute the cross map skill (measured with Pearsons $\rho$) for each variable pair. Note that `CCM()` computes the cross map in both "directions". 

```{r compute_ccm_matrix_thrips, results='hold', cache = TRUE}
vars       = colnames(Thrips[3:6])
var_pairs  = combn( vars, 2 )  # Combinations of vars, 2 at a time
libSize    = paste( NROW(Thrips) - E, NROW(Thrips) - E, 10, collapse = " " )
ccm_matrix = array(NA, dim = c(length(vars), length(vars)), 
                   dimnames = list(vars, vars))

for( i in 1:ncol( var_pairs ) ) {
   ccm_out = CCM(dataFrame = Thrips,
                 columns = var_pairs[1,i], target = var_pairs[2,i],
                 libSizes = libSize, Tp = 0, E = E, sample = 100)
                 
   outVars = names( ccm_out )

   var_out = unlist( strsplit( outVars[2], ':' ) )
   ccm_matrix[ var_out[2], var_out[1] ] = ccm_out[1,2]
                 
   var_out = unlist( strsplit( outVars[3], ':' ) )
   ccm_matrix[ var_out[2], var_out[1] ] = ccm_out[1,3]
}
```

We note that `ccm_matrix` rows are the second of the `CCM()` returned variables, while columns are the first variable.  As outlined earlier, influences are quantified from the `target` to the `columns` variable so that here, rows are considered the 'target`, influencing variable, and columns the `columns` influenced variable. 

For comparison we also compute the lagged cross-correlation, allowing lags of up to $\pm 6$ months.

```{r compute_corr_matrix_thrips, cache = TRUE}
corr_matrix <- array(NA, dim = c(length(vars), length(vars)), 
                     dimnames = list(vars, vars))

for(ccm_from in vars) {
    for(ccm_to in vars[vars != ccm_from]) {
        ccf_out <- ccf(Thrips[,ccm_from], Thrips[,ccm_to], 
                       type = "correlation", lag.max = 6, plot = FALSE)$acf
        corr_matrix[ccm_from, ccm_to] <- max(abs(ccf_out))
    }
}
```

We compare the two matrices.

```{r xmap_vs_corr_matrix_thrips}
ccm_matrix
corr_matrix
```

We can see that the cross map strengths are not symmetric.  In general CCM(X1 : X2) != CCM(X2 : X1). We also notice that the cross map and correlation between temperature and the seasonal indicator are high, with the cross map results suggesting that the seasonal variable can almost perfectly recover the temperature, $\rho = `r format(ccm_matrix["Season", "maxT_degC"], 2)`$. This makes interpretation more complicated, because we have to consider the possibility that cross mapping is simply identifying the shared seasonality between two time series. In other words, cross mapping between temperature and any variable with a seasonal cycle, might suggest an interaction even if there is no actual causal mechanism.

<!------------------------------------------------------------------------->
#### Convergent Cross-Mapping
<!------------------------------------------------------------------------->

With this in mind, we examine convergence in cross-map predictability, i.e. we compute `rho` as a function of library size `L`. The magnitude of the cross-correlation between *Thrips* and the cross mapped variable is shown as a black dashed line for comparison. 

```{r thrips_setup, echo = FALSE}
par( mfrow = c( 1, 3 ), mar = c( 3.5, 1.3, 2.3, 0.2 ), mgp = c(1.5, 0.5, 0) )

CCMPlot = function( ccm.df, E, ylim = c( 0.2, 1 ) ) {
    libSize = ccm.df $ LibSize
    V1      = names( ccm.df )[2]
    V2      = names( ccm.df )[3]

    title = paste( V1, " : ", V2, "\nE=" , E )
    
    print( title )
    print( libSize )

    plot( libSize, ccm.df[ , V1 ],
          ylim = ylim, main = title, col = "blue", type = "l", lwd = 3,
          xlab = 'Library Size', ylab = expression(Prediction~Skill~(rho)) )
    lines( libSize, ccm.df[ , V2 ], col = "red", lwd = 3 )
    abline( h = 0 )
    legend( 'topright', c( V1, V2 ), 
            fill = c( 'blue', 'red' ), bty = 'n', cex = 1.2 )
}
```

```{r ccm_thrips, fig.cap = "Thrips cross mapped to climatic variables. Vertical axis is cross map prediction skill (rho).", fig.width = 6.5, fig.height=5}
thrips_xmap_maxT <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                        columns = "Thrips_imaginis", target = "maxT_degC",
                        libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_maxT, E )
abline(h = corr_matrix['Thrips_imaginis', 'maxT_degC'], col = "black", lty = 2)

thrips_xmap_Rain <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                        columns = "Thrips_imaginis", target = "Rain_mm",
                        libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_Rain, E )
abline(h = corr_matrix['Thrips_imaginis', 'Rain_mm'], col = "black", lty = 2)

thrips_xmap_Season <- CCM(dataFrame = Thrips, E = E, Tp = 0,
                          columns = "Thrips_imaginis", target = "Season",
                          libSizes = "13 73 3", sample = 300, showPlot = FALSE)
CCMPlot( thrips_xmap_Season, E )
abline(h = corr_matrix['Thrips_imaginis', 'Season'], col = "black", lty = 2)
```

```{r thrips_setdown, echo = FALSE}
par(mar = c(4, 4, 2, 1), mgp = c(2.5, 1, 0))
```

The results show evidence of convergence for *Thrips* cross mapping to temperature and season variables, with the $\rho$ at maximum library size `L` significantly exceeding linear correlation. The rain variable does not indicate a substantially different cross map interaction, appearing confounded as to causal influence.

In addition, we are still left with the conundrum that temperature and to a lesser extent, rainfall, are easily predicted from the seasonal cycle, and so we cannot immediately ignore the possibility that the cross map results are an artifact of shared seasonal forcing.

To reframe, we wish to reject the null hypothesis that the level of cross mapping we obtain for `maxT_degC` and `Rain_mm` can be solely explained by shared seasonality. This hypothesis can be tested using randomization tests based on surrogate data. The idea is to generate surrogate time series with the same level of shared seasonality. Cross mapping between the real time series and these surrogates thus generates a null distribution for $\rho$, against which the actual cross map $\rho$ value can be compared.

<!------------------------------------------------------------------------->
#### Seasonal Surrogate Test
<!------------------------------------------------------------------------->

```{r seasonal_surrogates_thrips, cache = TRUE}
# Create matrix with temperature and rain surrogates (1000 time series vectors)
surr_maxT = SurrogateData( Thrips $ maxT_degC, method = "seasonal", T_period = 12, num_surr = 1000, alpha = 3)
surr_rain = SurrogateData( Thrips $ Rain_mm, method = "seasonal", T_period = 12, num_surr = 1000, alpha = 3)

# Rain cannot be negative
surr_rain = apply( surr_rain, 2, function(x){ i = which(x<0); x[i] = 0; x } )

# data.frame to hold CCM rho values between Thrips abundance and variable
rho_surr <- data.frame( maxT = numeric(1000), Rain = numeric(1000) )

# data.frames with time, Thrips, and 1000 surrogate climate variables for CCM()
maxT_data = as.data.frame( cbind( seq(1:nrow(Thrips)), Thrips $ Thrips_imaginis, surr_maxT ) )
names( maxT_data ) = c( 'time', 'Thrips_imaginis', paste( 'T', as.character( seq( 1,1000 ) ), sep = '' ) )

rain_data = as.data.frame( cbind( seq(1:nrow(Thrips)), Thrips $ Thrips_imaginis, surr_rain ) )
names( rain_data ) = c( 'time', 'Thrips_imaginis', paste( 'R', as.character( seq( 1,1000 ) ), sep = '' ) )

# Cross mapping
for (i in 1:1000) {
  targetCol = paste( 'T', i, sep = '' ) # as in maxT_data
  
  # Note this is a single sample CCM at full library size
  ccm_out = CCM( dataFrame = maxT_data, E = E, Tp = 0,
                 columns = 'Thrips_imaginis', target = targetCol,
                 libSizes = "73", sample = 1 )
                 
  col = paste( 'Thrips_imaginis', ':', targetCol, sep = '' )
   
  rho_surr $ maxT[i] = ccm_out[1,col]
}

for (i in 1:1000) {
  targetCol = paste( 'R', i, sep = '' ) # as in rain_data
  
  # Note this is a single sample CCM at full library size
  ccm_out = CCM( dataFrame = rain_data, E = E, Tp = 0,
                 columns = 'Thrips_imaginis', target = targetCol,
                 libSizes = "73", sample = 1 )
                 
  col = paste( 'Thrips_imaginis', ':', targetCol, sep = '' )
   
  rho_surr $ Rain[i] = ccm_out[1,col]
}
```

We now have a null distribution, and can estimate a $p$-value for rejecting the null hypothesis of mutual seasonality.

```{r significance_randomization_test}
1 - ecdf( rho_surr $ maxT )( ccm_matrix["maxT_degC", "Thrips_imaginis" ] )
1 - ecdf( rho_surr $ Rain )( ccm_matrix["Rain_mm",   "Thrips_imaginis" ] )
```

In the case of temperature, the CCM influence we estimated (`r ccm_matrix["maxT_degC", "Thrips_imaginis" ]`) is higher than the linear correlation (`r corr_matrix["maxT_degC", "Thrips_imaginis" ]`), and is highly significant in relation to a surrogate null distribution.  Regarding rainfall, the CCM influence (`r ccm_matrix["Rain_mm", "Thrips_imaginis" ]`) is higher than the linear correlate (`r corr_matrix["Rain_mm", "Thrips_imaginis" ]`), but not significant at the 96th percentile of the surrogate null distribution.  We note that the original Thrips data collections were at a much higher frequency than those available through the GPDD, and that monthly accumulated rainfall may be inadequate to resolve lifecycle influences on a species with a lifecycle of approximately one month. With more highly resolved data, it may well be possible to establish significance.

```{r reset_par, include = FALSE}
par( par.orig )
```

<!------------------------------------------------------------------------->
# Package Core
<!------------------------------------------------------------------------->
The **rEDM** package is implemented as a wrapper to the [`cppEDM`](https://github.com/SugiharaLab/cppEDM) library. All EDM algorithms are executed in the core `cppEDM` library and interfaced through the [`Rcpp`](https://cran.r-project.org/package=Rcpp) package.

<!------------------------------------------------------------------------->
## Data Input
<!------------------------------------------------------------------------->
Data can be input as an R `data.frame`, or read from a `.csv` file. In either case, the first column must define a time vector, and all columns are expected to be named. The time vector can be a string encoding a Date, or Datetime format.  All subsequent columns are expected to be numeric.

<!------------------------------------------------------------------------->
### S-Map coefficients and embedded data
<!------------------------------------------------------------------------->
`SMap()` should be called with a DataFrame that has columns explicity corresponding to dimensions `E`.  This means that if a multivariate data set is used, it should not be called with an embedding from `Embed()` since `Embed()` will add lagged coordinates for each variable.  These extra columns will then not correspond to the intended dimensions in the matrix inversion and prediction reconstruction and subsequent S-map coefficients. In this case, use the `embedded = TRUE` parameter with the multivariate data so that columns selected correspond to the proper dimension.

<!------------------------------------------------------------------------->
# Acknowledgements
<!------------------------------------------------------------------------->
**rEDM** is an evolving implementation of EDM code. Previous versions were developed by George Sugihara, Alan Trombla, Richard Penner, Victor Wong, Martin Casdagli, Jerome Cartagena, Mohsen Azarbayejani, Ava Pierce, Jennifer Trezzo, and Hao Ye.

<!------------------------------------------------------------------------->
# References
<!------------------------------------------------------------------------->

\name{PlotObsPred}
\alias{PlotObsPred}
\title{Plot observed versus predicted values}
\description{Base-graphics plot of Observations and Predictions against time.}
\usage{
PlotObsPred(df, dataName = "", E = NULL, Tp = NULL)
}
\arguments{
\item{df}{A data.frame with a time column plus Observations and Predictions.}
\item{dataName}{Optional label used in the plot title.}
\item{E}{Embedding dimension.}
\item{Tp}{Forecast interval (prediction horizon).}
}
\value{Invisibly \code{NULL}.}
\keyword{hplot}

\docType{data}
\name{block_3sp}
\alias{block_3sp}
\title{Time series for a three-species coupled model.}
\format{A data frame with 198 rows and 10 columns: 
\describe{
  \item{\code{time}}{time index (# of generations)}
  \item{\code{x_t}  }{abundance of simulated species x at time t}
  \item{\code{x_t-1}}{abundance of simulated species x at time t-1}
  \item{\code{x_t-2}}{abundance of simulated species x at time t-2}
  \item{\code{y_t}  }{abundance of simulated species y at time t}
  \item{\code{y_t-1}}{abundance of simulated species y at time t-1}
  \item{\code{y_t-2}}{abundance of simulated species y at time t-2}
  \item{\code{z_t}  }{abundance of simulated species z at time t}
  \item{\code{z_t-1}}{abundance of simulated species z at time t-1}
  \item{\code{z_t-2}}{abundance of simulated species z at time t-2}
}}
\usage{
block_3sp
}
\description{
Time series generated from a discrete-time coupled 
  Lotka-Volterra model exhibiting chaotic dynamics.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EDM_FileIO.R
\name{RequireE}
\alias{RequireE}
\title{Resolve / require the embedding dimension E.}
\usage{
RequireE(fn, E, embedded, columns)
}
\description{
pyEDM makes E a required argument for Simplex, SMap, PredictInterval and
PredictNonlinear. When embedded, E is inferred as length(columns) and any
user value is ignored; otherwise E must be supplied (non-NULL, >= 1).
}
\keyword{internal}

\name{PredictNonlinear}
\alias{PredictNonlinear}
\title{Prediction skill versus S-map localisation}
\description{
\code{\link{PredictNonlinear}} uses \code{\link{SMap}} to evaluate
prediction accuracy as a function of the state-space localization parameter
\code{theta}.
}
\usage{
PredictNonlinear(dataFrame = NULL, columns, target, lib, pred, theta = NULL, E = NULL,
                 Tp = 1, knn = 0, tau = -1, exclusionRadius = 0,
                 embedded = FALSE, validLib = logical(0), noTime = FALSE,
                 ignoreNan = TRUE, numProcess = 4, backend = "RANN",
                 pathIn = "./", dataFile = "", pathOut = "./", predictFile = "",
                 parameterList = FALSE, showPlot = FALSE, verbose = FALSE)
}
\arguments{
\item{dataFrame}{A data.frame of input data. The first column must be a time
index or time values unless \code{noTime = TRUE}. The columns must be named.}
\item{columns}{Column name(s) to build the embedding: character vector or space-separated string.}
\item{target}{Target column name to predict.}
\item{lib}{Library (training) index range as (start end) pairs.}
\item{pred}{Prediction index range as (start end) pairs.}
\item{theta}{S-map localisation parameter (\code{0} = global linear map).}
\item{E}{Embedding dimension. Required (no default); must be a positive integer unless \code{embedded = TRUE}, in which case it is inferred as the number of \code{columns}.}
\item{Tp}{Forecast interval (prediction horizon).}
\item{knn}{Number of nearest neighbours. \code{0} selects the method default.}
\item{tau}{Embedding delay (negative selects past lags).}
\item{exclusionRadius}{Temporal (Theiler) exclusion radius around each prediction point.}
\item{embedded}{If \code{TRUE}, \code{columns} already form the embedding.}
\item{validLib}{Logical vector marking admissible library rows (or length 0 for all).}
\item{noTime}{If \code{TRUE}, synthesise a 1..N time index instead of using column 1.}
\item{ignoreNan}{Remove rows with NaN in the embedding from the library and prediction sets.}
\item{numProcess}{Number of worker processes for the parameter sweep or task grid.}
\item{backend}{Nearest-neighbour backend: \code{"RANN"} (default) or \code{"brute"}.}
\item{pathIn}{File path for input \code{dataFile}.}
\item{dataFile}{Input dataFile, .csv format.}
\item{pathOut}{Output file path for \code{predictFile}}
\item{predictFile}{Output file name, .csv format.}
\item{parameterList}{Append named list of parameters/values to return.}
\item{showPlot}{If \code{TRUE}, draw a base-graphics plot of the result.}
\item{verbose}{Not used.}
}
\value{A data.frame with columns Theta and rho.}

\examples{
data(TentMapNoise)
theta.rho = PredictNonlinear( dataFrame = TentMapNoise, E = 2,
lib = "1 100", pred = "201 500", columns = "TentMap",
target = "TentMap", showPlot = FALSE )
}

\keyword{nonlinear}

\name{PredictInterval}
\alias{PredictInterval}
\title{Prediction skill versus forecast interval}
\description{
\code{\link{PredictInterval}} uses \code{\link{Simplex}} to evaluate
prediction accuracy as a function of forecast interval Tp.
}
\usage{
PredictInterval(dataFrame = NULL, columns, target, lib, pred, maxTp = 10, E = NULL,
                tau = -1, exclusionRadius = 0, embedded = FALSE,
                validLib = logical(0), noTime = FALSE, ignoreNan = TRUE,
                numProcess = 4, backend = "RANN", pathIn = "./",
                dataFile = "", pathOut = "./", predictFile = "",
                parameterList = FALSE, showPlot = FALSE)
}
\arguments{
\item{dataFrame}{A data.frame of input data. The first column must be a time
index or time values unless \code{noTime = TRUE}. The columns must be named.}
\item{columns}{Column name(s) to build the embedding: character vector or space-separated string.}
\item{target}{Target column name to predict.}
\item{lib}{Library (training) index range as (start end) pairs.}
\item{pred}{Prediction index range as (start end) pairs.}
\item{maxTp}{Largest forecast interval to evaluate.}
\item{E}{Embedding dimension. Required (no default); must be a positive integer unless \code{embedded = TRUE}, in which case it is inferred as the number of \code{columns}.}
\item{tau}{Embedding delay (negative selects past lags).}
\item{exclusionRadius}{Temporal (Theiler) exclusion radius around each prediction point.}
\item{embedded}{If \code{TRUE}, \code{columns} already form the embedding.}
\item{validLib}{Logical vector marking admissible library rows (or length 0 for all).}
\item{noTime}{If \code{TRUE}, synthesise a 1..N time index instead of using column 1.}
\item{ignoreNan}{Remove rows with NaN in the embedding from the library and prediction sets.}
\item{numProcess}{Number of worker processes for the parameter sweep or task grid.}
\item{backend}{Nearest-neighbour backend: \code{"RANN"} (default) or \code{"brute"}.}
\item{pathIn}{File path for input \code{dataFile}.}
\item{dataFile}{Input dataFile, .csv format.}
\item{pathOut}{Output file path for \code{predictFile}}
\item{predictFile}{Output file name, .csv format.}
\item{parameterList}{Append named list of parameters/values to return.}
\item{showPlot}{If \code{TRUE}, draw a base-graphics plot of the result.}
}
\value{A data.frame with columns \code{Tp, rho}.}
\examples{
data(TentMap)
Tp.rho = PredictInterval( dataFrame = TentMap, lib = "1 100",
pred = "201 500", E = 2, columns = "TentMap", target = "TentMap",
showPlot = FALSE )
}

\keyword{nonlinear}

\docType{data}
\name{TentMapNoise}
\alias{TentMapNoise}
\title{Time series of tent map plus noise.}
\format{Data frame with 999 rows and 2 columns
\describe{
  \item{\code{Time}}{time index.}
  \item{\code{TentMap}}{tent map values.}
}
}
\usage{
TentMapNoise
}
\description{
First-differenced time series generated from the tent map
  recurrence relation with mu = 2 and random noise.
}
\keyword{datasets}

\name{Simplex}
\alias{Simplex}
\title{Simplex projection}

\description{
\code{\link{Simplex}} performs time series forecasting based on
weighted nearest neighbors projection in the time series phase space as
described in \cite{Sugihara and May}.
}

\usage{
Simplex(dataFrame = NULL, columns, target, lib, pred, E = NULL, Tp = 1, knn = 0,
        tau = -1, exclusionRadius = 0, embedded = FALSE,
        validLib = logical(0), noTime = FALSE, ignoreNan = TRUE,
        backend = "RANN", pathIn = "./", dataFile = "", pathOut = "./",
        predictFile = "", parameterList = FALSE, includeState = FALSE,
        verbose = FALSE, showPlot = FALSE, .tieBreak = TRUE)
}

\arguments{
\item{dataFrame}{A data.frame of input data. The first column is time
  unless \code{noTime = TRUE}. The columns must be named.}
\item{columns}{Column name(s) to build the embedding: character vector string of whitespace separated column name(s), or vector
of column names used to create the library. If individual column names
contain whitespace place names in a vector, or, append ',' to the name.}
\item{target}{Target column name to predict.}
\item{lib}{Library (training) index range as (start end) pairs.  Mulitple row index
pairs can be specified with each pair defining the first and last
rows of time series observation segments used to create the library.}
\item{pred}{Prediction index range as (start end) pairs. A single contiguous range is supported.}
\item{E}{Embedding dimension. Required (no default); must be a positive integer unless \code{embedded = TRUE}, in which case it is inferred as the number of \code{columns}.}
\item{Tp}{Forecast interval (prediction horizon, number of time column rows).}
\item{knn}{Number of nearest neighbours. If \code{knn=0} knn is set to E+1..}
\item{tau}{Embedding delay (negative selects past lags).}
\item{exclusionRadius}{Temporal (Theiler) exclusion radius around each
  prediction point excludes vectors from the search space of nearest 
neighbors if their relative time index is within \code{exclusionRadius}.}
\item{embedded}{If \code{TRUE} \code{columns} already form the embedding.}
\item{validLib}{Logical vector marking admissible library rows (or length 0 for all).}
\item{noTime}{If \code{TRUE} synthesise a 1..N time index instead of using column 1.}
\item{ignoreNan}{Remove rows with NaN in the embedding from the library
  and prediction sets.}
\item{backend}{Nearest-neighbour backend: \code{"RANN"} (default) or \code{"brute"}.}
\item{pathIn}{File path for input \code{dataFile}.}
\item{dataFile}{Input dataFile, .csv format. The first column must be a time
index or time values unless noTime is TRUE. The first row must be column names.}
\item{pathOut}{Output file path for \code{predictFile}}
\item{predictFile}{Output file name, .csv format.}
\item{parameterList}{Append named list of parameters/values to return.}
\item{includeState}{If \code{TRUE}, also return an \code{internal} list of engine state: \code{knn_neighbors}, \code{knn_distances}, \code{lib_i}, \code{pred_i}, \code{targetVec}, \code{embedding} (1-based row indices; 0 = neighbour sentinel).}
\item{verbose}{Emit diagnostic messages.}
\item{showPlot}{If \code{TRUE} draw a base-graphics plot of the result.}
\item{.tieBreak}{If \code{TRUE} implement nearest neighbor tie breaking.}
}

\value{A data.frame with columns Time, Observations, Predictions and
  Pred_Variance.

  If \code{includeState = TRUE} a named list "internal" is added.
}

\examples{
data( block_3sp )
smplx = Simplex(block_3sp, "x_t", "x_t", lib = c(1, 100), pred = c(101, 195), E = 3)
ComputeError( smplx $ Predictions, smplx $ Observations )
}

\details{
  If embedded is \code{FALSE}, the data \code{column(s)} are embedded to
  dimension \code{E} with time lag \code{tau}. This embedding forms an
  E-dimensional phase space for the \code{\link{Simplex}} projection.
  If embedded is \code{TRUE}, the data are assumed to contain an
  E-dimensional embedding with E equal to the number of \code{columns}.
  Predictions are made using leave-one-out cross-validation, i.e.
  observation vectors are excluded from the prediction simplex. 

  To assess an optimal embedding dimension \code{\link{EmbedDimension}}
  can be applied. Accuracy statistics can be estimated by
  \code{\link{ComputeError}}.
}

\references{Sugihara G. and May R. 1990. Nonlinear forecasting as a way
of distinguishing chaos from measurement error in time series.
Nature, 344:734-741.
}

\keyword{nonlinear}

\name{SurrogateData}
\alias{SurrogateData}
\title{Surrogate time series for significance testing}
\description{Generate surrogate series (random_shuffle, ebisuzaki phase randomisation, or seasonal).}
\usage{
SurrogateData(ts, method = c("random_shuffle", "ebisuzaki", "seasonal"),
              num_surr = 100, T_period = 1, alpha = 0)
}
\arguments{
\item{ts}{Numeric time-series vector to generate surrogates for.}
\item{method}{Surrogate method: \code{"random_shuffle"}, \code{"ebisuzaki"}, or \code{"seasonal"}.}
\item{num_surr}{Number of surrogate series to generate.}
\item{T_period}{Seasonal period (seasonal method).}
\item{alpha}{Gaussian-noise standard deviation (seasonal method).}
}
\value{A numeric matrix with \code{length(ts)} rows and \code{num_surr} columns.}
\examples{
SurrogateData(sin(1:100), method = "ebisuzaki", num_surr = 5)
}
\keyword{datagen}

\docType{data}
\name{paramecium_didinium}
\alias{paramecium_didinium}
\title{Time series for the Paramecium-Didinium laboratory experiment}
\usage{
paramecium_didinium
}
\description{
Time series of Paramecium and Didinium abundances (#/mL) from 
an experiment by Veilleux (1979)
}
\keyword{datasets}

\name{CCM}
\alias{CCM}
\title{Convergent cross mapping}
\description{
  The state-space of a multivariate dynamical system (not a purely
  stochastic one) encodes coherent phase-space variable trajectories. If
  enough information is available, one can infer the presence or absence
  of cross-variable interactions associated with causal links between
  variables. \code{\link{CCM}} measures the extent to which states of
  variable Y can reliably estimate states of variable X. This can happen
  if X is causally influencing Y. 

  If cross-variable state predictability converges as more state-space
  information is provided, this indicates a causal link. \code{\link{CCM}}
  performs this cross-variable mapping using Simplex, with convergence
  assessed across a range of observational library sizes as described in
  \cite{Sugihara et al. 2012}.
}

\usage{
CCM(dataFrame = NULL, columns, target, E, Tp = 0, knn = 0, tau = -1,
    exclusionRadius = 0, libSizes, sample = 30, seed = NULL,
    embedded = FALSE, validLib = logical(0), includeData = FALSE,
    numProcess = 4, backend = "RANN", pathIn = "./", dataFile = "",
    pathOut = "./", predictFile = "", parameterList = FALSE, showPlot = FALSE)
}
\arguments{
\item{dataFrame}{A data.frame of input data. The first column must be a time
index or time values unless \code{noTime = TRUE}. The columns must be named.}
\item{columns}{Column name(s) to build the embedding: character vector
or space-separated string of columns to create the library. If
individual column names contain whitespace place names in a vector, or,
append ',' to the name.}
\item{target}{Target column name to predict.}
\item{E}{Embedding dimension.}
\item{Tp}{Forecast interval (prediction horizon).}
\item{knn}{Number of nearest neighbours. \code{0} selects the method default.}
\item{tau}{Embedding delay (negative selects past lags).}
\item{exclusionRadius}{Temporal (Theiler) exclusion radius around each prediction point.}
\item{libSizes}{Library sizes: string of 3 whitespace separated integer values
specifying the intial library size, the final library size,
and the library size increment. Can also be a list of strictly
 ncreasing library sizes.}
\item{sample}{Number of random subsamples per library size.}
\item{seed}{Random seed for reproducibility (not bit-comparable to pyEDM/NumPy).}
\item{embedded}{If \code{TRUE}, \code{columns} already form the embedding.}
\item{validLib}{Logical vector marking admissible library rows (or length 0 for all).}
\item{includeData}{Add per-library-size sample-variance columns to the result.}
\item{numProcess}{Number of worker processes for the parameter sweep or task grid.}
\item{backend}{Nearest-neighbour backend: \code{"RANN"} (default) or \code{"brute"}.}
\item{pathIn}{File path for input \code{dataFile}.}
\item{dataFile}{Input dataFile, .csv format. The first column must be a time
index or time values unless noTime is TRUE. The first row must be column names.}
\item{pathOut}{Output file path for \code{predictFile}}
\item{predictFile}{Output file name, .csv format.}
\item{parameterList}{Append named list of parameters/values to return.}
\item{showPlot}{If \code{TRUE}, draw a base-graphics plot of the result.}
}
\value{
  A data.frame with 3 columns. The first column is \code{LibSize}
  specifying the subsampled library size.  Columns 2 and 3 report
  Pearson correlation coefficients for the prediction of X from Y, and
  Y from X.

  If \code{includeData = TRUE} and \code{parameterList = TRUE} a
  named list "parameters" is added.
}

\references{Sugihara G., May R., Ye H., Hsieh C., Deyle E., Fogarty M., Munch S., 2012. Detecting Causality in Complex Ecosystems. Science 338:496-500.
}

\examples{
data(sardine_anchovy_sst)
df = CCM( dataFrame = sardine_anchovy_sst, E = 3, Tp = 0, columns = "anchovy",
target = "np_sst", libSizes = "10 70 10", sample = 100 )
}
\keyword{nonlinear}

\name{ComputeError}
\alias{ComputeError}
\title{Prediction skill statistics}
\description{Pearson rho, MAE, CAE and RMSE between observations and predictions.}
\usage{
ComputeError(obs, pred, digits = 6)
}
\arguments{
\item{obs}{Numeric vector of observations.}
\item{pred}{Numeric vector of predictions aligned with \code{obs}.}
\item{digits}{Rounding applied to each statistic.}
}
\value{A named list with elements rho, MAE, CAE and RMSE.}
\examples{
ComputeError(1:10 + 0, c(1:9 + 0, 8))
}
\keyword{regression}

\docType{data}
\name{sardine_anchovy_sst}
\alias{sardine_anchovy_sst}
\title{Time series for the California Current Anchovy-Sardine-SST system}
\format{\describe{
  \item{\code{year}}{year of measurement}
  \item{\code{anchovy}}{anchovy landings, scaled to mean = 0, sd = 1}
  \item{\code{sardine}}{sardine landings, scaled to mean = 0, sd = 1}
  \item{\code{sio_sst}}{3-year running average of sea surface temperature at 
    SIO pier, scaled to mean = 0, sd = 1}
  \item{\code{np_sst}}{3-year running average of sea surface temperature at 
    Newport pier, scaled to mean = 0, sd = 1}
}}
\usage{
sardine_anchovy_sst
}
\description{
Time series of Pacific sardine landings (CA), Northern anchovy 
  landings (CA), and sea-surface temperature (3-year average) at the SIO 
  pier and Newport pier
}
\keyword{datasets}

\docType{data}
\name{Thrips}
\alias{Thrips}
\title{Apple-blossom Thrips time series}
\description{
Seasonal outbreaks of Thrips imaginis.
}
\references{
Davidson and Andrewartha, Annual trends in a natural population of
Thrips imaginis \emph{Thysanoptera}, Journal of Animal Ecology, 17,
193-199, 1948.
}

\docType{package}
\name{rEDM}
\alias{EDM}
\alias{rEDM-package}
\title{Empirical dynamic modeling}
\description{
  \pkg{rEDM} provides tools for Empirical Dynamic Modeling time series
  analyses. It is based on reconstructing multivariate state space
  representations from uni- or multivariate time series, then projecting
  state changes using various metrics applied to nearest neighbors.
  
  Functionality includes:
  \itemize{
    \item Simplex projection (Sugihara and May 1990)
    \item Sequential Locally Weighted Global Linear Maps (S-map) (Sugihara 1994)
    \item Multivariate embeddings (Dixon et. al. 1999)
    \item Convergent cross mapping (Sugihara et. al. 2012)
    \item Multiview embedding (Ye and Sugihara 2016)
  }
}
\references{
  Sugihara G. and May R. 1990. Nonlinear forecasting as a way of
  distinguishing chaos from measurement error in time series.
  Nature, 344:734-741.

  Sugihara G. 1994. Nonlinear forecasting for the classification of
  natural time series. Philosophical Transactions: Physical Sciences
  and Engineering, 348 (1688) : 477-495.

  Dixon, P. A., M. Milicich, and G. Sugihara, 1999. Episodic
  fluctuations in larval supply. Science 283:1528-1530.

  Sugihara G., May R., Ye H., Hsieh C., Deyle E., Fogarty M.,
  Munch S., 2012. Detecting Causality in Complex Ecosystems.
  Science 338:496-500.

  Ye H., and G. Sugihara, 2016. Information leverage in
  interconnected ecosystems: Overcoming the curse of dimensionality.
  Science 353:922-925.
}
\details{

\strong{Main Functions}: 
  \itemize{
    \item \code{\link{Simplex}} - simplex projection
    \item \code{\link{SMap}} - S-map projection
    \item \code{\link{CCM}} - convergent cross mapping
    \item \code{\link{Multiview}} - multiview forecasting
  }
\strong{Helper Functions}: 
  \itemize{
    \item \code{\link{Embed}} - time delay embedding
    \item \code{\link{ComputeError}} - forecast skill metrics
    \item \code{\link{EmbedDimension}} - optimal embedding dimension
    \item \code{\link{PredictInterval}} - optimal prediction interval
    \item \code{\link{PredictNonlinear}} - evaluate nonlinearity
  }
}
\author{
\strong{Maintainer}: Joseph Park

\strong{Authors}: Joseph Park, Ethan Deyle, George Sugihara

% \strong{Contributors}: 
}
\keyword{package}

\name{Multiview}
\alias{Multiview}
\title{Multiview embedding forecast}
\description{Multiview embedding forecast (Ye & Sugihara 2016): rank
D-column views and average the best to find optimal combinations of
variables that best represent the dynamics.}
\usage{
Multiview(dataFrame = NULL, columns, target, lib = numeric(0), pred = numeric(0),
          D = 0, E = 1, Tp = 1, knn = 0, tau = -1, multiview = 0,
          exclusionRadius = 0, trainLib = TRUE, excludeTarget = FALSE,
          ignoreNan = TRUE, numProcess = 4, backend = "RANN",
          pathIn = "./", dataFile = "", pathOut = "./", predictFile = "",
          parameterList = FALSE, showPlot = FALSE)
}
\arguments{
\item{dataFrame}{A data.frame of input data. The first column is time unless \code{noTime = TRUE}.}
\item{columns}{Column name(s) to build the embedding: character vector
or space-separated string. If individual column names
contain whitespace place names in a vector, or, append ',' to the name.}
\item{target}{Target column name to predict.}
\item{lib}{Library (training) index range as (start end) pairs.}
\item{pred}{Prediction index range as (start end) pairs.}
\item{D}{View dimension for Multiview (number of embedding columns combined).}
\item{E}{Embedding dimension.}
\item{Tp}{Forecast interval (prediction horizon).}
\item{knn}{Number of nearest neighbours. \code{0} selects the method default.}
\item{tau}{Embedding delay (negative selects past lags).}
\item{multiview}{Number of top-ranked views to average (default \code{floor(sqrt(nCombo))}).}
\item{exclusionRadius}{Temporal (Theiler) exclusion radius around each prediction point.}
\item{trainLib}{Rank views on the library (in-sample) rather than the prediction set.}
\item{excludeTarget}{Exclude the target variable from candidate views.}
\item{ignoreNan}{Remove rows with NaN in the embedding from the library and prediction sets.}
\item{numProcess}{Number of worker processes for the parameter sweep or task grid.}
\item{backend}{Nearest-neighbour backend: \code{"RANN"} (default) or \code{"brute"}.}
\item{pathIn}{File path for input \code{dataFile}.}
\item{dataFile}{Input dataFile, .csv format.}
\item{pathOut}{Output file path for \code{predictFile}}
\item{predictFile}{Output file name, .csv format.}
\item{parameterList}{Append named list of parameters/values to return.}
\item{showPlot}{If \code{TRUE}, draw a base-graphics plot of the result.}
}
\value{
Named list with data.frames \code{[[View, Predictions]]}.

data.frame \code{View} columns:
\tabular{ll}{
  Col_1 \tab column index\cr
  ...   \tab column index\cr
  Col_D \tab column index\cr
  rho   \tab Pearson correlation\cr
  MAE   \tab mean absolute error\cr
  RMSE  \tab root mean square error\cr
  name_1 \tab column name\cr
  ...    \tab column name\cr
  name_D \tab column name\cr
}

  If \code{parameterList = TRUE} a named list "parameters" is added.
}

\references{Ye H., and G. Sugihara, 2016. Information leverage in 
interconnected ecosystems: Overcoming the curse of dimensionality.
Science 353:922-925.
}

\details{Multiview embedding is a method to identify variables in a
  multivariate dynamical system that are most likely to contribute to
  the observed dynamics.  It is a multistep algorithm with these general
  steps:
  \enumerate{
    \item Compute D-dimensional variable combination forecasts.
    \item Rank forecasts.
    \item Compute predictions of top combinations.
    \item Compute multiview averaged prediction.
  }
  If \code{E>1}, all variables are embedded to dimension E. 
  If \code{trainLib} is \code{TRUE} initial forecasts and ranking are
  done in-sample (\code{lib=pred}) and predictions using the top ranked
  combinations use the specified \code{lib} and \code{pred}.
  If \code{trainLib} is \code{FALSE} initial forecasts and ranking use
  the specified \code{lib} and \code{pred}, the step of computing
  predictions of the top combinations is skipped. 
}

\examples{
data(block_3sp)
L = Multiview( dataFrame = block_3sp, lib = "1 100", pred = "101 190",
E = 2, columns = "x_t y_t z_t", target = "x_t" )
}

\keyword{nonlinear}

\name{SMap}
\alias{SMap}
\title{Sequential locally weighted global linear map (S-map)}
\description{
  \code{\link{SMap}} performs time series forecasting based on localised
  (or global) nearest neighbor projection in the time series phase space as
  described in \cite{Sugihara 1994}. 
}
\usage{
SMap(dataFrame = NULL, columns, target, lib, pred, E = NULL, Tp = 1, knn = 0,
     tau = -1, theta = 0, exclusionRadius = 0, embedded = FALSE,
     validLib = logical(0), noTime = FALSE, ignoreNan = TRUE,
     backend = "RANN", pathIn = "./", dataFile = "", pathOut = "./",
     predictFile = "", parameterList = FALSE, includeState = FALSE,
     verbose = FALSE, showPlot = FALSE)
}
\arguments{
\item{dataFrame}{A data.frame of input data. The first column must be a time
index or time values unless \code{noTime = TRUE}. The columns must be named.}
\item{columns}{Column name(s) to build the embedding: character vector
or space-separated string, or vector
of column names used to create the library. If individual column names
contain whitespace place names in a vector, or, append ',' to the name.}
\item{target}{Target column name to predict.}
\item{lib}{string or vector with start and stop indices of input data
rows used to create the library from observations. Mulitple row index
pairs can be specified with each pair defining the first and last
rows of time series observation segments used to create the library.}
\item{pred}{string with start and stop indices of input data rows used for
predictions. A single contiguous range is supported.}
\item{E}{Embedding dimension. Required (no default); must be a positive integer unless \code{embedded = TRUE}, in which case it is inferred as the number of \code{columns}.}
\item{Tp}{Forecast interval (prediction horizon).}
\item{knn}{Number of nearest neighbours. \code{0} selects the method default.}
\item{tau}{Embedding delay (negative selects past lags).}
\item{theta}{S-map localisation parameter (\code{0} = global linear map).}
\item{exclusionRadius}{Temporal (Theiler) exclusion radius around each
prediction point. Excludes vectors from the search space of nearest 
neighbors if their relative time index is within exclusionRadius.}
\item{embedded}{If \code{TRUE}, \code{columns} already form the embedding.}
\item{validLib}{Logical vector marking admissible library rows (or length 0 for all).}
\item{noTime}{If \code{TRUE}, synthesise a 1..N time index instead of using column 1.}
\item{ignoreNan}{Remove rows with NaN in the embedding from the library and prediction sets.}
\item{pathIn}{File path for input \code{dataFile}.}
\item{dataFile}{Input dataFile, .csv format. The first column must be a time
index or time values unless noTime is TRUE. The first row must be column names.}
\item{pathOut}{Output file path for \code{predictFile}}
\item{predictFile}{Output file name, .csv format.}
\item{backend}{Nearest-neighbour backend: \code{"RANN"} (default) or \code{"brute"}.}
\item{parameterList}{Append named list of parameters/values to return.}
\item{includeState}{If \code{TRUE}, also return an \code{internal} list of engine state: \code{knn_neighbors}, \code{knn_distances}, \code{lib_i}, \code{pred_i}, \code{targetVec}, \code{embedding} (1-based row indices; 0 = neighbour sentinel).}
\item{verbose}{Emit diagnostic messages.}
\item{showPlot}{If \code{TRUE}, draw a base-graphics plot of the result.}
}
\value{
  A named list with three data.frames
  \code{[[predictions, coefficients, singularValues]]}.
  \code{predictions} has columns \code{Observations, Predictions}.
  The first column contains time or index values.

  \code{coefficients} data.frame has time or index values in the first column.
  Columns 2 through E+2 (E+1 columns) are the SMap coefficients.

  \code{singularValues} data.frame has time or index values in the first column.
  Columns 2 through E+2 (E+1 columns) are the SVD singularValues. The
  first value corresponds to the SVD bias (intercept) term.

  If \code{parameterList = TRUE} a named list "parameters" is added.

  If \code{includeState = TRUE} a named list "internal" is added.
}

\references{Sugihara G. 1994. Nonlinear forecasting for the classification of natural time series. Philosophical Transactions: Physical Sciences and Engineering, 348 (1688):477-495.}

\details{
  If \code{embedded} is \code{FALSE}, the data \code{column(s)} are embedded
  to dimension \code{E} with time lag \code{tau}. This embedding forms an
  n-columns * E-dimensional phase space for the \code{\link{SMap}} projection.
  If embedded is \code{TRUE}, the data are assumed to contain an
  E-dimensional embedding with E equal to the number of \code{columns}.
  See the Note below for proper use of multivariate data (number of
  \code{columns} > 1).

  If \code{ignoreNan} is \code{TRUE}, the library (\code{lib}) is
  internally redefined to exclude nan embedding vectors. If
  \code{ignoreNan} is \code{FALSE} no library adjustment is made. The
  (\code{lib}) can be explicitly specified to exclude nan library vectors.
  
  Predictions are made using leave-one-out cross-validation, i.e.
  observation rows are excluded from the prediction regression.

  In contrast to \code{\link{Simplex}}, \code{\link{SMap}} uses all
  available neighbors and weights them with an exponential decay
  in phase space distance with exponent \code{theta}. \code{theta}=0
  uses all neighbors corresponding to a global autoregressive model.
  As \code{theta} increases, neighbors closer in vicinity to the
  observation are considered. 
}

\note{
\code{\link{SMap}} should be called with columns explicitly corresponding to
dimensions E. In the univariate case (number of \code{columns} = 1) with
default \code{embedded = FALSE}, the time series will be time-delay
embedded to dimension E, SMap coefficients correspond to each dimension. 

If a multivariate data set is used (number of \code{columns} > 1) it
must use \code{embedded = TRUE} with E equal to the number of columns.
This prevents the function from internally time-delay embedding the
multiple columns to dimension E.  If the internal time-delay embedding
is performed, then state-space columns will not correspond to the
intended dimensions in the matrix inversion, coefficient assignment,
and prediction. In the multivariate case, the user should first prepare
the embedding (using \code{\link{Embed}} for time-delay embedding), then
pass this embedding to \code{\link{SMap}} with appropriately specified
\code{columns}, \code{E}, and \code{embedded = TRUE}.
}

\keyword{nonlinear}

\name{Embed}
\alias{Embed}
\title{Time-delay embedding}
\description{Takens time-delay embedding of one or more columns, returned as a
named data.frame matching pyEDM's \code{Embed()}.}
\usage{
Embed(dataFrame = NULL, E = 0, tau = -1, columns = "",
      includeTime = FALSE, pathIn = "./", dataFile = "")
}
\arguments{
\item{dataFrame}{A data.frame of input data. Column 1 is the time vector.}
\item{E}{Embedding dimension.}
\item{tau}{Embedding delay in time-column rows (negative selects past lags).}
\item{columns}{Column name(s) to embed: a whitespace-separated string, or a
character vector of column names.}
\item{includeTime}{If \code{TRUE}, prepend column 1 of \code{dataFrame} (the
time vector), keeping its original name, as the first output column.}
\item{pathIn}{Directory for \code{dataFile} (used when reading input from disk).}
\item{dataFile}{CSV file name to read as input; when set it takes precedence
over \code{dataFrame}.}
}
\value{A data.frame of the embedding (with the time column first when
\code{includeTime = TRUE}).}
\details{
  Each \code{columns} item produces E lagged vectors. For column \code{c} and
  step \code{i = 0..E-1} the lag is \code{i * |tau|}, and the column is named
  \code{c(t-lag)} when \code{tau < 0} or \code{c(t+lag)} when \code{tau > 0}.
  For example columns X, Y with E = 2, tau = -1 yield
  \code{X(t-0) X(t-1) Y(t-0) Y(t-1)}; with tau = -2 they yield
  \code{X(t-0) X(t-2) Y(t-0) Y(t-2)}.

  If \code{dataFile} is supplied the CSV at \code{file.path(pathIn, dataFile)}
  is read (preserving column names) and used as the input, taking precedence
  over \code{dataFrame}.

  All rows are retained; the first \code{|tau| * (E-1)} partial rows contain
  \code{NA}.
}
\examples{
Embed(data.frame(x = 1:10), E = 3, tau = -1, columns = "x")
}
\keyword{nonlinear}

\docType{data}
\name{Lorenz5D}
\alias{Lorenz5D}
\title{5-D Lorenz'96}
\format{Data frame with 1000 rows and 6 columns
\describe{
  \item{\code{Time}}{Time.}
  \item{\code{V1}}{variable 1.}
  \item{\code{V2}}{variable 2.}
  \item{\code{V3}}{variable 3.}
  \item{\code{V4}}{variable 4.}
  \item{\code{V5}}{variable 5.}
}
}
\usage{
Lorenz5D
}
\description{5-D Lorenz'96 timeseries with F = 8.
}
\references{
Lorenz, Edward (1996). Predictability - A problem partly solved,
Seminar on Predictability, Vol. I, ECMWF.
}
\keyword{datasets}

\docType{data}
\name{EvergladesFlow}
\alias{EvergladesFlow}
\title{Water flow to NE Everglades}
\format{A data frame with 1379 rows and 2 columns: 
\describe{
  \item{\code{Date}}{Date.}
  \item{\code{S12CD_S333_CFS}}{Cumulative weekly flow (CFS).}
}
}
\usage{
EvergladesFlow
}
\description{
  Cumulative weekly water flow into northeast Everglades from water
  control structures S12C, S12D and S333 from 1980 through 2005.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EDM_Plot.R
\name{PlotCCM}
\alias{PlotCCM}
\title{CCM cross-map skill versus library size, both directions.}
\usage{
PlotCCM(df, E = NULL)
}
\arguments{
\item{df}{CCM result data.frame (LibSize plus two rho columns).}

\item{E}{embedding dimension, shown in the title.}
}
\value{
Invisibly \code{NULL}.
}
\description{
CCM cross-map skill versus library size, both directions.
}
\keyword{internal}

\docType{data}
\name{circle}
\alias{circle}
\title{2-D timeseries of a circle.}
\format{A data frame with 200 rows and 3 columns: 
\describe{
  \item{\code{Time}}{time index.}
  \item{\code{x}}{sin component.}
  \item{\code{y}}{cos component.}
}}
\usage{
circle
}
\description{
Time series of of circle in 2-D (sin and cos).
}
\keyword{datasets}

\name{PlotSmap}
\alias{PlotSmap}
\title{Plot S-map predictions and coefficients}
\description{Base-graphics multi-panel plot of S-map predictions and coefficient series.}
\usage{
PlotSmap(SmapList, dataName = "", E = NULL, Tp = NULL)
}
\arguments{
\item{SmapList}{A list returned by \code{SMap} (predictions and coefficients).}
\item{dataName}{Optional label used in the plot title.}
\item{E}{Embedding dimension.}
\item{Tp}{Forecast interval (prediction horizon).}
}
\value{Invisibly \code{NULL}.}
\keyword{hplot}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EDM_Plot.R
\name{PlotSweep}
\alias{PlotSweep}
\title{rho versus a swept parameter (EmbedDimension / PredictInterval / ...).}
\usage{
PlotSweep(df, xlab, title = NULL)
}
\description{
rho versus a swept parameter (EmbedDimension / PredictInterval / ...).
}
\keyword{internal}
